import * as basket from "../eaiws/basket";
import type { CalculationSheetInfo } from "../eaiws/basket";
import { GroupedMap } from "../utils";
import type { BaseItemManager } from "./BaseItemManager";
export declare const ItemUpdateFlags: {
    readonly None: 0;
    readonly Item: 1;
    readonly ItemProps: 2;
    readonly Image: 4;
    readonly Calculation: 8;
    readonly Tax: 16;
    readonly ConfigData: 32;
    readonly UserImage: 64;
    readonly ExtItemNumber: 128;
    readonly UserItemDescription: 256;
    readonly UserArticleDescription: 256;
    readonly AdditionalImages: 512;
    readonly Comments: 1024;
    readonly IncludeChildren: 2147483648;
    /** UserFields = UserImage | ExtItemNumber | UserItemDescription | AdditionalImages */
    readonly UserFields: number;
    readonly All: 4294967295;
};
export type ItemUpdateFlags = number;
export interface EditableLabelItem {
    setLabel(pLabel: string): Promise<void>;
}
export interface ArticleImageCameraSetup {
    alpha: number;
    beta: number;
    fov: number;
    mode: "Perspective" | "Orthographic";
}
export type ItemInactiveState = "Alternative" | "Optional" | "None";
export declare abstract class BaseItem<T extends BaseItem<T, M>, M extends BaseItemManager<T, M>> {
    get id(): string;
    get label(): string;
    get type(): basket.BasketItemType;
    get item(): basket.BasketItem;
    get parent(): T | null;
    get children(): ReadonlyArray<T>;
    /** Index of this item inside the children array of its parent. */
    get childIndex(): number;
    get isSelected(): boolean;
    get isDisposed(): boolean;
    /**
     * Can be used to store runtime user/custom data. This data is not persistent.
     * Use BasketService.setItemAppData() for persistent data.
     */
    get userData(): GroupedMap;
    get isGroup(): boolean;
    /**
     * Returns true if this item represents an configurable OFML article.
     */
    get isOfmlArticle(): boolean;
    constructor(pManager: M, pParent: T | null, pChildIndex: number, pEaiwsItem: basket.BasketItem);
    canEditLabel(): this is EditableLabelItem;
    /**
     * Returns true if this item is excluded from the calculations for the currently active
     * pricing procedure (e.g. its an alternative or optional item). Returns undefined if the state is
     * unknown or its an unsupported item. Undefined will be also returned if the ItemProperties
     * are not available. get or fetchItemProperties() should be called before to ensure that they are available.
     */
    isExcludedFromCalculations(): boolean | undefined;
    /**
     * Returns if this item is marked as alternative or optional for the currently active pricing procedure.
     * Returns undefined if the state is unknown or its an unsupported item. Undefined will be also
     * returned if the ItemProperties are not available. get or fetchItemProperties() should be
     * called before to ensure that they are available.
     */
    getInactiveState(): ItemInactiveState | undefined;
    /**
     * Note: The setArticleId which is part of the ItemProperties may not be up-to-date.
     * Use this.item.setArticleId or this.getSetArticle() instead.
     */
    getItemProperties(): Promise<basket.ItemProperties>;
    getCachedItemProperties(): basket.ItemProperties | undefined;
    /**
     * Fetches the item properties if necessary and adds them to the cache.
     * @param pIncludeChildren If true the item properties of child items will be fetched too.
     * @param pShouldFetchCallback Can be used to ignore certain items. The callback has to return false in this case.
     */
    fetchItemProperties(pIncludeChildren: boolean, pShouldFetchCallback?: (pItem: T) => boolean): Promise<void>;
    getImageUrl(): Promise<string | null>;
    getCachedImageUrl(): string | null | undefined;
    getUserImageUrl(): Promise<string | null>;
    getCachedUserImageUrl(): string | null | undefined;
    /**
     * Fetches the user defined images if necessary and adds them to the cache.
     * @param pIncludeChildren If true the user image of child items will be fetched too.
     * @param pShouldFetchCallback Can be used to ignore certain items. The callback has to return false in this case.
     */
    fetchUserImageUrls(pIncludeChildren: boolean, pShouldFetchCallback?: (pItem: T) => boolean): Promise<void>;
    /**
     * Returns the additional images for this item.
     * Note: `undefined` images should be handled as missing/broken images
     */
    getAdditionalImageUrls(): Promise<Array<string | undefined>>;
    /**
     * Returns the additional images for this item from cache.
     * Note: `undefined` images should be handled as missing/broken images
     */
    getCachedAdditionalImageUrls(): Array<string | undefined> | undefined;
    /**
     * Fetches the user defined additional images if necessary and adds them to the cache.
     * @param pIncludeChildren If true the additional images of child items will be fetched too.
     * @param pShouldFetchCallback Can be used to ignore certain items. The callback has to return false in this case.
     */
    fetchAdditionalImageUrls(pIncludeChildren: boolean, pShouldFetchCallback?: (pItem: T) => boolean): Promise<void>;
    /**
     * Returns the item calculation.
     * @param pIgnoreHeaderCalculation If true the cached calculation will be returned even if the
     *      header calculation has changed in the meantime. Every condition which depends on a header condition
     *      may have a wrong/outdated value in this case. (default: false)
     */
    getCalculation(pIgnoreHeaderCalculation?: boolean): Promise<CalculationSheetInfo>;
    /**
     * Returns the cached item calculation.
     * @param pIgnoreHeaderCalculation If true the cached calculation will be returned even if the
     *      header calculation has changed in the meantime. Every condition which depends on a header condition
     *      may have a wrong/outdated value in this case. (default: false)
     */
    getCachedCalculation(pIgnoreHeaderCalculation?: boolean): CalculationSheetInfo | undefined;
    /**
     * Fetches the calculations if necessary and adds them to the cache.
     * @param pIncludeChildren If true the calculations of child items will be fetched too.
     * @param pShouldFetchCallback Can be used to ignore certain items. The callback has to return false in this case.
     * @param pIgnoreHeaderCalculation If true the cached calculation will be returned even if the
     *      header calculation has changed in the meantime. Every condition which depends on a header condition
     *      may have a wrong/outdated value in this case. (default: false)
     */
    fetchCalculations(pIncludeChildren: boolean, pShouldFetchCallback?: (pItem: T) => boolean, pIgnoreHeaderCalculation?: boolean): Promise<void>;
    getExtItemNumber(): Promise<string | null>;
    getCachedExtItemNumber(): string | null | undefined;
    /**
     * Fetches the external item numbers if necessary and adds them to the cache.
     * @param pIncludeChildren If true the item number of child items will be fetched too.
     * @param pShouldFetchCallback Can be used to ignore certain items. The callback has to return false in this case.
     */
    fetchExtItemNumbers(pIncludeChildren: boolean, pShouldFetchCallback?: (pItem: T) => boolean): Promise<void>;
    getComments(): Promise<string | null>;
    getCachedComments(): string | null | undefined;
    /**
     * Fetches the comments if necessary and adds them to the cache.
     * @param pIncludeChildren If true the comments of child items will be fetched too.
     * @param pShouldFetchCallback Can be used to ignore certain items. The callback has to return false in this case.
     */
    fetchComments(pIncludeChildren: boolean, pShouldFetchCallback?: (pItem: T) => boolean): Promise<void>;
    invalidateGroupCalculations(): void;
    getParentGroup(): T | null;
    updateItem(pFlags: ItemUpdateFlags): Promise<void>;
    isChildOf(pParent: T, pDirectOnly: boolean): boolean;
    getPrevChild(pChild: T): T | null;
    getNextChild(pChild: T): T | null;
    getItemNumber(): string;
    protected mManager: M;
    protected mParent: T | null;
    protected mItem: basket.BasketItem;
    protected mImageUrl: string | null | undefined;
    protected mUserImageUrl: string | null | undefined;
    protected mExtItemNumber: string | null | undefined;
    protected mComments: string | null | undefined;
    protected mAdditionalImageUrls: Array<string | undefined> | undefined;
    protected mItemProps: basket.ItemProperties | undefined;
    protected mCalculation: CalculationSheetInfo | undefined;
    protected mCalculationHdrVersion: number;
    protected mChildren: Array<T>;
    protected mChildIndex: number;
    protected mItemNumber: string | undefined;
    protected mItemNumberArray: Array<number> | undefined;
    protected mUserData: GroupedMap;
    protected mIsSelected: boolean;
    protected mIsDisposed: boolean;
    protected abstract readonly asType: T;
    protected abstract registerItem(): void;
    protected invalidateCalculation(pTriggerUpdateEvent: boolean): void;
    protected isCachedCalculationValid(pIgnoreHeaderCalculation: boolean): boolean;
    protected getGeneratedImage(pCamSetup: ArticleImageCameraSetup | null): Promise<string | null>;
    protected getCatalogImage(): Promise<string | null>;
    protected getStoredImage(): Promise<basket.ImageInfo | null>;
    protected getCameraSetupFromImage(pImage: basket.ImageInfo): ArticleImageCameraSetup | null;
    protected generateImageUrl(): Promise<void>;
    protected setItemProperties(pProps: basket.ItemProperties, pUpdateItem: boolean, pUpdateCalculation: boolean, pAdditionalUpdateFlags?: ItemUpdateFlags): Promise<void>;
    protected internalGetSetArticle(): T | null;
    protected performUpdates(pFlags: ItemUpdateFlags): Promise<void>;
    protected handleChangedItemProps(): void;
    private useItemPropsFromCache;
    private useCalculationFromCache;
    private getImageAttributeValue;
    private setParent;
    private invalidateItemNumber;
    private fetchItemFields;
    private collectItemsToFetch;
    private isValidImageUrl;
    private getInactivePositionState;
}
