import * as basket from "../eaiws/basket";
import type { EaiwsSession } from "../eaiws";
import { ListenableEvent } from "../utils";
import { ArticleItem } from "./ArticleItem";
import { BaseItemManager } from "./BaseItemManager";
import { BasketItem } from "./BasketItem";
import { CalculationManager } from "./CalculationManager";
import { ProjectInfo } from "./ProjectInfo";
import { View } from "./View";
export type EventDataItemsPasted = Array<{
    /** The original copied item. */
    originalItem: BasketItem | null;
    /** The pasted item */
    item: BasketItem;
}>;
export interface LoadFromSessionOptions {
    /** If a name is provided, the pricing procedure will be activated during session loading. */
    pricingProcedureName?: string;
}
export declare class Basket extends BaseItemManager<BasketItem, Basket> {
    get calculation(): CalculationManager;
    get views(): ReadonlyArray<View>;
    get projectInfo(): ProjectInfo;
    /**
     * Defines the image resolution for server generated images. (default: 512)
     */
    get generatedImageResolution(): number;
    set generatedImageResolution(pValue: number);
    /**
     * If true calculation line errors will be also included in the `inconsistencyReason` of the article `ItemProperties`.
     * @defaultValue true
     */
    includeCalculationErrorsInInconsistencies: boolean;
    readonly eventOfmlArticleUpdated: ListenableEvent<ArticleItem, Basket>;
    readonly eventOfmlUpdateStateChanged: ListenableEvent<ArticleItem, Basket>;
    /**
     * Will be triggered everytime if items got pasted into the basket. (e.g. copy/paste or OEX import)
     */
    readonly eventItemsPasted: ListenableEvent<EventDataItemsPasted, Basket>;
    /**
     * Will be triggered if this basket instance gets modified.
     */
    readonly eventModified: ListenableEvent<void, Basket>;
    /**
     * Will be triggered from `Basket.updateSessionLanguage()`
     */
    readonly eventUpdateSessionLanguage: ListenableEvent<void, Basket>;
    /**
     * Will be triggered from `Basket.clear()`
     */
    readonly eventClear: ListenableEvent<void, Basket>;
    /**
     * Creates a new Basket instance.
     * @param pSession Session the basket should operate on.
     * @param pProjectInfo Optional pre-created `ProjectInfo` instance. Will be ignored if `pSession` and `pProjectInfo.session` are not equal.
     */
    constructor(pSession: EaiwsSession, pProjectInfo?: ProjectInfo);
    loadFromSession(pOptions?: LoadFromSessionOptions): Promise<void>;
    /**
     * Should be called if the locale and or data language of the eaiws session has changed to
     * update all language related data.
     */
    updateSessionLanguage(): Promise<void>;
    clear(): void;
    addView(pViewIdOrConfig: string | basket.BasketViewConfig): Promise<View>;
    /** Removes a view and disposes it. */
    removeView(pView: View): void;
    fetchOfmlUpdateStates(pItems: ReadonlyArray<BasketItem>): Promise<void>;
    invalidateHeaderCalculation(): void;
    /**
     * Checks the consistency of all articles.
     * Returns false if at least one of the articles or its configuration is inconsistent.
     * Note: Errors of invisible calculation lines will be ignored.
     * @param pCheckCalculation If true the calculation will be also checked for error. (default: true)
     */
    checkConsistency(pCheckCalculation?: boolean): Promise<boolean>;
    /**
     * Create necessary basket columns if they don't exist yet in the provided session.
     * Note: This function will be also called from `Basket.loadFromSession()`.
     */
    static PrepareBasketColumns(pSession: EaiwsSession): Promise<void>;
    protected createItem(pParent: BasketItem | null, pChildIndex: number, pEaiwsItem: basket.BasketItem): BasketItem;
    private static readonly FETCH_UPDATE_STATE_CHUNK_SIZE;
    private static readonly DEFAULT_GENERATED_IMAGE_OPTIONS;
    private mCalcManager;
    private mViews;
    private mProjectInfo;
    private mGeneratedImageResolution;
    private mGeneratedImageOptions;
    private onNotifyModified;
    private setGeneratedImageResolution;
}
