import { ListenableEvent } from "../utils";
import type { Basket } from "./Basket";
import type { BasketItem } from "./BasketItem";
/**
 * Flags which are used to lock certain item related operations.
 */
export declare const ItemLockFlags: {
    /** nothing is locked */
    readonly None: 0;
    /** locks the delete operation => item can't be deleted */
    readonly Delete: 1;
    /** locks the configuration of an ofml article => article can't be configured */
    readonly Configuration: 2;
    /** locks the calculation of an item => item can't be calculated */
    readonly Calculation: 4;
    /** locks the item properties => properties like texts or article number can't be changed */
    readonly ItemProperties: 8;
    /** locks the item image => the image of an item can't be changed */
    readonly Image: 16;
    /** locks the article quantity => the quantity of an article can't be changed */
    readonly Quantity: 32;
    /** locks the ofml article update => updating an ofml article is not possible */
    readonly OfmlArticleUpdate: 64;
    /** locks all operations */
    readonly All: 4294967295;
};
export type ItemLockFlags = number;
/**
 * The ItemLockManager is used to lock certain item related operations.
 * The operations are locked only on UI side. The API will still allow these operations.
 */
export declare class ItemLockManager {
    /** will be triggered if the item locks have changed */
    readonly eventLocksChanged: ListenableEvent<void, ItemLockManager>;
    /** list of currently locked items (basket item ids) and there lock flags */
    get lockedItems(): ReadonlyMap<string, ItemLockFlags>;
    /**
     * Creates a new `ItemLockManager` instance
     * @param pLockedItems Map from basket item id to lock flags.
     */
    constructor(pBasket: Basket, pLockedItems?: ReadonlyMap<string, ItemLockFlags>);
    /** Should be called to clean-up the ItemLockManager if it is no longer used. */
    dispose(): void;
    /** Sets a new list of locked items. */
    setLockedItems(pLockedItems: ReadonlyMap<string, ItemLockFlags>): void;
    /** Returns true if the given lock flag is set on `pItem`. */
    isItemLocked(pItem: BasketItem, pFlags: ItemLockFlags): boolean;
    /** Returns true if the flag `ItemLockFlags.Delete` is set for one of the given items. */
    isDeleteLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    /** Returns true if the flag `ItemLockFlags.Configuration` is set for one of the given items. */
    isConfigurationLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    /** Returns true if the flag `ItemLockFlags.Calculation` is set for one of the given items. */
    isCalculationLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    /** Returns true if the flag `ItemLockFlags.Image` is set for one of the given items. */
    isImageLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    /** Returns true if the flag `ItemLockFlags.Quantity` is set for one of the given items. */
    isQuantityLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    /** Returns true if the flag `ItemLockFlags.OfmlArticleUpdate` is set for one of the given items. */
    isOfmlArticleUpdateLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    /** Returns true if the flag `ItemLockFlags.Calculation` is set for one of the locked items. */
    isHeaderCalculationLocked(): boolean;
    /** Returns true if the flag `ItemLockFlags.ItemProperties` is set for one of the given items. */
    areItemPropertiesLocked(pItems: ReadonlyArray<BasketItem>): boolean;
    private mBasket;
    private mLockedItems;
    private onItemDisposed;
    private areItemsLocked;
    private areSetPartsLocked;
    private areChildsLocked;
}
