import { ListenableEvent } from "../utils";
import { ArticleItem } from "./ArticleItem";
/**
 * Flags which are used to lock certain item related operations.
 */
export const ItemLockFlags = {
    /** nothing is locked */
    None: 0x00000000,
    /** locks the delete operation => item can't be deleted */
    Delete: 0x00000001,
    /** locks the configuration of an ofml article => article can't be configured */
    Configuration: 0x00000002,
    /** locks the calculation of an item => item can't be calculated */
    Calculation: 0x00000004,
    /** locks the item properties => properties like texts or article number can't be changed */
    ItemProperties: 0x00000008,
    /** locks the item image => the image of an item can't be changed */
    Image: 0x00000010,
    /** locks the article quantity => the quantity of an article can't be changed */
    Quantity: 0x00000020,
    /** locks the ofml article update => updating an ofml article is not possible */
    OfmlArticleUpdate: 0x00000040,
    /** locks all operations */
    All: 0xFFFFFFFF
};
/**
 * The ItemLockManager is used to lock certain item related operations.
 * The operations are locked only on UI side. The API will still allow these operations.
 */
export class ItemLockManager {
    /** list of currently locked items (basket item ids) and there lock flags */
    get lockedItems() {
        return (this.mLockedItems);
    }
    /**
     * Creates a new `ItemLockManager` instance
     * @param pLockedItems Map from basket item id to lock flags.
     */
    constructor(pBasket, pLockedItems = new Map()) {
        /** will be triggered if the item locks have changed */
        this.eventLocksChanged = new ListenableEvent(this);
        this.onItemDisposed = (pItem) => {
            if (this.mLockedItems.delete(pItem.id)) {
                this.eventLocksChanged.trigger();
            }
        };
        this.mBasket = pBasket;
        this.setLockedItems(pLockedItems);
        this.mBasket.eventItemDisposed.addListener(this.onItemDisposed);
    }
    /** Should be called to clean-up the ItemLockManager if it is no longer used. */
    dispose() {
        this.mBasket.eventItemDisposed.removeListener(this.onItemDisposed);
    }
    /** Sets a new list of locked items. */
    setLockedItems(pLockedItems) {
        this.mLockedItems = new Map();
        pLockedItems.forEach((pFlags, pItemId) => {
            var _a;
            const tItem = this.mBasket.getItemById(pItemId);
            if (tItem != null) {
                if (tItem instanceof ArticleItem) {
                    //only add main articles
                    const tMainArticle = tItem.getMainArticle();
                    //merge with current flags
                    const tCurrentFlags = (_a = this.mLockedItems.get(tMainArticle.id)) !== null && _a !== void 0 ? _a : ItemLockFlags.None;
                    this.mLockedItems.set(tMainArticle.id, tCurrentFlags | pFlags);
                }
                else {
                    this.mLockedItems.set(pItemId, pFlags);
                }
            }
        });
        this.eventLocksChanged.trigger();
    }
    /** Returns true if the given lock flag is set on `pItem`. */
    isItemLocked(pItem, pFlags) {
        //check main article only
        if (pItem instanceof ArticleItem) {
            const tMainArticle = pItem.getMainArticle();
            if (tMainArticle !== pItem) {
                return (this.isItemLocked(tMainArticle, pFlags));
            }
        }
        const tLockedFlags = this.mLockedItems.get(pItem.id);
        if (tLockedFlags == null)
            return (false);
        return ((tLockedFlags & pFlags) !== 0);
    }
    /** Returns true if the flag `ItemLockFlags.Delete` is set for one of the given items. */
    isDeleteLocked(pItems) {
        if (this.areItemsLocked(pItems, ItemLockFlags.Delete)) {
            return (true);
        }
        //check set and folders
        for (const tItem of pItems) {
            if (tItem.isGroup && this.areChildsLocked(tItem, ItemLockFlags.Delete)) {
                return (true);
            }
            //special case: if the calculation of a set part is locked we are not allowed to delete the set (quantity will change)
            if (tItem.type === "SetArticle" && this.areSetPartsLocked(tItem, ItemLockFlags.Calculation)) {
                return (true);
            }
        }
        return (false);
    }
    /** Returns true if the flag `ItemLockFlags.Configuration` is set for one of the given items. */
    isConfigurationLocked(pItems) {
        return (this.areItemsLocked(pItems, ItemLockFlags.Configuration));
    }
    /** Returns true if the flag `ItemLockFlags.Calculation` is set for one of the given items. */
    isCalculationLocked(pItems) {
        if (this.areItemsLocked(pItems, ItemLockFlags.Calculation)) {
            return (true);
        }
        //check set and folders
        for (const tItem of pItems) {
            if (tItem.type === "SetArticle" && this.areSetPartsLocked(tItem, ItemLockFlags.Calculation)) {
                return (true);
            }
            if (tItem.type === "Folder" && this.areChildsLocked(tItem, ItemLockFlags.Calculation)) {
                return (true);
            }
        }
        return (false);
    }
    /** Returns true if the flag `ItemLockFlags.Image` is set for one of the given items. */
    isImageLocked(pItems) {
        return (this.areItemsLocked(pItems, ItemLockFlags.Image));
    }
    /** Returns true if the flag `ItemLockFlags.Quantity` is set for one of the given items. */
    isQuantityLocked(pItems) {
        return (this.areItemsLocked(pItems, ItemLockFlags.Quantity));
    }
    /** Returns true if the flag `ItemLockFlags.OfmlArticleUpdate` is set for one of the given items. */
    isOfmlArticleUpdateLocked(pItems) {
        return (this.areItemsLocked(pItems, ItemLockFlags.OfmlArticleUpdate));
    }
    /** Returns true if the flag `ItemLockFlags.Calculation` is set for one of the locked items. */
    isHeaderCalculationLocked() {
        let tLocked = false;
        this.mLockedItems.forEach((pFlags, pItemId) => {
            if ((pFlags & ItemLockFlags.Calculation) !== 0) {
                tLocked = true;
                return;
            }
        });
        return (tLocked);
    }
    /** Returns true if the flag `ItemLockFlags.ItemProperties` is set for one of the given items. */
    areItemPropertiesLocked(pItems) {
        return (this.areItemsLocked(pItems, ItemLockFlags.ItemProperties));
    }
    areItemsLocked(pItems, pFlags) {
        if (pItems.length === 0 || pFlags === ItemLockFlags.None)
            return (false);
        for (const tItem of pItems) {
            if (this.isItemLocked(tItem, pFlags)) {
                return (true);
            }
        }
        return (false);
    }
    areSetPartsLocked(pSet, pFlags) {
        const tParts = pSet.getSetParts();
        for (const tPart of tParts) {
            if (this.isItemLocked(tPart, pFlags)) {
                return (true);
            }
        }
        return (false);
    }
    areChildsLocked(pParent, pFlags) {
        for (const tChild of pParent.children) {
            if (this.isItemLocked(tChild, pFlags) || this.areChildsLocked(tChild, pFlags)) {
                return (true);
            }
        }
        return (false);
    }
}
