import * as basket from "../../eaiws/basket";
import type * as catalog from "../../eaiws/catalog";
import { ArticleItem } from "../ArticleItem";
import { FolderItem } from "../FolderItem";
import type { ItemInactiveState } from "../BaseItem";
import { SetArticleItem } from "../SetArticleItem";
import type { TextItem } from "../TextItem";
import { UserArticleItem } from "../UserArticleItem";
import type { Basket } from "../Basket";
import type { BasketItem } from "../BasketItem";
import type { ItemLockManager } from "../ItemLockManager";
export type EventDataItemsPasted = Array<{
    /** The original copied item. */
    originalItem: BasketItem | null;
    /** The pasted item */
    item: BasketItem;
}>;
export interface InsertFolderOptions {
    /** @defaultValue false */
    showSubTotal: boolean;
}
export interface DeleteItemsOptions {
    /** @defaultValue true */
    wholeComposite: boolean;
    /** @defaultValue true */
    children: boolean;
}
export interface CopyItemsOptions {
    /** @defaultValue true */
    wholeComposite: boolean;
    /** @defaultValue true */
    children: boolean;
    /** @defaultValue false */
    cut: boolean;
}
export interface PasteItemsOptions {
    /**
     * If specified, the items will be pasted as children of the parent.
     * If not the insert position is based on the current selection.
     * @defaultValue null
     */
    parent: BasketItem | null;
    /**
     * If specified, the items will be pasted before the specified item.
     * Will be ignored if parent is not specified.
     * @defaultValue null
     */
    beforeItem: BasketItem | null;
}
export interface MoveItemsOptions {
    /**
     * If true the items will reordered during move to match the order specified by the given item array.
     * This option will be ignored if the copy option is true.
     * @defaultValue false
     */
    reorder: boolean;
    /**
     * If true the items will be copied to the new location, otherwise they are moved.
     * @defaultValue false
     */
    copy: boolean;
}
export interface MergeArticlesOptions {
    /**
     * If the given item list contains folders, `folderMode` defines how to handle them.
     * - Ignore: folders will be ignored, only articles contained in the item list will be merged
     * - Delete: folders will be deleted but their children will not be deleted (break up folder structure)
     * - MergeChildren: every child article of a folder will also be considered for merging
     * @defaultValue "Ignore"
     */
    folderMode: "Ignore" | "Delete" | "MergeChildren";
}
export interface UpdateOfmlArticleError {
    item: ArticleItem;
    error: unknown;
}
export interface UpdateOfmlArticlesResult {
    updated: number;
    invalid: number;
    migratable: number;
    errors: Array<UpdateOfmlArticleError>;
}
export declare function insertArticle(pBasket: Basket, pArticle: catalog.ArticleCatalogItem | basket.InsertInfo, pOptions?: basket.InsertArticleOptions): Promise<ArticleItem>;
export declare function insertContainerItem(pBasket: Basket, pContainer: catalog.CatalogItem): Promise<Array<BasketItem>>;
export declare function insertFolder(pBasket: Basket, pName: string, pOptions?: Partial<InsertFolderOptions>): Promise<FolderItem>;
export declare function insertSetArticle(pBasket: Basket, pName: string): Promise<SetArticleItem>;
export declare function insertUserArticle(pBasket: Basket, pShortText: string): Promise<UserArticleItem>;
export declare function insertTextItem(pBasket: Basket, pText: string): Promise<TextItem>;
export declare function deleteItems(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pOptions?: Partial<DeleteItemsOptions>): Promise<void>;
/**
 * Copies the given items into an OBX file. Returns the url of the OBX file.
 */
export declare function copyItems(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pOptions?: Partial<CopyItemsOptions>): Promise<string>;
/**
 * Inserts the items which are part of the given OBX. Only eaiws session urls are accepted here.
 * Use importObx() for external files.
 */
export declare function pasteItems(pBasket: Basket, pObxUrl: string, pOptions?: Partial<PasteItemsOptions>): Promise<Array<BasketItem>>;
export declare function relocateItems(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pNewParent: BasketItem, pBeforeItem: BasketItem | null, pOptions?: Partial<MoveItemsOptions>): Promise<Array<BasketItem>>;
/**
 * Moves the given items in the desired direction if possible.
 * `getPossibleMoveDirections()` should be used in advance to check if a movement is possible.
 */
export declare function moveItems(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pDirection: basket.MoveItemsDirection): Promise<basket.MoveItemsResult>;
/**
* Returns all currently possible move directions for the given items.
*/
export declare function getPossibleMoveDirections(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pLockManager?: ItemLockManager): Set<basket.MoveItemsDirection>;
/** @deprecated  Use `setItemInactiveState` instead. */
export declare function setItemAlternativeState(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pAlternative: boolean): Promise<void>;
export declare function setItemInactiveState(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pNewState: ItemInactiveState): Promise<void>;
export declare function convertToUserArticles(pBasket: Basket, pItems: ReadonlyArray<BasketItem>): Promise<Array<UserArticleItem>>;
/**
 * Splits up composite articles into individual articles.
 * Returns the new items which originally represented the main article.
 */
export declare function splitUpCompositeArticles(pBasket: Basket, pItems: ReadonlyArray<BasketItem>): Promise<Array<BasketItem>>;
export declare function convertToSetArticle(pBasket: Basket, pFolder: FolderItem): Promise<SetArticleItem>;
export declare function breakUpSetArticle(pBasket: Basket, pSetArticle: SetArticleItem): Promise<FolderItem>;
/**
 * Tries to merge the given articles. Returns true if articles were merged.
 */
export declare function mergeArticles(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pOptions?: Partial<MergeArticlesOptions>): Promise<boolean>;
/**
 * Updates or migrates the given article items.
 * If the optional `pProgressCallback` returns false, the process will be aborted.
 * Errors during article update will be collected and returned as part of the result. They are not thrown as an exception.
 */
export declare function updateOfmlArticles(pBasket: Basket, pItems: ReadonlyArray<BasketItem>, pMigrate: boolean, pProgressCallback?: (pFinshed: number, pTotal: number) => boolean, pPriceDate?: basket.IsoDateString): Promise<UpdateOfmlArticlesResult>;
/**
 * Imports the given Obx file.
 * Returns null if the file upload was aborted.
 */
export declare function importObx(pBasket: Basket, pObxData: ArrayBuffer | Blob, pOnUploadProgress?: (pEvent: ProgressEvent, pXhr: XMLHttpRequest) => any): Promise<Array<BasketItem> | null>;
/**
 * Imports the given pec file.
 * Returns null if the file upload was aborted.
 */
export declare function importPec(pBasket: Basket, pPecData: ArrayBuffer | Blob, pOnUploadProgress?: (pEvent: ProgressEvent, pXhr: XMLHttpRequest) => any): Promise<Array<BasketItem> | null>;
export declare function cloneItem(pBasket: Basket, pItem: BasketItem): Promise<BasketItem>;
