import * as catalog from "../../eaiws/catalog";
import type { ArticleElement, ArticleManager, MainArticleElement } from "../../cf";
import type { CameraZoomToFitOptions, View } from "../../core/view";
import { ListenableEvent } from "../../utils";
import { ArticleItem } from "../ArticleItem";
import type { ArticleImageCameraSetup } from "../BaseItem";
import { ViewArticleItem } from "../ViewArticleItem";
import type { Basket } from "../Basket";
import { MultiArticlePropertyProvider } from "./MultiArticlePropertyProvider";
export interface ArticleConfigurationOptions {
    /**
     * If a view is given its camera orientation will be initialized for the article.
     */
    view?: View;
    /**
     * If the options and a view are defined, zoom to fit will be performed.
     */
    zoomToFitOptions?: CameraZoomToFitOptions;
    /**
     * If true the basket will be updated for every article change. Otherwise it will be updated only if finish() is called.
     */
    liveBasketUpdate?: boolean;
}
export interface AssignArticleImageCameraOptions {
    /** If true, the camera will be assigned to all articles part of the composite. */
    wholeComposite?: boolean;
    /**
     * Updates the article images immediately (camera setup only persists after updating the image).
     * note: only supported if configuration was started with `liveBasketUpdate` enabled
     */
    updateArticleImages?: boolean;
    /** Optional progress handler for the image update. False should be returned to abort the process.*/
    onUpdateImagesProgress?: (pFinshed: number, pTotal: number) => boolean;
}
export type RestorePreConfigurationFunc = (pArticleManager: ArticleManager, pOptions?: ArticleConfigurationOptions) => Promise<Array<MainArticleElement>>;
export declare class ArticleConfiguration {
    get started(): boolean;
    get isPreConfiguration(): boolean;
    get articleManager(): ArticleManager | null;
    /**
     * Will be triggered everytime the basket (or its items) has changed.
     */
    readonly eventBasketChanged: ListenableEvent<void, ArticleConfiguration>;
    constructor(pBasket: Basket);
    /**
     * Start configuration for an existing article. Configuration can't be aborted in this case.
     */
    start(pArticleManager: ArticleManager, pArticle: ArticleItem | ViewArticleItem, pOptions?: ArticleConfigurationOptions): Promise<Array<MainArticleElement>>;
    /**
     * Start configuration for a new article. Configuration can be aborted.
     * The liveBasketUpdate option is not supported.
     */
    start(pArticleManager: ArticleManager, pArticle: catalog.ArticleCatalogItem, pOptions?: ArticleConfigurationOptions): Promise<Array<MainArticleElement>>;
    /**
    * Start configuration of multiple articles.
    */
    startMultiConfiguration(pArticles: ReadonlyArray<ArticleItem>): Promise<MultiArticlePropertyProvider>;
    /**
     * Assign camera view to article which is used to generate the image for the ArticleItem.
     * If pView is null the default camera setup will be used (can be used to reset the article camera).
     */
    assignArticleImageCamera(pArticle: ArticleElement, pView: View | null, pOptions?: AssignArticleImageCameraOptions): Promise<void>;
    /**
     * Returns the currently assigned camera setup for the given article.
     */
    getArticleImageCamera(pArticle: ArticleElement): ArticleImageCameraSetup | null;
    /**
     * Resets the camera of the given view to the initial camera setup for the current article.
     */
    resetView(pView: View, pZoomToFitOptions?: CameraZoomToFitOptions): void;
    /**
     * Finish configuration.
     */
    finish(): Promise<void>;
    /**
     * Aborts the configuration process. Only supported for new articles.
     */
    abort(): Promise<void>;
    /**
     * If an OBK was saved during pre configuration of an catalog item, the session still
     * contains this item. This function should be called before calling `Basket.loadFromSession()`
     * to delete the item from the session prior to loading.
     * A function to restore the configuration state of this item be returend in this case.
     * If this function is not called no special handling for this item will be performed.
     * It will be handled as a regular basket item.
     */
    loadPreConfigurationFromSession(): Promise<RestorePreConfigurationFunc | null>;
    private static readonly APP_KEY;
    private static readonly APP_DATA_PRECONFIGURATION_ITEM;
    private static readonly DEFAULT_CAMERA_SETUP;
    private static readonly MAX_CONCURRENT_IMAGE_UPDATES;
    private mBasket;
    private mStarted;
    private mLiveBasketUpdate;
    private mArticleManager;
    private mApp;
    private mMultiPropProvider;
    private mIsPreConfiguration;
    private mMainParent;
    private mMainItem;
    private mPreConfigMainItemId;
    private mParentItemsToUpdate;
    private mChangedMainElements;
    private mRestoredInstances;
    private mAssignedCameras;
    private mOldFatherItemId;
    private mOldPricingProcedure;
    private mOldIncludeCalculationErrorsInInconsistencies;
    private onUpdateSessionLanguage;
    private onClear;
    private onArticleChanged;
    private onArticleAdded;
    private onArticleRemoved;
    private onArticleInstanceRestored;
    private onModelModified;
    private reset;
    private clearTrackedChanges;
    private registerEvents;
    private deregisterEvents;
    private startForArticle;
    private collectArticleBasketItems;
    private restorePreConfiguration;
    private startPreConfiguration;
    private setPreConfigurationItemId;
    private getPreConfigurationItemId;
    private setupCamera;
    private addAssignedCamera;
    private fillMainElementCache;
    private fillBasketCache;
    private fillBasketConfigDataCache;
    private addArticleDataToCache;
    private addConfigDataToCache;
    private synchronizeBasket;
}
