import * as basket from "../eaiws/basket";
import type * as catalog from "../eaiws/catalog";
import type { Application } from "../core/Application";
import type { BaseScreenshotOptions } from "../core/view/Viewer";
import type { EaiwsSession } from "../eaiws";
import type { PricingProcedureInfo } from "../eaiws/basket";
import { ListenableEvent } from "../utils";
import { MainArticleElement } from "./MainArticleElement";
import type { PlaceholderCreator } from "./PlaceholderCreator";
import type { ArticleElement } from "./ArticleElement";
export declare class ArticleChangedEventData {
    article: ArticleElement;
    wholeComposite: boolean;
    constructor(pArticle: ArticleElement, pWholeComposite: boolean);
}
export interface OapErrorEventData {
    error: Error;
    article: ArticleElement;
    interactor?: basket.OAPInteractor;
}
/**
     * Article screenshot generation options.
     */
export interface ArticleScreenshotOptions extends BaseScreenshotOptions {
    /** margin for the zoom (default: 0.05 (5%)) */
    margin?: number;
    /** use orthographic camera mode (default: false) */
    orthographic?: boolean;
    /** renders only the main article if true (default: false) */
    hideSubArticles?: boolean;
    /** alpha camera rotation (y-axis) in degree (default: -15) */
    cameraAlpha?: number;
    /** beta camera rotation (x-axis) in degree (default: 20) */
    cameraBeta?: number;
    /** camera field of view in degree (default: 30) */
    cameraFov?: number;
}
export interface ArticleImportOptions {
    /**
     * If this callback is given the ofml update state of each article will be checked and if it is
     * Updatable or Migratable this callback will be called. If true is returned from the callback
     * the article will be Updated/Migrated. (default: undefined)
     */
    shouldUpdateCallback?: (pItemProps: basket.ItemProperties) => boolean;
    /**
     * If true articles which do not have a position will be also imported and placed at
     * an automatically determined position, otherwise they will be ignored. (default: false)
     */
    importArticlesWithoutPosition?: boolean;
}
export interface InitializeSessionOptions {
    /**
     * If true the OAP planning mode will be enabled.
     * Should be used only for applications with planning support. (default: false)
     */
    oapPlanningMode?: boolean;
}
export declare class ArticleManager {
    get app(): Application;
    get session(): EaiwsSession;
    get maxGfjTextureSize(): number;
    /**
     * The currently active pricing procedure.
     */
    get pricingProcedure(): PricingProcedureInfo | null;
    /**
     * Optional father item which will be used as father/parent in the basket hierarchy for
     * newly inserted/created articles.
     */
    fatherItemId: string | null;
    /**
     * Placeholder creator for articles without geometry. Can be overriden with a custom implementation.
     * The DefaultPlaceholderCreator is used as the default implementation.
     */
    placeholderCreator: PlaceholderCreator;
    /**
     * If true calculation line errors will be also included in the `inconsistencyReason` of the article `ItemProperties`.
     * @defaultValue true
     */
    includeCalculationErrorsInInconsistencies: boolean;
    readonly eventArticleChanged: ListenableEvent<ArticleChangedEventData, ArticleManager>;
    /**
     * Will be triggered if an article was added to the model.
     */
    readonly eventArticleAdded: ListenableEvent<ArticleElement, ArticleManager>;
    /**
     * Will be triggered if an article was removed/deleted from the model.
     */
    readonly eventArticleRemoved: ListenableEvent<ArticleElement, ArticleManager>;
    /**
     * Will be triggered if an article instance (basket id) was restored.
     */
    readonly eventArticleInstanceRestored: ListenableEvent<MainArticleElement, ArticleManager>;
    /**
     * Will be triggered for every error during execution of OAP actions.
     */
    readonly eventOapError: ListenableEvent<OapErrorEventData, ArticleManager>;
    constructor(pApp: Application, pEaiwsSession: EaiwsSession);
    /**
     * Has to be called for every eaiws session to enable new OAP features.
     * If it is not called the EAIWS will run in compatibility mode.
     */
    initializeSession(pOptions?: InitializeSessionOptions): Promise<void>;
    /**
     * Inserts a new article into the EAIWS session and creates a MainArticleElement.
     * Note: The MainArticleElement will not be added to the ModelManager here. This should be done afterwards.
     */
    insertArticle(pArticle: catalog.ArticleCatalogItem | basket.InsertInfo, pOptions?: basket.InsertArticleOptions): Promise<MainArticleElement>;
    /**
     * Inserts pArticle as a sub article of pMainArticle and returns the basket id of the new sub article.
     */
    insertSubArticle(pMainArticle: MainArticleElement, pArticle: catalog.ArticleCatalogItem, pOptions?: basket.InsertArticleOptions): Promise<string>;
    /**
     * Creates a MainArticleElement for the given basket id which should belong to an already
     * existing main article in the EAIWS session.
     * Note: The MainArticleElement will not be added to the ModelManager here. This should be done afterwards.
     * @param pBasketId The id of the article to load.
     * @param pLoadTransformations If true the transformations will be loaded. (Default: false)
     * @param pBasketItems If the basket.BasketItems of the articles are already known. They can be passed here to speed up the operation.
     * The list of items has to be complete. It has to conatin all main articles and all of its sub articles.
     * Further, the geometryId and setArticleId (if the item is part of a set) of each BasketItem should be available.
     */
    createArticleById(pBasketId: string, pLoadTransformations?: boolean, pBasketItems?: Array<basket.BasketItem>): Promise<MainArticleElement>;
    /**
     * Creates MainArticleElements for the given basket ids which should belong to already
     * existing main articles in the EAIWS session.
     * Note: The MainArticleElements will not be added to the ModelManager here. This should be done afterwards.
     * @param pBasketIds The ids of the articles to load.
     * @param pLoadTransformations If true the transformations will be loaded. (default: false)
     * @param pBasketItems If the basket.BasketItems of the articles are already known. They can be passed here to speed up the operation.
     * The list of items has to be complete. It has to conatin all main articles and all of its sub articles.
     * Further, the geometryId and setArticleId (if the item is part of a set) of each BasketItem should be available.
     * @param pGfjUrls Map from basketId to gfj url. Can be used to speed up insertion in case the urls are already known.
     */
    createArticlesById(pBasketIds: Array<string>, pLoadTransformations?: boolean, pBasketItems?: Array<basket.BasketItem>, pGfjUrls?: Map<string, string>): Promise<Array<MainArticleElement>>;
    setGfjLayerControl(pLayerControl: string): void;
    setGfjBasketIdsEnabled(pEnabled: boolean): void;
    setGfjMaxTextureSize(pMaxSize: number): void;
    synchronizeSession(pSaveArticleTransformations: boolean): Promise<void>;
    /**
     * Imports articles from the current eaiws session.
     */
    importFromSession(pOptions?: ArticleImportOptions): Promise<void>;
    /**
     * Imports articles from the given obx.
     * Note: The loaded articles will not be added to the ModelManager here. This should be done afterwards.
     */
    importObx(pObxData: ArrayBuffer | Blob, pOptions?: ArticleImportOptions): Promise<Array<MainArticleElement>>;
    exportObx(pArticles: Array<MainArticleElement>): Promise<string>;
    saveToObk(pSaveArticleTransformations: boolean): Promise<string>;
    /**
     * Returns a list of all main article which are currently part of the scene.
     */
    getAllMainArticles(): Array<MainArticleElement>;
    /**
     * Returns the article element with the given id.
     */
    getArticleById(pBasketId: string): ArticleElement | null;
    /**
     * Set active pricing procedure.
     */
    setPricingProcedure(pPricingProcedure: PricingProcedureInfo | null): void;
    /**
     * Resets cached ArticleData and ItemProperties for all existing articles.
     */
    resetCachedArticleInfos(pNotifyPropertyChange: boolean): void;
    /**
     * Creates a screenshot and returns it as data uri or blob.
     */
    createArticleScreenshot(pArticle: ArticleElement, pOptions: ArticleScreenshotOptions): Promise<string>;
    createArticleScreenshot(pArticle: ArticleElement, pOptions: ArticleScreenshotOptions, pAsBlob: false): Promise<string>;
    createArticleScreenshot(pArticle: ArticleElement, pOptions: ArticleScreenshotOptions, pAsBlob: true): Promise<Blob>;
    createArticleScreenshot(pArticle: ArticleElement, pOptions: ArticleScreenshotOptions, pAsBlob: boolean | undefined): Promise<string | Blob>;
    private mApp;
    private mSession;
    private mPricingProcedure;
    private mInstancedArticles;
    private mGfjLoader;
    private mGfjOptions;
    private mMaxGfjTextureSize;
    private ensureArticlesAreInstanced;
    private saveArticleTransformations;
    private deleteArticlesFromSession;
    private collectMainArticlesInGroup;
}
