import type { TransformNode } from "@babylonjs/core/Meshes/transformNode";
import type { Application } from "../core/";
import type { DocumentInputContext, DocumentOutputContext } from "../core/io/Types";
import { SceneElement } from "../core/mdl";
import type * as mdlioTypes from "../core/mdl/io/Types";
import * as basket from "../eaiws/basket";
import { CalculationSheetInfo } from "../eaiws/basket";
import { ListenableEvent } from "../utils";
import { ArticleElement } from "./ArticleElement";
import { SubArticleElement } from "./SubArticleElement";
import type { SubArticleHighlight } from "./SubArticleHighlight";
import type { ArticleConfigData } from "./ArticleTypes";
import type { ArticleManager } from "./ArticleManager";
export declare class MainArticleElement extends ArticleElement {
    static readonly CLASS_ID: string;
    get hasPlaceholderGeometry(): boolean;
    /**
     * If true the main article will not be visible. Only the sub-articles will be visible.
     */
    get hideMainArticle(): boolean;
    set hideMainArticle(pHide: boolean);
    /**
     * Array of sub-article ids. If not null only the sub-articles contained in this array are visible
     * otherwise all sub-articles will be visible.
     */
    get visibleSubArticles(): ReadonlyArray<string> | null;
    set visibleSubArticles(pSubArticles: ReadonlyArray<string> | null);
    get showDimensions(): boolean;
    set showDimensions(pShow: boolean);
    /**
     * Custom scale factor of the dimension.
     * @defaultValue 1
     */
    get dimensionsScaleFactor(): number;
    set dimensionsScaleFactor(pFactor: number);
    /**
     * Color of the dimensions as hex string.
     * @defaultValue "#595959"
     */
    get dimensionsColor(): string;
    set dimensionsColor(pColor: string);
    /**
     * Will be triggered inside updateGeometry() if the geometry was successfully updated.
     */
    readonly eventGeometryChanged: ListenableEvent<void, MainArticleElement>;
    constructor(pBasketId: string, pManager: ArticleManager, pNode: TransformNode);
    canSelectSubElement(): boolean;
    dispose(): void;
    isInstanced(): boolean;
    ensureIsInstanced(): Promise<void>;
    getMainArticle(): MainArticleElement;
    getParentArticle(): ArticleElement | null;
    isMainArticle(): this is MainArticleElement;
    isSubArticle(): this is SubArticleElement;
    hasSubArticles(pSelectableOnly: boolean): boolean;
    /**
     * Retrieves the basket items which are not already cached (MainArticle + SubArticles)
     * from the EAIWS and adds them to the cache.
     */
    fetchBasketItems(): Promise<void>;
    /**
     * Fills the internal cache with the given items.
     */
    fillBasketItemCache(pItems: Array<basket.BasketItem>): void;
    /**
     * Retrieves the item properties which are not already cached (MainArticle + SubArticles)
     * from the EAIWS and adds them to the cache.
     */
    fetchItemProperties(): Promise<void>;
    /**
     * Fills the internal cache with the given item properties.
     */
    fillItemPropertiesCache(pItemProps: Array<basket.ItemProperties>): void;
    /**
     * Retrieves the calculations which are not already cached (MainArticle + SubArticles)
     * from the EAIWS and adds them to the cache.
     */
    fetchCalculations(): Promise<void>;
    /**
     * Fills the internal cache with the given calculations.
     */
    fillCalculationCache(pCalculations: Map<string, CalculationSheetInfo>): void;
    /**
     * Fills the internal cache with the given data.
     */
    fillConfigDataCache(pData: Map<string, ArticleConfigData>): void;
    /**
     * Returns a calculation for the main article including all its sub articles.
     */
    getCompositeCalculation(): Promise<CalculationSheetInfo>;
    /**
     * Notify this main article that there was a change.
     * @param pArticle The changed article. Can be a sub article of the main article or the main article itself.
     * @param pWholeComposite Shoud be true if the given article and also other articles have changed.
     */
    notifyArticleChanged(pArticle: ArticleElement, pWholeComposite: boolean): void;
    updateGeometry(pValidateSelection?: boolean): Promise<void>;
    beginUpdateGeometry(): void;
    endUpdateGeometry(): Promise<void>;
    /**
     * If there is a pending geometry update this function will wait until it is finished,
     * otherwise it will resovle immediately.
     */
    waitUntilGeometryUpdated(): Promise<void>;
    clone(pNewParent?: TransformNode | null): SceneElement | null;
    serialize(pContext: DocumentOutputContext): mdlioTypes.SceneElement | null;
    /**
     * Returns the currently known ofml update state.
     */
    getKnownOfmlUpdateState(): basket.OFMLUpdateState;
    getOfmlUpdateState(): Promise<basket.OFMLUpdateState>;
    /**
     * Updates the ofml article and its sub articles to use the currently installed product data.
     * If an update is necessary can be checked with 'getOfmlUpdateState()'. Returns true if the
     * update was successful.
     * @pMigrate set to true if migration should be performed if necessary
     * @pUpdateGeometry set to true if geometry should also be updated
     * @pPriceDate price date to use for the update
     */
    updateOfmlArticle(pMigrate: boolean, pUpdateGeometry: boolean, pPriceDate?: basket.IsoDateString): Promise<boolean>;
    /**
     * Enables highlighting of selectabe sub articles for this main article. The highlighting will be disabled
     * automatically e.g. if the article changed or a tool was started.
     * Returns false if the default tool is currently not running (if a default tool exists)
     * or the highlighting is enabled already.
     * @param pCreateCallback Called after creation of the highlighting.
     * @param pDisposeCallback Called if the highighting was diposed/disabled.
     */
    highlightSubArticles(pCreateCallback: (pHighlights: ReadonlyArray<SubArticleHighlight>) => void, pDisposeCallback?: () => void): boolean;
    protected handleImmediateChanges(): void;
    protected copyDataFrom(pElement: SceneElement): void;
    protected static LoadArticleElementNode(pApp: Application, pContext: DocumentInputContext, pIOElement: mdlioTypes.SceneElement): {
        node: TransformNode | null;
        basketId: string | null;
        geometryIncomplete: boolean;
    } | null;
    protected supportScaleEaiwsTransform(): boolean;
    private mArticleDataCache;
    private mArticleDataOptionsCache;
    private mBasketItemCache;
    private mItemPropertiesCache;
    private mCalculationCache;
    private mCompositeCalculation;
    private mCatalogImageUrlCache;
    private mOapInteractorsCache;
    private mOapInteractorsCacheLock;
    private mLastKnownUpdateState;
    private mGfjUrl?;
    private mSubArticleCache;
    private mHasSubArticles;
    private mHasSelectableSubArticles;
    private mUpdateGeometryDepth;
    private mPendingGeometryUpdate;
    private mWaitGeometryUpdatePromise;
    private mWaitGeometryUpdateResolve;
    private mObxUrl;
    private mEnsureObxPromise;
    private mEnsureIsInstancedPromise;
    private mHideMainArticle;
    private mVisibleSubArticles;
    private mSubArticleHighlights;
    private mEaiwsTransformUpToDate;
    private mHasPlaceholderGeometry;
    private mBoxDimensions;
    private mDimensionsScaleFactor;
    private mDimensionsColor;
    private recoverInstance;
    private createOrUpdateSubArticleElements;
    private createSubArticleElements;
    private ensureObx;
    private getEaiwsBasketItems;
    private getEaiwsItemProps;
    private getEaiwsItemCalculation;
    private getEaiwsCompositeCalculation;
    private updateMainArticleVisibility;
    private updateSubArticleVisibility;
    private updateBoxDimensions;
}
