import { Texture as BJSTexture } from "@babylonjs/core/Materials/Textures/texture";
import { Matrix, Vector4 } from "@babylonjs/core/Maths/math.vector";
import { TransformNode } from "@babylonjs/core/Meshes/transformNode";
import { CreateBox } from "@babylonjs/core/Meshes/Builders/boxBuilder";
import { NodeUserData } from "../core/mdl/NodeUserData";
import { PbrMaterial } from "../core/mdl/PbrMaterial";
import { getDataPath, Log } from "../utils";
import { PI_DIV2 } from "../utils/math";
import { joinPath } from "../utils/string";
import { ajax, Lock } from "../utils/async";
import { loadImage } from "../utils/image";
import { ARTICLEELEMENT_NODEDATA_BASKETID } from "./Constants";
/**
 * Creates an empty TransformNode for the given article.
 */
export class DummyPlaceholderCreator {
    constructor(pArticleManager) {
        this.mApp = pArticleManager.app;
    }
    create(pArticle, pSubArticleIds) {
        const tNode = new TransformNode("article_dummyPlaceholder", this.mApp.scene);
        NodeUserData.SetData(tNode, ARTICLEELEMENT_NODEDATA_BASKETID, pArticle.basketId);
        tNode.freezeWorldMatrix();
        //create sub articles
        for (const tSubId of pSubArticleIds) {
            const tSubNode = new TransformNode("subarticle_dummyPlaceholder", this.mApp.scene);
            NodeUserData.SetData(tSubNode, ARTICLEELEMENT_NODEDATA_BASKETID, tSubId);
            tSubNode.parent = tNode;
            tSubNode.freezeWorldMatrix();
        }
        return (Promise.resolve(tNode));
    }
}
export class DefaultPlaceholderCreator {
    constructor(pArticleManager) {
        this.mApp = pArticleManager.app;
        this.mLoadImageLock = new Lock();
        this.mFrameImg = null;
        this.mDefaultCatalogImg = null;
    }
    async create(pArticle, pSubArticleIds) {
        const tItemProps = await pArticle.getItemProperties();
        const tCatImage = await pArticle.getCatalogImageUrl();
        const tTexture = await this.createTexture(tItemProps, tCatImage);
        const tMesh = this.createMesh(tItemProps, tTexture);
        //create sub articles
        for (const tSubId of pSubArticleIds) {
            const tSubNode = new TransformNode("subarticle_placeholder", this.mApp.scene);
            NodeUserData.SetData(tSubNode, ARTICLEELEMENT_NODEDATA_BASKETID, tSubId);
            tSubNode.parent = tMesh;
            tSubNode.freezeWorldMatrix();
        }
        return (tMesh);
    }
    async createTexture(pItemProps, pCatalogImageUrl) {
        const tFrameImg = await this.loadFrameImg();
        let tCatalogImg = null;
        if (pCatalogImageUrl != null)
            tCatalogImg = await this.loadImgFromUrl(pCatalogImageUrl);
        //fallback to default image
        if (tCatalogImg == null)
            tCatalogImg = await this.loadDefaultCatalogImg();
        const tCanvas = document.createElement("canvas");
        tCanvas.width = 950;
        tCanvas.height = 346;
        const tContext = tCanvas.getContext("2d");
        if (tContext == null)
            throw new Error(" Failed to get 2d canvas context");
        //prepare text
        tContext.font = "400 36px " + this.mApp.theme.fontFamily;
        const tShorTextLines = this.getTextLines(tContext, pItemProps.article.shortText, 545);
        tContext.font = "400 24px " + this.mApp.theme.fontFamily;
        const tManSeriesLines = this.getTextLines(tContext, pItemProps.article.manufacturerName + " | " + pItemProps.article.seriesName, 545);
        this.limitLines(tShorTextLines, tManSeriesLines.length > 1 ? 2 : 3);
        this.limitLines(tManSeriesLines, 2);
        //frame
        tContext.drawImage(tFrameImg, 0, 0);
        //catalog image
        if (tCatalogImg != null) {
            if (tCatalogImg.naturalWidth > tCatalogImg.naturalHeight) {
                const tRatio = tCatalogImg.naturalHeight / tCatalogImg.naturalWidth;
                const tHeight = Math.round(250 * tRatio);
                const tY = Math.round(48 + 125 - tHeight * 0.5);
                tContext.drawImage(tCatalogImg, 48, tY, 250, tHeight);
            }
            else {
                const tRatio = tCatalogImg.naturalWidth / tCatalogImg.naturalHeight;
                const tWidth = Math.round(250 * tRatio);
                const tX = Math.round(48 + 125 - tWidth * 0.5);
                tContext.drawImage(tCatalogImg, tX, 48, tWidth, 250);
            }
        }
        let tLineHeight = 36;
        let tLineSpacing = tLineHeight * 0.15;
        let tPosY = 94;
        //article number
        tContext.font = "700 36px " + this.mApp.theme.fontFamily;
        tContext.fillStyle = "#4f4f4f";
        tContext.fillText(pItemProps.article.baseArticleNumber, 335, tPosY);
        //short text
        tPosY += tLineHeight + 15;
        tContext.font = "400 36px " + this.mApp.theme.fontFamily;
        for (let i = 0; i < tShorTextLines.length; ++i) {
            if (i > 0)
                tPosY += tLineHeight + tLineSpacing;
            tContext.fillText(tShorTextLines[i], 335, tPosY);
        }
        //manufacturer and series
        tPosY += tLineHeight + 15;
        tLineHeight = 24;
        tLineSpacing = tLineHeight * 0.15;
        tContext.fillStyle = "#b2b2b2";
        tContext.font = "400 24px " + this.mApp.theme.fontFamily;
        for (let i = 0; i < tManSeriesLines.length; ++i) {
            if (i > 0)
                tPosY += tLineHeight + tLineSpacing;
            tContext.fillText(tManSeriesLines[i], 335, tPosY);
        }
        return (tCanvas.toDataURL("image/jpg"));
    }
    createMesh(pItemProps, pTextureData) {
        const tName = "article_placeholder_" + pItemProps.article.baseArticleNumber;
        const tMaterialId = "article_placeholder_" + pItemProps.itemId;
        //dispose existing material and texture
        const tMat = this.mApp.scene.getMaterialById(tMaterialId);
        if (tMat != null) {
            tMat.dispose(false, true);
        }
        //mesh
        const tFaceUVs = new Array();
        tFaceUVs.push(new Vector4(1, 1, 0, 0));
        tFaceUVs.push(new Vector4(0, 0, 1, 1));
        tFaceUVs.push(new Vector4(0, 0, 0, 0));
        tFaceUVs.push(new Vector4(0, 0, 0, 0));
        tFaceUVs.push(new Vector4(0, 0, 0, 0));
        tFaceUVs.push(new Vector4(0, 0, 0, 0));
        const tMesh = CreateBox(tName, {
            width: 1.0,
            height: 0.364,
            depth: 0.01,
            faceUV: tFaceUVs
        }, this.mApp.scene);
        tMesh.bakeTransformIntoVertices(Matrix.RotationX(PI_DIV2));
        tMesh.bakeTransformIntoVertices(Matrix.Translation(0.5, 0.005, -0.182));
        tMesh.freezeWorldMatrix();
        //material
        const tMaterial = new PbrMaterial(tName, this.mApp.model.materials);
        tMaterial.id = tMaterialId;
        tMaterial.backFaceCulling = false;
        tMaterial.metallic = 0;
        tMaterial.roughness = 1.0;
        tMaterial.albedoTexture = BJSTexture.LoadFromDataString(pItemProps.itemId, pTextureData, this.mApp.scene);
        tMaterial.albedoTexture.anisotropicFilteringLevel = this.mApp.engine.getCaps().maxAnisotropy;
        tMaterial.freeze();
        tMesh.material = tMaterial;
        //set basket id
        NodeUserData.SetData(tMesh, ARTICLEELEMENT_NODEDATA_BASKETID, pItemProps.itemId);
        tMesh.freezeWorldMatrix();
        return (tMesh);
    }
    async loadFrameImg() {
        const tLock = await this.mLoadImageLock.acquireLock();
        try {
            if (this.mFrameImg == null) {
                this.mFrameImg = await loadImage(joinPath(getDataPath(), "cf/images/placeholder_frame.png"));
            }
            return (this.mFrameImg);
        }
        finally {
            tLock.unlock();
        }
    }
    async loadDefaultCatalogImg() {
        const tLock = await this.mLoadImageLock.acquireLock();
        try {
            if (this.mDefaultCatalogImg == null) {
                this.mDefaultCatalogImg = await loadImage(joinPath(getDataPath(), "cf/images/placeholder_catalog_default.png"));
            }
            return (this.mDefaultCatalogImg);
        }
        finally {
            tLock.unlock();
        }
    }
    async loadImgFromUrl(pUrl) {
        const tLock = await this.mLoadImageLock.acquireLock();
        try {
            //download image in advance to avoid "unsecure operation" exception in mixed context scenarios
            const tArrayBuffer = await ajax("GET", pUrl, undefined, { responseType: "arraybuffer" });
            const tImage = await loadImage(new Blob([tArrayBuffer]));
            return (tImage);
        }
        catch (pError) {
            Log.error("loadImgFromUrl failed", pError);
            return (null);
        }
        finally {
            tLock.unlock();
        }
    }
    getTextLines(pContext, pText, pMaxWidth) {
        if (pText == null)
            return ([""]);
        const words = pText.split(" ");
        const lines = [];
        let currentLine = words[0];
        for (let i = 1; i < words.length; i++) {
            const word = words[i];
            const width = pContext.measureText(currentLine + " " + word).width;
            if (width < pMaxWidth) {
                currentLine += " " + word;
            }
            else {
                lines.push(currentLine);
                currentLine = word;
            }
        }
        lines.push(currentLine);
        return lines;
    }
    limitLines(pLines, pMaxLines) {
        if (pLines.length > pMaxLines) {
            pLines.splice(pMaxLines);
            //replace last char of last line with ellipsis
            let tLastLine = pLines[pMaxLines - 1];
            tLastLine = tLastLine.substr(0, tLastLine.length - 1);
            tLastLine += "\u2026";
            pLines[pMaxLines - 1] = tLastLine;
        }
    }
}
