import { Vector3 } from "@babylonjs/core/Maths/math.vector";
import { FileNotFoundError } from "../../core/io/Types";
import { ZipDocumentContainer } from "../../core/io/ZipDocumentContainer";
import { BasicElement } from "../../core/mdl/BasicElement";
import { CameraMode } from "../../core/view/CameraControl";
import * as basket from "../../eaiws/basket";
import { Log, Progress } from "../../utils";
import { addElements } from "../../utils/array";
import { ERR_INVALID_ARGUMENT, ERR_INVALID_STATE, newError } from "../../utils/error";
import { isNotNullOrEmpty, isNullOrEmpty } from "../../utils/string";
import { ArticleImport } from "../ArticleImport";
import { MainArticleElement } from "../MainArticleElement";
import { UserArticleElement } from "../UserArticleElement";
import { GfjLoader } from "./GfjLoader";
import { ObxReader } from "./ObxReader";
export async function importPec(pArticleManager, pPecData, pOptions = {}) {
    if (pPecData == null)
        throw new Error(ERR_INVALID_ARGUMENT);
    const tContainer = await ZipDocumentContainer.LoadContainer(pPecData);
    return (importPecFromContainer(pArticleManager, tContainer, pOptions));
}
export function importPecFromContainer(pArticleManager, pContainer, pOptions = {}) {
    const tImport = new PecImport(pArticleManager);
    return (tImport.importFromContainer(pContainer, pOptions));
}
export function importPecFromCatalog(pArticleManager, pCatalogContainerItem, pOptions = {}) {
    const tImport = new PecImport(pArticleManager);
    return (tImport.importFromCatalog(pCatalogContainerItem, pOptions));
}
class PecImport {
    constructor(pArticleManager) {
        this.mArticleManager = pArticleManager;
    }
    async importFromContainer(pContainer, pOptions = {}) {
        if (pContainer == null)
            throw new Error(ERR_INVALID_ARGUMENT);
        const tContext = {};
        tContext.itemProps = null;
        tContext.loadedElements = [];
        tContext.progress = new Progress(pOptions.onProgress, true);
        tContext.gfjLoader = new GfjLoader(this.mArticleManager.app.model);
        tContext.gfjLoader.onResolveUrl = (pUrl) => this.resolveGfjUrl(pUrl, tContext);
        try {
            this.mArticleManager.app.viewer.beginDisableRendering();
            tContext.progress.startSection(5);
            tContext.container = pContainer;
            await this.readPecXml(tContext, pOptions.view);
            tContext.progress.startSection(5);
            await this.importObx(tContext, pOptions.dataLanguage);
            tContext.progress.startSection(60);
            await this.loadArticles(tContext, pOptions.shouldUpdateCallback);
            tContext.progress.startSection(30);
            await this.loadScene(tContext);
            tContext.progress.setProgress(100);
        }
        finally {
            this.mArticleManager.app.viewer.requestRenderFrame();
            this.mArticleManager.app.viewer.endDisableRendering();
        }
        return (tContext.loadedElements);
    }
    async importFromCatalog(pContainerItem, pOptions = {}) {
        if (pContainerItem == null)
            throw new Error(ERR_INVALID_ARGUMENT);
        const tContext = {};
        tContext.itemProps = null;
        tContext.loadedElements = [];
        tContext.progress = new Progress(pOptions.onProgress, true);
        tContext.gfjLoader = new GfjLoader(this.mArticleManager.app.model);
        try {
            this.mArticleManager.app.viewer.beginDisableRendering();
            tContext.progress.startSection(10);
            await this.pasteContainerItem(tContext, pContainerItem);
            tContext.progress.startSection(60);
            await this.loadArticles(tContext, pOptions.shouldUpdateCallback);
            tContext.progress.startSection(30);
            await this.loadScene(tContext);
            tContext.progress.setProgress(100);
        }
        finally {
            this.mArticleManager.app.viewer.requestRenderFrame();
            this.mArticleManager.app.viewer.endDisableRendering();
        }
        return (tContext.loadedElements);
    }
    async readPecXml(pContext, pView) {
        if (pContext.container == null)
            throw newError(ERR_INVALID_STATE, "container is null");
        const tXmlString = await pContext.container.getFile("pec.xml", false);
        const tParser = new DOMParser();
        const tDoc = tParser.parseFromString(tXmlString, "application/xml");
        if (tDoc == null || tDoc.documentElement == null)
            throw new Error("Failed to read pec.xml.");
        const tRoot = tDoc.documentElement;
        if (tRoot.tagName !== "pec")
            throw new Error("Invalid root element in pec.xml.");
        //get resource path mapping
        pContext.uriResourcePathMap = new Map();
        pContext.gfjResourcePathMap = new Map();
        const tResourcesList = tRoot.querySelectorAll("resources > resource");
        if (tResourcesList != null) {
            for (let i = 0; i < tResourcesList.length; ++i) {
                const tResourceElement = tResourcesList.item(i);
                const tPathElement = tResourceElement.querySelector("path");
                if (tPathElement != null) {
                    const tPath = tPathElement.textContent;
                    if (!isNullOrEmpty(tPath)) {
                        const tIsGfj = tPath.endsWith(".gfj");
                        //uri resource mapping
                        const tUri = tResourceElement.getAttribute("uri");
                        if (!isNullOrEmpty(tUri)) {
                            pContext.uriResourcePathMap.set(tUri, tPath);
                            if (tIsGfj)
                                pContext.gfjResourcePathMap.set(tUri, tPath);
                        }
                        //checksum resource mapping
                        const tGeomCkSum = tResourceElement.getAttribute("geometryChecksum");
                        if (!isNullOrEmpty(tGeomCkSum) && tIsGfj)
                            pContext.gfjResourcePathMap.set(tGeomCkSum, tPath);
                    }
                }
            }
        }
        //get scene gfj path
        const tSceneElement = tRoot.querySelector("scene");
        if (tSceneElement != null) {
            const tUri = tSceneElement.getAttribute("uri");
            if (isNotNullOrEmpty(tUri))
                pContext.sceneGfjUrl = pContext.gfjResourcePathMap.get(tUri);
        }
        //load camera
        if (pView != null)
            this.loadCamera(pView, tRoot);
    }
    loadCamera(pView, pPecXmlRoot) {
        const tCamElement = pPecXmlRoot.querySelector("camera");
        if (tCamElement == null)
            return;
        //mode
        const tType = tCamElement.getAttribute("type");
        if (tType === "orthographic") {
            pView.cameraControl.cameraMode = CameraMode.Orthographic;
            //ortho Y
            const tOrthoElement = tCamElement.querySelector("orthoY");
            if (tOrthoElement != null) {
                pView.cameraControl.setCameraOrtho(parseFloat(tOrthoElement.getAttribute("value")));
            }
        }
        else {
            pView.cameraControl.cameraMode = CameraMode.Perspective;
        }
        //position
        const tPosElement = tCamElement.querySelector("position");
        if (tPosElement != null) {
            const tPosition = Vector3.Zero();
            tPosition.x = parseFloat(tPosElement.getAttribute("x"));
            tPosition.y = parseFloat(tPosElement.getAttribute("y"));
            tPosition.z = -parseFloat(tPosElement.getAttribute("z"));
            pView.cameraControl.setPosition(tPosition);
        }
        //direction
        const tDirElement = tCamElement.querySelector("direction");
        if (tDirElement != null) {
            const tDir = Vector3.Zero();
            tDir.x = parseFloat(tDirElement.getAttribute("x"));
            tDir.y = parseFloat(tDirElement.getAttribute("y"));
            tDir.z = -parseFloat(tDirElement.getAttribute("z"));
            pView.cameraControl.setDirection(tDir);
        }
    }
    async importObx(pContext, pDataLanguage) {
        try {
            if (pContext.container == null)
                throw newError(ERR_INVALID_STATE, "container is null");
            //read obx from container
            const tObxData = await pContext.container.getFile("document.obx", true);
            const tObxBlob = new Blob([tObxData]);
            const tSession = this.mArticleManager.session;
            if (tSession.isValid) {
                //upload obx to eaiws
                const tUploadUrl = await tSession.uploadFile("CutBuffer", tObxBlob);
                if (isNullOrEmpty(tUploadUrl))
                    throw new Error("Failed to upload obx.");
                //paste obx
                const tPastedIds = await tSession.basket.paste(this.mArticleManager.fatherItemId, null, tUploadUrl);
                if (tPastedIds == null || tPastedIds.length === 0)
                    return;
                //delete uploaded obx
                await tSession.deleteFile(tUploadUrl);
                //get all items
                const tOptions = new basket.GetAllItemsOptions();
                tOptions.subItems = true;
                tOptions.wholeComposite = true;
                tOptions.geometryIds = true;
                tOptions.setArticleIds = true;
                const tPropsOptions = this.mArticleManager._GetDefaultItemPropsOptions();
                tPropsOptions.subItems = true;
                tPropsOptions.wholeComposite = true;
                pContext.items = await tSession.basket.getAllItems(tPastedIds, tOptions);
                pContext.itemProps = await tSession.basket.getItemProperties(tPastedIds, tPropsOptions);
            }
            else {
                //fallback to client side obx parsing
                const tObxReader = new ObxReader();
                if (pDataLanguage != null) {
                    tObxReader.setLanguages(pDataLanguage);
                }
                await tObxReader._loadObx(tObxBlob);
                pContext.items = tObxReader.getAllItems();
                pContext.itemProps = tObxReader.getAllItemProperties();
            }
        }
        catch (pError) {
            //document.obx is optional
            if (!(pError instanceof FileNotFoundError))
                throw (pError);
        }
    }
    async pasteContainerItem(pContext, pContainerItem) {
        var _a;
        if (pContainerItem.type !== "Container")
            throw new Error("Invalid catalog item type.");
        //get container uri
        let tUri = null;
        if (pContainerItem.resources != null) {
            for (const tRes of pContainerItem.resources) {
                if (tRes.type === "XCF:PEC") {
                    tUri = (_a = tRes.url) !== null && _a !== void 0 ? _a : tRes.value;
                    break;
                }
            }
        }
        if (isNullOrEmpty(tUri))
            throw new Error("Catalog item does not have XCF:PEC resource.");
        const tSession = this.mArticleManager.session;
        const tPasteOptions = new basket.PasteContainerOptions();
        tPasteOptions.setArticleIds = true;
        tPasteOptions.geometryIds = true;
        const tPasteResult = await tSession.basket.pasteContainer(null, null, tUri, tPasteOptions);
        if (tPasteResult == null)
            return;
        //collect item ids and geometry-uris
        const tPastedIds = new Array();
        const tGeomUris = new Array();
        if (tPasteResult.items != null && tPasteResult.itemGeometries != null) {
            for (let i = 0; i < tPasteResult.items.length; ++i) {
                tPastedIds.push(tPasteResult.items[i].itemId);
                const tGeomUri = tPasteResult.itemGeometries[i];
                if (isNullOrEmpty(tGeomUri)) {
                    //add dummy url to prevent invalid url exception of resolveURIs
                    tGeomUris.push("imp:00000000000000000000000000000000");
                }
                else
                    tGeomUris.push(tGeomUri);
            }
        }
        //resolve uris (including sceneGeometry)
        pContext.basketIdGfjMap = new Map();
        if (!isNullOrEmpty(tPasteResult.sceneGeometry))
            tGeomUris.push(tPasteResult.sceneGeometry);
        const tResolvedUrls = await tSession.session.resolveURIs(tGeomUris);
        for (let i = 0; i < tPastedIds.length; ++i) {
            pContext.basketIdGfjMap.set(tPastedIds[i], tResolvedUrls[i]);
        }
        if (!isNullOrEmpty(tPasteResult.sceneGeometry))
            pContext.sceneGfjUrl = tResolvedUrls[tResolvedUrls.length - 1];
        //get all item properties
        const tPropsOptions = this.mArticleManager._GetDefaultItemPropsOptions();
        tPropsOptions.subItems = true;
        tPropsOptions.wholeComposite = true;
        pContext.items = tPasteResult.items;
        pContext.itemProps = await tSession.basket.getItemProperties(tPastedIds, tPropsOptions);
    }
    async loadArticles(pContext, pShouldUpdateCallback) {
        if (pContext.itemProps == null || pContext.itemProps.length === 0 ||
            pContext.items == null || pContext.items.length === 0)
            return;
        //check if pec contains gfj data (only for direct import)
        if (pContext.gfjResourcePathMap != null && pContext.gfjResourcePathMap.size === 0) {
            //Fallback: load gfj from eaiws
            const tArticles = await ArticleImport.Import(this.mArticleManager, pContext.items, pContext.itemProps, false, { shouldUpdateCallback: pShouldUpdateCallback });
            if (tArticles != null) {
                addElements(pContext.loadedElements, tArticles);
            }
            return;
        }
        //collect main articles
        const tMainArticles = new Array();
        for (const tProps of pContext.itemProps) {
            if (tProps.article != null && tProps.composableGeometry != null && tProps.composableGeometry.enabled) {
                tMainArticles.push(tProps);
            }
        }
        if (tMainArticles.length === 0)
            return;
        //build map from basket id to properties
        pContext.itemPropsMap = new Map();
        for (const tProps of pContext.itemProps) {
            pContext.itemPropsMap.set(tProps.itemId, tProps);
        }
        //build map from basket id to item
        pContext.itemsMap = new Map();
        for (const tItem of pContext.items) {
            pContext.itemsMap.set(tItem.itemId, tItem);
        }
        //load articles
        const tStep = 100 / tMainArticles.length;
        for (const tMainArticle of tMainArticles) {
            try {
                const tArticle = await this.loadArticle(tMainArticle, pContext);
                if (tArticle != null)
                    pContext.loadedElements.push(tArticle);
            }
            catch (pError) {
                Log.error("Failed to load article " + tMainArticle.itemId, pError);
            }
            pContext.progress.setProgress(pContext.progress.sectionProgress + tStep);
        }
    }
    async loadArticle(pMainArticle, pContext) {
        const tSession = this.mArticleManager.session;
        const tGfjPath = this.getArticleGfjPath(pMainArticle, pContext);
        if (isNullOrEmpty(tGfjPath))
            throw new Error("gfj path not found");
        const tArticleItems = this.mArticleManager._getCompositeItems(pMainArticle.itemId, pContext.itemsMap);
        const tNodes = await pContext.gfjLoader.load(tGfjPath, pMainArticle.itemId, tArticleItems, pContext.container);
        let tNode = null;
        if (tNodes != null && tNodes.length > 0) {
            tNode = tNodes[0];
        }
        let tArticle = null;
        if (pMainArticle.itemType === "UserArticle") {
            tArticle = new UserArticleElement(pMainArticle.itemId, this.mArticleManager, tNode);
        }
        else {
            tArticle = new MainArticleElement(pMainArticle.itemId, this.mArticleManager, tNode);
        }
        //fill props cache
        //get item properties for loaded basket ids
        const tItemProperties = new Array();
        for (const tItem of tArticleItems) {
            const tProps = pContext.itemPropsMap.get(tItem.itemId);
            if (tProps != null)
                tItemProperties.push(tProps);
        }
        tArticle.fillItemPropertiesCache(tItemProperties);
        //reset imported geometry for user articles
        if (tSession.isValid && tArticle instanceof UserArticleElement)
            await tArticle._resetImpGeometry();
        if (tArticle.node == null)
            await tArticle._initPlaceholderGeometry(tArticleItems);
        //load transformation
        tArticle._loadEaiwsTransform(pMainArticle.composableGeometry);
        return (tArticle);
    }
    getArticleGfjPath(pMainArticle, pContext) {
        if (pContext.basketIdGfjMap != null) {
            return (pContext.basketIdGfjMap.get(pMainArticle.itemId));
        }
        if (pContext.gfjResourcePathMap != null) {
            if (pMainArticle.itemType === "UserArticle") {
                //use geometry for path mapping
                const tGeometry = pMainArticle.composableGeometry.geometry;
                if (tGeometry == null)
                    throw new Error("missing geometry for user article");
                return (pContext.gfjResourcePathMap.get(tGeometry));
            }
            else {
                //use geometryId for path mapping
                const tGeomId = this.normalizeBase64(pMainArticle.article.geometryId);
                if (isNullOrEmpty(tGeomId))
                    throw new Error("missing geometryId");
                return (pContext.gfjResourcePathMap.get(tGeomId));
            }
        }
        return (null);
    }
    async loadScene(pContext) {
        if (isNullOrEmpty(pContext.sceneGfjUrl))
            return;
        try {
            const tResults = await pContext.gfjLoader.load(pContext.sceneGfjUrl, null, null, pContext.container);
            if (tResults != null && tResults.length > 0) {
                for (const tResult of tResults) {
                    if (tResult != null) {
                        const tElement = new BasicElement(this.mArticleManager.app.model, tResult);
                        pContext.loadedElements.push(tElement);
                    }
                }
            }
        }
        catch (pError) {
            //scene.gfj is optional
            if (!(pError instanceof FileNotFoundError))
                throw (pError);
        }
    }
    resolveGfjUrl(pUrl, pContext) {
        if (pUrl == null)
            return ("");
        if (pContext.uriResourcePathMap != null && pUrl.startsWith("imp:")) {
            const tUrl = pContext.uriResourcePathMap.get(pUrl);
            return (tUrl != null ? tUrl : "");
        }
        return (pUrl);
    }
    normalizeBase64(pBase64) {
        if (pBase64 == null)
            return (null);
        let tBase64 = pBase64;
        tBase64 = tBase64.replace(/_/g, "/");
        tBase64 = tBase64.replace(/-/g, "+");
        return (tBase64);
    }
}
