import { Scene } from "@babylonjs/core/scene";
import { Engine } from "@babylonjs/core/Engines/engine";
import { ListenableEvent } from "../utils";
import { AppDocument } from "./AppDocument";
import type { AppModule } from "./AppModule";
import { AppSettings } from "./AppSettings";
import { CommandManager } from "./cmd/CommandManager";
import { I18nManager } from "./I18nManager";
import { ModelManager } from "./mdl/ModelManager";
import type { SceneElement } from "./mdl/SceneElement";
import type { Property } from "./prop/Property";
import { RenderingManager } from "./rendering/RenderingManager";
import { ToolManager } from "./tool/ToolManager";
import { Viewer } from "./view/Viewer";
export interface AppInitOptions {
    /**
     * Enables hardware antialiasing if supported from the hardware.
     * Note: Hardware antialiasing does not work in combination with post processing. If you want to use it
     * you have to disable SSAO and FXAA.
     * @defaultValue false
     */
    hardwareAntialiasing: boolean;
    /**
     * If enabled, allows direct capturing of the current frame from the rendering canvas.
     * Has to be enabled if `Viewer.createScreenshot()` is used.
     * @defaultValue true
     */
    preserveDrawingBuffer: boolean;
    /**
     * Enables stencil buffer access.
     * @defaultValue true
     */
    stencilBuffer: boolean;
    /**
     * If enabled, the viewer will be adjust the rendering resolution to match the device pixel ratio.
     * @defaultValue true
     */
    adaptToDeviceRatio: boolean;
    /**
     * Maximum device pixel ration used for rendering.
     * @defaultValue 2.0
     */
    maximumDeviceRatio: number;
    /**
     * If enabled, the viewer will be automatically resized. Viewer.resize() has to be called manually if disabled.
     * @defaultValue true
     */
    autoResizeViewer: boolean;
    /**
     * If enabled, the BabylonJs audio engine will be initialized.
     * @defaultValue false
     */
    audioEngine: boolean;
    /**
     * Can be used to disable WebGL 2 and force the use of WebGL 1.
     * Because the support for WebGL 2 is still buggy on some devices/browsers (e.g. IOS 15.0 or some low-end Android devices), it is recommended
     * to enable WebGL 2 only on well known devices/browsers or to offer an option to the user where he can disable it in case of problems.
     * @defaultValue true
     */
    disableWebGL2Support: boolean;
}
export interface AppTheme {
    mainColorBright: string;
    mainColorDark: string;
    fontFamily: string;
}
export interface AppCallbacks {
    /**
     * If called the host application should show a property editor.
     * If only one property is shown, the editor should be closed automatically after the user changed the value
     * of the property.
     * @param pElement The element which will be edited.
     * @param pGetPropertiesCallback Callback to retrieve the properties which should be shown. Has to be called everytime a property value was changed.
     * @param pOptions Customization options.
     */
    showPropertyEditor?: (pElement: SceneElement, pGetPropertiesCallback: () => Promise<Array<Property>>, pOptions: AppCallbacks.ShowPropertyEditorOptions) => void;
    /**
     * If called the host application should show a action selection dialog.
     * After the user selected an action the dialog has to be closed.
     * @param pElement The element which will be edited.
     * @param pActions The actions which should be shown.
     * @param pOptions Customization options.
     */
    showActionChoice?: (pActions: Array<AppCallbacks.ChoiceAction>, pOptions: AppCallbacks.ShowActionChoiceOptions) => void;
    /**
     * If called the host application should show the provided media.
     * @param pMedia The media to show.
     * @param pOptions Customization options.
     */
    showMedia?: (pMedia: AppCallbacks.Media, pOptions: AppCallbacks.ShowMediaOptions) => void;
    /**
     * If called the host application should show the provided message to the user.
     * @param pMessage The media to show.
     * @param pOptions Customization options.
     */
    showMessage?: (pMessage: AppCallbacks.Message, pOptions: AppCallbacks.ShowMessageOptions) => void;
    /**
     * Notifies the host application that an asynchronous operation was started (e.g. to show a loading animation).
     * Note: Nested calls to begin/endAsyncOperation should be supported.
     */
    beginAsyncOperation?: () => void;
    /**
     * Notifies the host application that an asynchronous operation has ended.
     */
    endAsyncOperation?: () => void;
}
export declare namespace AppCallbacks {
    interface BaseCallbackOptions {
        /** Optional origin reference for UI placement. */
        originReference?: ClientRect;
    }
    interface ShowPropertyEditorOptions extends BaseCallbackOptions {
        /** Optional title for the property editor. */
        title?: string;
    }
    interface ShowActionChoiceOptions extends BaseCallbackOptions {
        /** Optional title for the action choice dialog. */
        title?: string;
        /** Display mode for the actions. */
        displayMode: "list" | "tile";
        /** Tile size for the tiles if displayMode = "tile" */
        tileSize?: "small" | "medium" | "large";
        /** If defined the UI should show a back button and execute this function if it is clicked. */
        goBack?: () => void;
    }
    interface ShowMediaOptions extends BaseCallbackOptions {
    }
    interface ShowMessageOptions extends BaseCallbackOptions {
    }
    interface ChoiceAction {
        /** Display text for the action. */
        text?: string;
        /** Image of this action. */
        image?: string;
        /** Has to be executed if user selects this action. */
        execute: () => Promise<void>;
    }
    interface Media {
        /** Id of the media to show. */
        id: string;
        /**
         * Type of the media.
         *  - YouTube: In this case the `id` is the ID of an youtube video.
         */
        type: "YouTube";
    }
    interface Message {
        /** Text to show for this message. */
        text: string;
    }
}
export declare class Application {
    get version(): string;
    get buildInfo(): string;
    get engine(): Engine;
    get scene(): Scene;
    get i18n(): I18nManager;
    get settings(): AppSettings;
    get document(): AppDocument;
    get viewer(): Viewer;
    get model(): ModelManager;
    get commands(): CommandManager;
    get tools(): ToolManager;
    get rendering(): RenderingManager;
    get theme(): AppTheme;
    /**
     * Callbacks which should be implemented by the host application.
     */
    get appCallbacks(): AppCallbacks;
    /**
     * The list of registered modules.
     */
    get modules(): ReadonlyArray<AppModule>;
    /**
     * The options used to initialize the application.
     * If the aplication was not initialized yet these options represent the default options.
     */
    get initOptions(): Readonly<AppInitOptions>;
    /**
     * Name of the host application. (e.g. used as meta data for saved documents).
     */
    applicationName: string;
    /**
     * Version of the host application. (e.g. used as meta data for saved documents).
     */
    applicationVersion: string;
    /**
     * Will be triggered periodically from the applications main loop. Can be used for custom animations.
     */
    readonly eventAnimate: ListenableEvent<void, Application>;
    constructor();
    /**
     * Has to be called once to intialize the application.
     * The 3d rendering canvas will be created as a children of pViewerContainer.
     * Returns false if initialization failed.
     */
    initialize(pViewerContainer: HTMLDivElement, pOptions?: Partial<AppInitOptions>): boolean;
    /** Dispose application and free used resources. */
    dispose(): void;
    /** Registers a new application module. */
    registerModule(pModule: AppModule): void;
    /** Returns the module with the given id or null if the module was not found. */
    getModule(pModuleId: string): AppModule | null;
    private mEngine;
    private mScene;
    private mI18n;
    private mSettings;
    private mDocument;
    private mViewer;
    private mModelManager;
    private mCommandManager;
    private mToolManager;
    private mRenderingManager;
    private mTheme;
    private mAppCallbacks;
    private mInitOptions;
    private mModules;
    private mModulesArray;
    private processMainLoop;
}
