import type { GroupedMap, GroupedMapData } from "../../utils";
import { CustomError } from "../../utils/error";
import type { Model } from "../mdl/io/Types";
import type { Viewer } from "../view/io/Types";
export declare const DOCUMENT_MAJOR_VERSION = 2;
export declare const DOCUMENT_MINOR_VERSION = 1;
export type Quaternion = [number, number, number, number];
export type Vector2 = [number, number];
export type Vector3 = [number, number, number];
export type Color3 = [number, number, number];
export interface DataMap {
    [dataKey: string]: unknown;
}
export interface Document {
    header: DocumentHeader;
    appData?: GroupedMapData;
    moduleData?: GroupedMapData;
    viewer?: Viewer;
    model: Model;
}
export interface DocumentHeader {
    majorVer: number;
    minorVer: number;
    appName?: string;
    appVer?: string;
}
export interface DocumentIOContext {
    readonly customState: GroupedMap;
    /**
     * Can be used to register a task which needs to be finished before the the whole
     * load/save process is finished. Tasks will be processed sequentially.
     */
    registerPendingTask(pTask: () => Promise<void>): void;
}
export interface DocumentInputContext extends DocumentIOContext {
    readonly documentHeader: DocumentHeader;
    getGeometryIdByIndex(pIndex: number): string | null;
    getMaterialIdByIndex(pIndex: number): string | null;
    getTextureIdByIndex(pIndex: number): string | null;
    getLayerNameByIndex(pIndex: number): string | null;
    /**
     * Retrieves an embedded binary or text resource from the document container. If the file doesn't exist the promise
     * will be rejected with an FileNotFoundError.
     * @param pFilename Name of the file. May include a path. e.g. "textures/123.jpg".
     * @param pBinary If true the file containes binary data and is returned as ArrayBuffer otherwise as string.
     */
    getResource(pFilename: string, pBinary: false): Promise<string>;
    getResource(pFilename: string, pBinary: true): Promise<ArrayBuffer>;
    getResource(pFilename: string, pBinary: boolean): Promise<ArrayBuffer | string>;
    /**
     * Retrieves a binary or text file from the container. If the file doesn't exist the promise
     * will be rejected with an FileNotFoundError.
     * @param pFilename Name of the file. May include a path. e.g. "textures/123.jpg".
     * @param pBinary If true the file containes binary data and is returned as ArrayBuffer otherwise as string.
     */
    getFile(pFilename: string, pBinary: false): Promise<string>;
    getFile(pFilename: string, pBinary: true): Promise<ArrayBuffer>;
    getFile(pFilename: string, pBinary: boolean): Promise<ArrayBuffer | string>;
    /**
     * Returns a list of files which are part of the given folder and its sub folders.
     * The files are returned as a relative path including the file name.
     * If the folder doesn't exist the promise will be rejected with an FileNotFoundError.
     * @param pFolderPath Path of the folder.
     */
    listFiles(pFolderPath?: string): Promise<Array<string>>;
}
export interface DocumentOutputContext extends DocumentIOContext {
    readonly geometryIds: ReadonlyArray<string>;
    readonly materialIds: ReadonlyArray<string>;
    readonly textureIds: ReadonlyArray<string>;
    readonly layerNames: ReadonlyArray<string>;
    addGeometry(pGeomId: string): number;
    addMaterial(pMatId: string): number;
    addTexture(pTexId: string): number;
    addLayer(pName: string): number;
    /**
     * Adds a binary or text resource which will be embedded in the document container.
     * @param pData The data of the file.
     * @param pFilename Name of the file to add. May include a path. e.g. "textures/123.jpg". Has to be lowercase.
     * @param pCompress If true the file should be compressed (if supported by the implementation)
     * otherwise the file should not be compressed. (Default: true)
     */
    addResource(pData: ArrayBuffer | Blob | string, pFilename: string, pCompress?: boolean): Promise<void>;
    /**
     * Adds a binary or text resource which will be embedded in the document container using an url.
     * @param pUrl The file url.
     * @param pFilename Name of the file to add. May include a path. e.g. "textures/123.jpg". Has to be lowercase.
     * @param pBinary If true the file containes binary data.
     * @param pCompress If true the file should be compressed (if supported by the implementation)
     * otherwise the file should not be compressed. (Default: true)
     */
    addResourceFromUrl(pUrl: string, pFilename: string, pBinary: boolean, pCompress?: boolean): Promise<void>;
    /**
     * Adds an already cached resource to the container.
     * True should be returned if the file was added, otherwise false (e.g. if the file was not found inside the cache).
     * @param pFilename Name of the file to add. May include a path. e.g. "textures/123.jpg". Has to be lowercase.
     * @param pCompress If true the file should be compressed (if supported by the implementation)
     * otherwise the file should not be compressed. (Default: true)
     */
    addResourceFromCache(pFilename: string, pCompress?: boolean): Promise<boolean>;
}
export interface DocumentContainer {
    /**
     * Adds a binary or text file to the container.
     * @param pData The data of the file.
     * @param pFilename Name of the file to add. May include a path. e.g. "textures/123.jpg". Has to be lowercase.
     * @param pCompress If true the file should be compressed (if supported by the implementation)
     * otherwise the file should not be compressed. (Default: true)
     * @param pCacheable If true the file may be cached. The filename should be used as cache id in this case. (Default: false)
     */
    addFile(pData: ArrayBuffer | Blob | string, pFilename: string, pCompress?: boolean, pCacheable?: boolean): Promise<void>;
    /**
     * Adds a binary or text file to the container using an url.
     * @param pUrl The file url.
     * @param pFilename Name of the file to add. May include a path. e.g. "textures/123.jpg". Has to be lowercase.
     * @param pBinary If true the file containes binary data.
     * @param pCompress If true the file should be compressed (if supported by the implementation)
     * otherwise the file should not be compressed. (Default: true)
     * @param pCacheable If true the file may be cached. The filename should be used as cache id in this case. (Default: false)
     */
    addFileFromUrl(pUrl: string, pFilename: string, pBinary: boolean, pCompress?: boolean, pCacheable?: boolean): Promise<void>;
    /**
     * Adds an already cached file to the container.
     * True should be returned if the file was added, otherwise false (e.g. if the file was not found inside the cache).
     * @param pFilename Name of the file to add. May include a path. e.g. "textures/123.jpg". Has to be lowercase.
     * @param pCompress If true the file should be compressed (if supported by the implementation)
     * otherwise the file should not be compressed. (Default: true)
     */
    addFileFromCache(pFilename: string, pCompress?: boolean): Promise<boolean>;
    /**
     * Retrieves a binary or text file from the container. If the file doesn't exist the promise
     * has to be rejected with an FileNotFoundError.
     * @param pFilename Name of the file. May include a path. e.g. "textures/123.jpg".
     * @param pBinary If true the file containes binary data and is returned as ArrayBuffer otherwise as string.
     */
    getFile(pFilename: string, pBinary: false): Promise<string>;
    getFile(pFilename: string, pBinary: true): Promise<ArrayBuffer>;
    getFile(pFilename: string, pBinary: boolean): Promise<ArrayBuffer | string>;
    /**
     * Returns a list of files which are part of the given folder and its sub folders.
     * The files are returned as a relative path including the file name.
     * If the folder doesn't exist the promise has to be rejected with an FileNotFoundError.
     * @param pFolderPath Path of the folder.
     */
    listFiles(pFolderPath?: string): Promise<Array<string>>;
}
export declare class FileNotFoundError extends CustomError {
    constructor(pFilename: string);
}
export declare class AbortError extends CustomError {
    constructor(pReason: string);
}
