import type { BoundingInfo } from "@babylonjs/core/Culling/boundingInfo";
import { Matrix, Vector3 } from "@babylonjs/core/Maths/math.vector";
import type { AbstractMesh } from "@babylonjs/core/Meshes/abstractMesh";
import type { TransformNode } from "@babylonjs/core/Meshes/transformNode";
import type { Frustum } from "../utils/MathUtils";
import type { SceneElement } from "./SceneElement";
export declare class BoundingBox {
    /** Left side point indices. */
    static readonly SIDE_MIN_X: number[];
    /** Right side point indices. */
    static readonly SIDE_MAX_X: number[];
    /** Bottom side point indices. */
    static readonly SIDE_MIN_Y: number[];
    /** Top side point indices. */
    static readonly SIDE_MAX_Y: number[];
    /** Front side point indices. */
    static readonly SIDE_MIN_Z: number[];
    /** Back side point indices. */
    static readonly SIDE_MAX_Z: number[];
    /** Indices of the 12 box edges */
    static readonly EDGES: number[][];
    min: Vector3;
    max: Vector3;
    constructor(pMin?: Vector3, pMax?: Vector3);
    getCenter(pWorldMatrix?: Matrix | null): Vector3;
    getSize(): Vector3;
    getRadius(): number;
    /**
     * Returns an array with the 8 box corners in fixed order.
     *
     * ```text
     * Order of points
     * ..6--------5
     * ./ |     / |
     * 1--|----2  |
     * |  7----|--4
     * | /     | /
     * 0-------3
     *
     * 0 = (min.x, min.y, min.z)
     * 1 = (min.x, max.y, min.z)
     * 2 = (max.x, max.y, min.z)
     * 3 = (max.x, min.y, min.z)
     * 4 = (max.x, min.y, max.z)
     * 5 = (max.x, max.y, max.z)
     * 6 = (min.x, max.y, max.z)
     * 7 = (min.x, min.y, max.z)
     * ```
     */
    getPoints(pWorldMatrix?: Matrix | null): Array<Vector3>;
    isValid(): boolean;
    reset(): void;
    clone(): BoundingBox;
    expand(pBox: BoundingBox): BoundingBox;
    expandInPlace(pBox: BoundingBox): BoundingBox;
    expandByPoint(pPoint: Vector3): BoundingBox;
    expandByPointInPlace(pPoint: Vector3): BoundingBox;
    transform(pMatrix: Matrix): BoundingBox;
    transformInPlace(pMatrix: Matrix): BoundingBox;
    isInFrustum(pFrustum: Frustum, pWorldMatrix?: Matrix | null): boolean;
    isCompletelyInFrustum(pFrustum: Frustum, pWorldMatrix?: Matrix | null): boolean;
    /**
     * Checks if this bounding box intersects with another bounding box.
     * @param pBounds The other bounding box.
     * @param pBorderIntersections If true, it counts as an intersection if the two bounding boxes are touching eachother.
     * @param pEpsilon Tolerance value for the check (default: utils.math.EPSILON)
     */
    intersects(pBounds: BoundingBox, pBorderIntersections: boolean, pEpsilon?: number): boolean;
    /**
     * Checks if this bounding box intersects with the given point.
     * @param pPoint The point.
     * @param pBorderIntersections If true, it counts as an intersection if the point is directly on the border of this bounding box.
     * @param pEpsilon Tolerance value for the check (default: utils.math.EPSILON)
     */
    intersectsPoint(pPoint: Vector3, pBorderIntersections: boolean, pEpsilon?: number): boolean;
    /**
     * Checks if this bounding box intersects with another bounding box given as a min and max vector.
     * @param pMin The min vector of the bonding box.
     * @param pMax The max vector of the bonding box.
     * @param pBorderIntersections If true, it counts as an intersection if the two bounding boxes are touching eachother.
     * @param pEpsilon Tolerance value for the check (default: utils.math.EPSILON)
     */
    intersectsMinMax(pMin: Vector3, pMax: Vector3, pBorderIntersections: boolean, pEpsilon?: number): boolean;
    /**
     * Calculates the bounding box for the given node. If pVisibleOnly is true invisible nodes are ignored.
     */
    static FromNode(pNode: TransformNode, pVisibleOnly?: boolean): BoundingBox;
    /**
     * Calculates the bounding box for the given mesh. If pVisibleOnly is true invisible nodes are ignored.
     */
    static FromMesh(pMesh: AbstractMesh, pVisibleOnly?: boolean): BoundingBox;
    /**
     * Calculates the bounding box for the given elements. If pVisibleOnly is true invisible nodes are ignored.
     */
    static FromElements(pElements: Array<SceneElement> | ReadonlyArray<SceneElement>, pUseWorldTransform?: boolean, pVisibleOnly?: boolean): BoundingBox;
    /**
     * Returns true if the given BabylonJS BoundingInfo is valid.
     */
    static IsBoundingInfoValid(pInfo: BoundingInfo): boolean;
    private static IsMinMaxValid;
    private isBoundingSphereInFrustum;
    private expandByNodeInPlace;
}
