import { Mesh } from "@babylonjs/core/Meshes/mesh";
import { Log } from "../../utils";
import { isNullOrEmpty } from "../../utils/string";
import { ResourceCache } from "./ResourceCache";
import { KEY_IS_FALLBACK } from "./internal/GeometrySerialization";
export class GeometryManager {
    get scene() {
        return (this.mScene);
    }
    constructor(pModelManager) {
        this.onDisposeMesh = (pMesh, pCacheCleared) => {
            if (pMesh == null)
                return;
            pMesh.dispose();
        };
        this.mScene = pModelManager.scene;
        this.mMeshCache = new ResourceCache();
        this.mMeshCache.onDisposeResource = this.onDisposeMesh;
        this.mLastCacheCleanUp = Date.now();
        this.missingGeometries = new Set();
    }
    clear() {
        this.mLastCacheCleanUp = Date.now();
        this.mMeshCache.clear();
        this.missingGeometries.clear();
    }
    addGeometryMesh(pId, pGeometryMesh) {
        if (isNullOrEmpty(pId) || pGeometryMesh == null || pGeometryMesh.geometry == null)
            return (false);
        const tGeom = this.getGeometry(pId);
        if (tGeom == null || this.isFallbackGeometry(tGeom)) {
            //dispose fallback geometry
            if (tGeom != null)
                this.mMeshCache.remove(pId, true);
            this.mMeshCache.add(pId, pGeometryMesh);
            //force correct id
            pGeometryMesh.id = pId;
            pGeometryMesh.geometry.id = pId;
            this.mScene.removeMesh(pGeometryMesh); //remove from scene, only keep in cache
            //remove from missing list
            this.missingGeometries.delete(pId);
            return (true);
        }
        return (false);
    }
    getGeometryMesh(pGeomId) {
        if (isNullOrEmpty(pGeomId))
            return (null);
        //check cache
        const tMesh = this.mMeshCache.get(pGeomId);
        if (tMesh == null || tMesh.geometry == null)
            return (null);
        return (tMesh);
    }
    getGeometry(pId) {
        const tMesh = this.getGeometryMesh(pId);
        if (tMesh != null)
            return (tMesh.geometry);
        else
            return (null);
    }
    hasGeometry(pId) {
        if (isNullOrEmpty(pId))
            return (false);
        return (this.mMeshCache.has(pId));
    }
    isFallbackGeometry(pGeometry) {
        if (pGeometry == null)
            return (false);
        return (pGeometry[KEY_IS_FALLBACK] === true);
    }
    getMeshGeometryId(pMesh) {
        if (!(pMesh instanceof Mesh) || pMesh.geometry == null)
            return (null);
        if (!this.hasGeometry(pMesh.geometry.id))
            return (null);
        return (pMesh.geometry.id);
    }
    //// internal //////////////////////////////////////////////////////////////////////////////
    /** @internal */
    _cleanUpCache(pImmediately) {
        const tNow = Date.now();
        if (pImmediately || tNow - this.mLastCacheCleanUp > GeometryManager.CACHE_CLEANUP_INTERVAL) {
            this.mLastCacheCleanUp = tNow;
            const tMaxTimeUnused = pImmediately ? 0 : GeometryManager.CACHE_CLEANUP_INTERVAL;
            this.cleanUpMeshCache(tMaxTimeUnused);
        }
    }
    cleanUpMeshCache(pMaxTimeUnunsed) {
        //collect used geometries
        const tUsedGeoms = new Set();
        this.mMeshCache.forEach((pId, pMesh) => {
            if (pMesh != null && pMesh.geometry != null) {
                const tGeom = pMesh.geometry;
                if (tGeom._meshes != null) {
                    /* eslint-disable-next-line @typescript-eslint/no-unsafe-member-access */
                    if (tGeom._meshes.length > 1) {
                        tUsedGeoms.add(pId);
                    }
                }
                else
                    Log.error("Private member Geometry._meshes is undefined.");
            }
            else
                tUsedGeoms.add(pId); //do not reload empty geometries
        });
        //remove unused
        this.mMeshCache.removeUnused(tUsedGeoms, GeometryManager.CACHE_CLEANUP_INTERVAL);
    }
}
//// private ///////////////////////////////////////////////////////////////////////////////
GeometryManager.CACHE_CLEANUP_INTERVAL = 120000;
