import { Matrix } from "@babylonjs/core/Maths/math.vector";
import type { TransformNode } from "@babylonjs/core/Meshes/transformNode";
import type { Node } from "@babylonjs/core/node";
import { ListenableEvent } from "../../utils";
import type { DocumentOutputContext } from "../io/Types";
import type { Property, PropertyClass, PropertyProvider } from "../prop";
import { PropertyChangedResult } from "../prop";
import { BoundingBox } from "./BoundingBox";
import type * as ioTypes from "./io/Types";
import type { ModelManager } from "./ModelManager";
export declare const ElementChangedFlags: {
    readonly None: 0;
    readonly Added: 1;
    readonly Removed: 2;
    readonly BoundingBox: 4;
    readonly Transformation: 8;
    readonly Properties: 16;
    readonly PropertyValues: 32;
};
export type ElementChangedFlags = number;
export declare const ElementCategory: {
    readonly None: 0;
    readonly Interior: 1;
    readonly FloorPlan: 2;
    readonly Wall: 3;
    readonly WallOpening: 4;
    readonly WallAddition: 5;
};
export type ElementCategory = typeof ElementCategory[keyof typeof ElementCategory];
export declare const UndoOperation: {
    readonly Delete: 0;
    readonly Transform: 1;
};
export type UndoOperation = typeof UndoOperation[keyof typeof UndoOperation];
export declare class SceneElement implements PropertyProvider {
    get node(): TransformNode;
    get isSelected(): boolean;
    get isDisposed(): boolean;
    get transform(): Matrix;
    get transformInverted(): Matrix;
    get worldTransform(): Matrix;
    get worldTransformInverted(): Matrix;
    /**
     * Complete bounding box of this element inluding invisible nodes.
     */
    get boundingBox(): BoundingBox;
    /**
     * Bounding box of this element inluding only visible elements.
     */
    get boundingBoxVisible(): BoundingBox;
    get category(): ElementCategory;
    readonly eventAdded: ListenableEvent<void, SceneElement>;
    readonly eventRemoved: ListenableEvent<void, SceneElement>;
    readonly eventBoundingBoxChanged: ListenableEvent<void, SceneElement>;
    readonly eventTransformationChanged: ListenableEvent<void, SceneElement>;
    readonly eventPropertiesChanged: ListenableEvent<PropertyChangedResult, PropertyProvider>;
    constructor(pModel: ModelManager, pNode: TransformNode);
    /**
     * Should be called if the element was removed from the scene and will never be used again.
     */
    dispose(): void;
    canSelect(): boolean;
    canSelectSubElement(): boolean;
    canTranslate(): boolean;
    canRotate(): boolean;
    canDelete(): boolean;
    setTransform(pTransform: Matrix): boolean;
    /**
     * Should be called after the transformation of this.node was modified directly instead
     * of using this.setTransform(). This function updates frozen world matrices and
     * calls notifyChange with ElementChangedFlags.Transformation.
     */
    updateTransformation(pUpdateChildNodes?: boolean): void;
    getMainElement(): SceneElement;
    getParentElement(): SceneElement | null;
    getTopLevelElement(): SceneElement;
    /**
     * Returns this element if it is currently selectable otherwise the next parent
     * which is selectable.
     */
    getSelectableElement(): SceneElement;
    /**
     * Returns the next selectable parent or this if there is no parent.
     */
    getSelectableParentElement(): SceneElement;
    /**
     * Can be overridden to define a custom bounding box for the selection highlight.
     * Defaults to SceneElement.boundingBoxVisible (does not return a clone).
     */
    getSelectionHighlightBoundingBox(): BoundingBox;
    /**
     * Can be overridden to disable the default selection highlight for this element. (Default: true)
     */
    isSelectionHighlightEnabled(): boolean;
    notifyChange(pFlags: ElementChangedFlags, pNotifyParent?: boolean): void;
    beginChange(): void;
    endChange(): boolean;
    replaceNode(pNewNode: TransformNode, pKeepTransfom: boolean, pKeepParent: boolean, pNotifyParent?: boolean): void;
    clone(pNewParent?: TransformNode | null): SceneElement | null;
    isInScene(): boolean;
    getNodes(pIncludeChildElementNodes: boolean): Array<TransformNode>;
    serialize(pContext: DocumentOutputContext): ioTypes.SceneElement | null;
    showInteractors(pShow: boolean): void;
    /**
     * Can be used to provide additional undo related data for certain operations.
     * Will be called before the operation is performed.
     */
    saveUndoData(pOperation: UndoOperation): object | null;
    /**
     * Will be called for certain operations to load the undo data which was
     * saved with a previous saveUndoData() call.
     */
    loadUndoData(pOperation: UndoOperation, pUndoData: object | null): void;
    getProperties(): Promise<Array<Property> | null>;
    getPropertyClasses(): Promise<Array<PropertyClass> | null>;
    beginPropertyChange(): void;
    endPropertyChange(): void;
    static FromNode(pNode: Node, pCheckParents: boolean): SceneElement | null;
    protected mNode: TransformNode;
    protected mModel: ModelManager;
    protected setNode(pNode: TransformNode): void;
    protected copyDataFrom(pElement: SceneElement): void;
    protected handleImmediateChanges(): void;
    protected handleDelayedChanges(): void;
    protected wasAllFlagsChanged(pFlag: ElementChangedFlags): boolean;
    protected wasFlagChanged(pFlag: ElementChangedFlags): boolean;
    protected static cloneNode(pNode: TransformNode, pNewParent?: TransformNode | null, pChildNodes?: boolean): TransformNode;
    private static readonly NODEDATA_SCENEELEMENT;
    private mBoundingBox;
    private mBoundingBoxVisible;
    private mTransform;
    private mTransformInverted;
    private mWorldTransform;
    private mWorldTransformInverted;
    private mIsSelected;
    private mIsInScene;
    private mChangedFlags;
    private mChangeDepth;
    private mChangeNotifyParent;
    private updateCachedTransforms;
    private collectNodes;
}
