import { Matrix } from "@babylonjs/core/Maths/math.vector";
import { ListenableEvent, Log } from "../../utils";
import { PropertyChangedResult } from "../prop";
import { getLocalMatrix, removeFromScene, setLocalMatrix, updateFrozenWorldMatrix } from "../utils/MeshUtils";
import { BoundingBox } from "./BoundingBox";
import { NodeUserData } from "./NodeUserData";
export const ElementChangedFlags = {
    None: 0x00000000,
    Added: 0x00000001,
    Removed: 0x00000002,
    BoundingBox: 0x00000004,
    Transformation: 0x00000008,
    Properties: 0x00000010,
    PropertyValues: 0x00000020
};
export const ElementCategory = {
    None: 0,
    Interior: 1,
    FloorPlan: 2,
    Wall: 3,
    WallOpening: 4,
    WallAddition: 5
};
export const UndoOperation = {
    Delete: 0,
    Transform: 1
};
export class SceneElement {
    get node() {
        return (this.mNode);
    }
    get isSelected() {
        return (this.mIsSelected);
    }
    get isDisposed() {
        return (this.mNode === undefined); //see dispose()
    }
    get transform() {
        this.updateCachedTransforms();
        return (this.mTransform);
    }
    get transformInverted() {
        this.updateCachedTransforms();
        return (this.mTransformInverted);
    }
    get worldTransform() {
        this.updateCachedTransforms();
        return (this.mWorldTransform);
    }
    get worldTransformInverted() {
        this.updateCachedTransforms();
        return (this.mWorldTransformInverted);
    }
    /**
     * Complete bounding box of this element inluding invisible nodes.
     */
    get boundingBox() {
        if (this.mBoundingBox == null) {
            this.mBoundingBox = this.mNode == null ? new BoundingBox() :
                BoundingBox.FromNode(this.mNode);
        }
        return (this.mBoundingBox);
    }
    /**
     * Bounding box of this element inluding only visible elements.
     */
    get boundingBoxVisible() {
        if (this.mBoundingBoxVisible == null) {
            this.mBoundingBoxVisible = this.mNode == null ? new BoundingBox() :
                BoundingBox.FromNode(this.mNode, true);
        }
        return (this.mBoundingBoxVisible);
    }
    get category() {
        return (ElementCategory.None);
    }
    constructor(pModel, pNode) {
        this.eventAdded = new ListenableEvent(this);
        this.eventRemoved = new ListenableEvent(this);
        this.eventBoundingBoxChanged = new ListenableEvent(this);
        this.eventTransformationChanged = new ListenableEvent(this);
        this.eventPropertiesChanged = new ListenableEvent(this);
        this.mModel = pModel;
        this.mBoundingBox = null;
        this.mBoundingBoxVisible = null;
        this.mTransform = null;
        this.mTransformInverted = Matrix.Identity();
        this.mWorldTransform = null;
        this.mWorldTransformInverted = Matrix.Identity();
        this.mIsSelected = false;
        this.mIsInScene = true;
        this.mChangedFlags = ElementChangedFlags.None;
        this.mChangeDepth = 0;
        this.mChangeNotifyParent = false;
        this.setNode(pNode);
    }
    /**
     * Should be called if the element was removed from the scene and will never be used again.
     */
    dispose() {
        if (this.mNode != null) {
            this.mNode.dispose();
            NodeUserData.SetData(this.mNode, SceneElement.NODEDATA_SCENEELEMENT, null);
        }
        this.mNode = undefined;
        this.mBoundingBox = null;
        this.mBoundingBoxVisible = null;
        this.mIsInScene = false;
    }
    canSelect() {
        return (this.mNode != null && this.isInScene());
    }
    canSelectSubElement() {
        return (false);
    }
    canTranslate() {
        return (true);
    }
    canRotate() {
        return (true);
    }
    canDelete() {
        return (true);
    }
    setTransform(pTransform) {
        if (this.mNode == null)
            return (false);
        const tChanged = setLocalMatrix(this.mNode, pTransform);
        if (tChanged)
            this.updateTransformation();
        return (tChanged);
    }
    /**
     * Should be called after the transformation of this.node was modified directly instead
     * of using this.setTransform(). This function updates frozen world matrices and
     * calls notifyChange with ElementChangedFlags.Transformation.
     */
    updateTransformation(pUpdateChildNodes = true) {
        if (this.mNode != null) {
            updateFrozenWorldMatrix(this.mNode, pUpdateChildNodes);
            this.notifyChange(ElementChangedFlags.Transformation);
        }
    }
    getMainElement() {
        return (this);
    }
    getParentElement() {
        if (this.mNode != null && this.mNode.parent != null)
            return (SceneElement.FromNode(this.mNode.parent, true));
        return (null);
    }
    getTopLevelElement() {
        const tParent = this.getParentElement();
        if (tParent != null)
            return (tParent.getTopLevelElement());
        else
            return (this);
    }
    /**
     * Returns this element if it is currently selectable otherwise the next parent
     * which is selectable.
     */
    getSelectableElement() {
        if (this.canSelect())
            return (this);
        return (this.getSelectableParentElement());
    }
    /**
     * Returns the next selectable parent or this if there is no parent.
     */
    getSelectableParentElement() {
        const tParent = this.getParentElement();
        if (tParent != null)
            return (tParent.getSelectableElement());
        else
            return (this);
    }
    /**
     * Can be overridden to define a custom bounding box for the selection highlight.
     * Defaults to SceneElement.boundingBoxVisible (does not return a clone).
     */
    getSelectionHighlightBoundingBox() {
        return (this.boundingBoxVisible);
    }
    /**
     * Can be overridden to disable the default selection highlight for this element. (Default: true)
     */
    isSelectionHighlightEnabled() {
        return (true);
    }
    notifyChange(pFlags, pNotifyParent = true) {
        if (pFlags == null || pFlags === ElementChangedFlags.None)
            return;
        this.mChangedFlags |= pFlags;
        this.mChangeNotifyParent = this.mChangeNotifyParent || pNotifyParent;
        this.handleImmediateChanges();
        if (this.mChangeDepth === 0) {
            this.handleDelayedChanges();
            //trigger events
            if (this.wasFlagChanged(ElementChangedFlags.Added))
                this.eventAdded.trigger();
            let tWasRemoved = false;
            if (this.wasFlagChanged(ElementChangedFlags.Removed)) {
                this.eventRemoved.trigger();
                tWasRemoved = true;
            }
            if (this.wasFlagChanged(ElementChangedFlags.BoundingBox))
                this.eventBoundingBoxChanged.trigger();
            if (this.wasFlagChanged(ElementChangedFlags.Transformation))
                this.eventTransformationChanged.trigger();
            if (this.wasFlagChanged(ElementChangedFlags.Properties |
                ElementChangedFlags.PropertyValues)) {
                this.eventPropertiesChanged.trigger(this.wasFlagChanged(ElementChangedFlags.Properties) ?
                    PropertyChangedResult.Properties : PropertyChangedResult.Value);
            }
            //notify parent that its bounding box has changed
            if (pNotifyParent && this.wasFlagChanged(ElementChangedFlags.Added |
                ElementChangedFlags.Removed |
                ElementChangedFlags.BoundingBox |
                ElementChangedFlags.Transformation)) {
                const tParent = this.getParentElement();
                if (tParent != null)
                    tParent.notifyChange(ElementChangedFlags.BoundingBox);
            }
            //notify model that there was a modification
            if (this.mChangedFlags !== ElementChangedFlags.None && (tWasRemoved || this.isInScene()))
                this.mModel.notifyModification();
            this.mChangedFlags = ElementChangedFlags.None;
            this.mChangeNotifyParent = false;
        }
    }
    beginChange() {
        this.mChangeDepth++;
    }
    endChange() {
        this.mChangeDepth--;
        if (this.mChangeDepth <= 0) {
            this.mChangeDepth = 0;
            this.notifyChange(this.mChangedFlags, this.mChangeNotifyParent);
            return (true);
        }
        return (false);
    }
    replaceNode(pNewNode, pKeepTransfom, pKeepParent, pNotifyParent = true) {
        if (pNewNode == null)
            return;
        if (this.isDisposed) {
            Log.warn("New node assigned to already disposed SceneElement. New node will be disposed.");
            pNewNode.dispose();
            return;
        }
        if (!this.isInScene()) {
            //remove new node from scene
            removeFromScene(pNewNode, this.mModel.scene);
        }
        const tOldNode = this.mNode;
        if (tOldNode == null) {
            this.setNode(pNewNode);
            return;
        }
        let tChangeFlags = ElementChangedFlags.BoundingBox;
        this.mNode = pNewNode;
        if (pKeepParent)
            this.mNode.parent = tOldNode.parent;
        //assign NodeUserData to new node
        const tUserData = NodeUserData.FromNode(tOldNode, false);
        if (tUserData != null) {
            tUserData.assignToNode(this.mNode);
            tUserData.setData(SceneElement.NODEDATA_SCENEELEMENT, this);
        }
        //copy transformation from old to new node
        if (pKeepTransfom) {
            const tMatrix = getLocalMatrix(tOldNode);
            setLocalMatrix(this.mNode, tMatrix);
            updateFrozenWorldMatrix(this.mNode);
        }
        else
            tChangeFlags |= ElementChangedFlags.Transformation;
        //dispose old node
        tOldNode.dispose();
        this.notifyChange(tChangeFlags, pNotifyParent);
    }
    clone(pNewParent) {
        const tClone = new SceneElement(this.mModel, SceneElement.cloneNode(this.mNode, pNewParent));
        tClone.copyDataFrom(this);
        return (tClone);
    }
    isInScene() {
        if (!this.mIsInScene)
            return (false);
        //check parent
        const tParent = this.getParentElement();
        if (tParent != null)
            return (tParent.isInScene());
        return (true);
    }
    getNodes(pIncludeChildElementNodes) {
        const tNodes = new Array();
        if (this.mNode != null) {
            tNodes.push(this.mNode);
            this.collectNodes(this.mNode, tNodes, pIncludeChildElementNodes);
        }
        return (tNodes);
    }
    serialize(pContext) {
        return (null);
    }
    showInteractors(pShow) {
    }
    /**
     * Can be used to provide additional undo related data for certain operations.
     * Will be called before the operation is performed.
     */
    saveUndoData(pOperation) {
        return (null);
    }
    /**
     * Will be called for certain operations to load the undo data which was
     * saved with a previous saveUndoData() call.
     */
    loadUndoData(pOperation, pUndoData) {
    }
    ////
    //implementation of PropertyProvider
    getProperties() {
        return (Promise.resolve(null));
    }
    getPropertyClasses() {
        return (Promise.resolve(null));
    }
    beginPropertyChange() {
        this.beginChange();
    }
    endPropertyChange() {
        this.endChange();
    }
    static FromNode(pNode, pCheckParents) {
        if (pNode == null)
            return (null);
        const tSceneElement = NodeUserData.GetData(pNode, SceneElement.NODEDATA_SCENEELEMENT);
        if (tSceneElement == null) {
            if (pCheckParents && pNode.parent != null)
                return (SceneElement.FromNode(pNode.parent, pCheckParents));
            else
                return (null);
        }
        return (tSceneElement);
    }
    //// internal //////////////////////////////////////////////////////////////////////////////
    /** @internal */
    _setIsSelected(pSelected) {
        this.mIsSelected = pSelected;
    }
    setNode(pNode) {
        this.mNode = pNode;
        if (this.mNode != null) {
            //set 'this' as user data, necessary for fromNode function
            NodeUserData.SetData(this.mNode, SceneElement.NODEDATA_SCENEELEMENT, this);
        }
        this.notifyChange(ElementChangedFlags.BoundingBox | ElementChangedFlags.Transformation);
    }
    copyDataFrom(pElement) {
    }
    handleImmediateChanges() {
        //reset cached bounding box
        if (this.wasFlagChanged(ElementChangedFlags.BoundingBox)) {
            this.mBoundingBox = null;
            this.mBoundingBoxVisible = null;
        }
        //reset cached transform
        if (this.wasFlagChanged(ElementChangedFlags.Transformation)) {
            this.mTransform = null;
            this.mWorldTransform = null;
        }
        if (this.wasFlagChanged(ElementChangedFlags.Added))
            this.mIsInScene = true;
        if (this.wasFlagChanged(ElementChangedFlags.Removed))
            this.mIsInScene = false;
    }
    handleDelayedChanges() {
    }
    wasAllFlagsChanged(pFlag) {
        return ((this.mChangedFlags & pFlag) === pFlag);
    }
    wasFlagChanged(pFlag) {
        return ((this.mChangedFlags & pFlag) != 0);
    }
    static cloneNode(pNode, pNewParent, pChildNodes) {
        if (pChildNodes != null && pChildNodes) {
            // a child node may belong to a SceneElement, clone the SceneElement in this case
            const tElememt = SceneElement.FromNode(pNode, false);
            if (tElememt != null) {
                const tClonedElement = tElememt.clone(pNewParent);
                if (tClonedElement != null)
                    return (tClonedElement.node);
            }
        }
        //clone the node and its user data (keep the name)
        const tClone = pNode.clone(pNode.name, pNewParent, true);
        if (tClone == null)
            throw new Error("Failed to clone node");
        const tUserData = NodeUserData.FromNode(pNode);
        if (tUserData != null) {
            const tClonedUserData = tUserData.clone();
            tClonedUserData.assignToNode(tClone);
        }
        if (pNode.isWorldMatrixFrozen)
            tClone.freezeWorldMatrix();
        //clone the children
        const tChilds = pNode.getChildTransformNodes(true);
        for (const tChild of tChilds) {
            this.cloneNode(tChild, tClone, true);
        }
        return (tClone);
    }
    updateCachedTransforms() {
        if (this.mTransform == null) {
            this.mTransform = this.mNode == null ? Matrix.Identity() : getLocalMatrix(this.mNode);
            this.mTransform.invertToRef(this.mTransformInverted);
        }
        if (this.mWorldTransform == null) {
            if (this.mNode != null) {
                this.mNode.computeWorldMatrix(true); //force update of the worldmatrix
                this.mWorldTransform = this.mNode.getWorldMatrix();
            }
            else {
                this.mWorldTransform = Matrix.Identity();
            }
            this.mWorldTransform.invertToRef(this.mWorldTransformInverted);
        }
    }
    collectNodes(pParent, pNodes, pIncludeChildElementNodes) {
        if (pParent == null)
            return;
        const tChilds = pParent.getChildTransformNodes(true);
        for (const tChild of tChilds) {
            const tElement = SceneElement.FromNode(tChild, false);
            if (tElement == null || pIncludeChildElementNodes) {
                pNodes.push(tChild);
                this.collectNodes(tChild, pNodes, pIncludeChildElementNodes);
            }
        }
    }
}
//// private ///////////////////////////////////////////////////////////////////////////////
SceneElement.NODEDATA_SCENEELEMENT = "core:sceneElement";
