import { Engine } from "@babylonjs/core/Engines/engine";
import { Texture as BJSTexture } from "@babylonjs/core/Materials/Textures/texture";
import { getDataPath, Log } from "../../../utils";
import { ERR_INVALID_ARGUMENT } from "../../../utils/error";
import { isNotNullOrEmpty, joinPath } from "../../../utils/string";
import { FileNotFoundError } from "../../io/Types";
import { deserializePbrMaterial, serializePbrMaterial } from "./PbrMaterialSerialization";
/** @internal */
export function serializeMaterials(pManager, pContext) {
    if (pContext == null || pContext.materialIds == null)
        throw new Error(ERR_INVALID_ARGUMENT);
    const tIOMaterials = new Array();
    for (const tMatId of pContext.materialIds) {
        const tMat = pManager.getMaterial(tMatId);
        if (tMat != null) {
            const tIOMat = serializePbrMaterial(pContext, tMat);
            tIOMaterials.push(tIOMat);
        }
        else {
            Log.warn("Unknown material id: " + tMatId);
            tIOMaterials.push({ id: tMatId });
        }
    }
    return (tIOMaterials);
}
/** @internal */
export function deserializeMaterials(pManager, pMaterials, pContext) {
    if (pMaterials == null || pContext == null)
        return;
    for (const tIOMat of pMaterials) {
        const tMat = deserializePbrMaterial(tIOMat, pContext, pManager);
        if (tMat != null)
            pManager.addMaterial(tIOMat.id, tMat);
    }
}
/** @internal */
export function serializeTextures(pManager, pContext) {
    if (pContext == null || pContext.textureIds == null)
        throw new Error(ERR_INVALID_ARGUMENT);
    const tIOTextures = new Array();
    for (const tTextureId of pContext.textureIds) {
        const tTexture = pManager.getTexture(tTextureId);
        if (tTexture != null) {
            const tResTex = pManager._getResourceTextures().get(tTextureId);
            if (tResTex != null && tTexture.isResource) {
                const tIOTexture = {};
                //id
                tIOTexture.id = tTextureId;
                //type
                if (tResTex.type !== "jpg")
                    tIOTexture.type = tResTex.type;
                //name
                if (tTexture.name != tTextureId && isNotNullOrEmpty(tTexture.name))
                    tIOTexture.name = tTexture.name;
                //transparent
                if (tTexture.hasAlpha)
                    tIOTexture.transparent = true;
                tIOTextures.push(tIOTexture);
            }
        }
        else {
            Log.warn("Unknown texture id: " + tTextureId);
            tIOTextures.push({ id: tTextureId });
        }
    }
    return (tIOTextures);
}
/** @internal */
export async function addTexturesToContainer(pManager, pTextureIds, pContainer, pProgressCallback) {
    if (pTextureIds == null || pContainer == null)
        throw new Error(ERR_INVALID_ARGUMENT);
    const tCount = pTextureIds.length;
    for (let i = 0; i < tCount; ++i) {
        await addTextureToContainer(pManager, pTextureIds[i], pContainer);
        if (pProgressCallback != null)
            pProgressCallback((i / tCount) * 100.0);
    }
}
/** @internal */
export async function readTexturesFromContainer(pManager, pTextures, pContainer, pProgressCallback) {
    if (pTextures == null || pContainer == null)
        throw new Error(ERR_INVALID_ARGUMENT);
    const tCount = pTextures.length;
    for (let i = 0; i < tCount; ++i) {
        await readTextureFromContainer(pManager, pTextures[i], pContainer);
        if (pProgressCallback != null)
            pProgressCallback((i / tCount) * 100.0);
    }
}
//// internals /////////////////////////////////////////////////////////////////////////////////////
function addTextureToContainer(pManager, pTextId, pContainer) {
    const tResTex = pManager._getResourceTextures().get(pTextId);
    const tTexture = pManager.getTexture(pTextId);
    if (tResTex == null || tTexture == null) {
        Log.warn("Unknown texture id: " + pTextId);
        return (Promise.resolve());
    }
    if (tResTex.data == null) {
        Log.warn("Missing texture data for id: " + pTextId);
        return (Promise.resolve());
    }
    //do not embed fallback textures
    if (tTexture.isFallback)
        return (Promise.resolve());
    const tFilename = "textures/" + pTextId + "." + tResTex.type;
    //try use texture from cache first
    //no compression for textures, jpg and png are already compressed
    const tPromise = pContainer.addFileFromCache(tFilename, false).then((pAdded) => {
        if (pAdded)
            return (Promise.resolve());
        else
            return (pContainer.addFile(tResTex.data, tFilename, false, true));
    }).catch((pError) => {
        Log.error("Failed to embed texture " + pTextId, pError);
    });
    return (tPromise);
}
function readTextureFromContainer(pManager, pTex, pContainer) {
    if (pTex == null || pManager.hasTexture(pTex.id))
        return (Promise.resolve());
    let tType = "jpg";
    if (pTex.type != null)
        tType = pTex.type;
    const tHasAlpha = pTex.transparent != null && pTex.transparent;
    const tPromise = pContainer.getFile("textures/" + pTex.id + "." + tType, true).then((pBuffer) => {
        const tTexture = pManager._internalCreateResourceTexture(pTex.id, pTex.name, tType, pBuffer, tHasAlpha);
        if (tTexture == null) {
            Log.error("Failed to load texture " + pTex.id);
        }
    }).catch((pError) => {
        //use fallback texture
        createFallbackTexture(pManager, pTex.id, pTex.name, tType, tHasAlpha);
        //register as missing texture
        pManager.missingResourceTextures.add(pTex.id);
        if (!(pError instanceof FileNotFoundError))
            Log.error("Failed to get embedded texture " + pTex.id, pError);
    });
    return (tPromise);
}
function createFallbackTexture(pManager, pId, pName, pType, pHasAlpha) {
    //always create as RGBA beause it may be reused for both cases
    const tTexture = pManager._internalCreateTexture(pId, pName, joinPath(getDataPath(), "core/textures/fallback.png"), true, Engine.TEXTUREFORMAT_RGBA, BJSTexture.NEAREST_SAMPLINGMODE);
    tTexture.isFallback = true;
    tTexture.hasAlpha = pHasAlpha;
    const tResTex = {
        data: null,
        blobUrl: null,
        type: pType
    };
    pManager._getResourceTextures().set(pId, tResTex);
    return (tTexture);
}
