import { VertexBuffer } from "@babylonjs/core/Buffers/buffer";
import { Geometry } from "@babylonjs/core/Meshes/geometry";
import { Mesh } from "@babylonjs/core/Meshes/mesh";
import { VertexData } from "@babylonjs/core/Meshes/mesh.vertexData";
import { Log } from "../../../utils";
import { fromDataUri } from "../../../utils/arraybuffer";
import { forEachParallel, ajax } from "../../../utils/async";
import { getFilePath, isDataUri, isNullOrEmpty } from "../../../utils/string";
import { GltfMesh, GltfTypes } from "./GltfMesh";
/** @internal */
export class GltfMeshImport {
    constructor(pScene) {
        this.onBufferLoaded = (pBufferId, pData) => {
            if (pData == null || this.mArrayBuffers == null) {
                Log.warn("Failed to load buffer " + pBufferId);
                return;
            }
            //bufferLength is optional in glTF v1.0
            const tBufferLength = this.mGltf.buffers[pBufferId].byteLength;
            if (tBufferLength != null && pData.byteLength != tBufferLength) {
                Log.warn("Buffer " + pBufferId + " has length " + pData.byteLength +
                    ". Expected: " + tBufferLength);
            }
            this.mArrayBuffers[pBufferId] = pData;
        };
        this.mScene = pScene;
        this.mConvertToLeftHanded = true;
        this.mIsVersion1 = false;
        this.mRootUrl = "";
        this.mArrayBuffers = null;
        this.mBinaryGltfOffset = 0;
    }
    importGltf(pGltf, pGltfUrl) {
        if (pGltf == null || pGltfUrl == null)
            return (Promise.resolve(null));
        this.mRootUrl = getFilePath(pGltfUrl);
        this.mGltf = pGltf;
        this.prepareLoading();
        //load buffers first
        const tPromise = this.loadBuffers().then(() => {
            return (this.importMeshes());
        });
        return (tPromise);
    }
    importBinaryGltf(pBuffer) {
        if (pBuffer == null)
            return (null);
        const tResult = this.parseBinaryGltfHeader(pBuffer);
        if (tResult == null)
            throw new Error("Failed to load binary glTF");
        this.mGltf = tResult.gltf;
        this.prepareLoading();
        if (this.mIsVersion1)
            this.mArrayBuffers["KHR_binary_glTF"] = tResult.arrayBuffer;
        else
            this.mArrayBuffers.push(tResult.arrayBuffer);
        this.mBinaryGltfOffset = tResult.binaryOffset;
        return (this.importMeshes());
    }
    static LoadArrayBuffer(pUrl) {
        if (isNullOrEmpty(pUrl))
            return (Promise.resolve(null));
        return (ajax("GET", pUrl, null, { responseType: "arraybuffer" }));
    }
    prepareLoading() {
        //check version
        this.mIsVersion1 = false;
        if (this.mGltf.asset.version == null || this.mGltf.asset.version.startsWith("1"))
            this.mIsVersion1 = true;
        //check if gltf uses a left handed coordinate system
        //this is necessary for backwards compatibility with old W-CF documents
        this.mConvertToLeftHanded = true;
        if (this.mGltf.asset.extras != null) {
            if (this.mGltf.asset.extras.EGR_leftHandedCS != null) {
                this.mConvertToLeftHanded = !this.mGltf.asset.extras.EGR_leftHandedCS;
            }
        }
        if (this.mIsVersion1)
            this.mArrayBuffers = {};
        else
            this.mArrayBuffers = [];
    }
    importMeshes() {
        const tMeshes = new Array();
        if (this.mIsVersion1) {
            for (const tGltfMeshId in this.mGltf.meshes) {
                /* eslint-disable-next-line no-prototype-builtins */
                if (this.mGltf.meshes.hasOwnProperty(tGltfMeshId)) {
                    const tMesh = this.importMesh(this.mGltf.meshes[tGltfMeshId]);
                    if (tMesh != null)
                        tMeshes.push(tMesh);
                }
            }
        }
        else {
            for (const tGltfMesh of this.mGltf.meshes) {
                const tMesh = this.importMesh(tGltfMesh);
                if (tMesh != null)
                    tMeshes.push(tMesh);
            }
        }
        return (tMeshes);
    }
    getStringFromTypedArray(pUInt8Array, pByteOffset, pByteLength) {
        const tOffset = pByteOffset != null ? pByteOffset : 0;
        let tLength = pByteLength != null ? pByteLength : pUInt8Array.byteLength - tOffset;
        tLength = Math.min(tLength, pUInt8Array.byteLength);
        const tSubArray = pUInt8Array.subarray(tOffset, tOffset + tLength);
        let tResult = "";
        for (let i = 0; i < tSubArray.length; ++i) {
            tResult += String.fromCharCode(tSubArray[i]);
        }
        return (tResult);
    }
    parseBinaryGltfHeader(pArrayBuffer) {
        const tHeader = new Uint32Array(pArrayBuffer, 0, 5);
        //check header
        if (tHeader[0] !== GltfMesh._GLTF_HEADER_MAGIC) {
            Log.error("Not a valid binary glTF file");
            return (null);
        }
        //check version
        const tVersion = tHeader[1];
        if (tVersion !== 1 && tVersion !== 2) {
            Log.error("Invalid binary glTF version: " + tVersion);
            return (null);
        }
        const tContentLength = tHeader[3];
        const tContentOffset = 20;
        let tBinOffset = tContentOffset + tContentLength;
        if (tVersion > 1)
            tBinOffset += 8;
        const tJsonString = this.getStringFromTypedArray(new Uint8Array(pArrayBuffer), tContentOffset, tContentLength);
        //check if gltf has 4 byte alignment
        if (tBinOffset % 4 === 0) {
            //use buffer directly
            return {
                gltf: JSON.parse(tJsonString),
                arrayBuffer: pArrayBuffer,
                binaryOffset: tBinOffset
            };
        }
        else {
            //cut off header
            return {
                gltf: JSON.parse(tJsonString),
                arrayBuffer: pArrayBuffer.slice(tBinOffset),
                binaryOffset: 0
            };
        }
    }
    loadBuffers() {
        //load up to 5 buffers parallel
        const tLoadTasks = new Array();
        if (this.mIsVersion1) {
            for (const tBuffer in this.mGltf.buffers) {
                tLoadTasks.push(() => this.loadBuffer(tBuffer));
            }
        }
        else {
            for (let i = 0; i < this.mGltf.buffers.length; ++i) {
                tLoadTasks.push(() => this.loadBuffer(i));
            }
        }
        return (forEachParallel(tLoadTasks, (pIndex, pTask) => pTask(), 5).then(() => { }));
    }
    loadBuffer(pBufferId) {
        const tBuffer = this.mGltf.buffers[pBufferId];
        if (tBuffer == null)
            return (Promise.reject(new Error("Buffer " + pBufferId + " not found")));
        if (isDataUri(tBuffer.uri)) {
            const tArrayBuffer = fromDataUri(tBuffer.uri);
            this.onBufferLoaded(pBufferId, tArrayBuffer);
            return (Promise.resolve());
        }
        else {
            return (this.loadBufferFromUrl(pBufferId, this.mRootUrl + tBuffer.uri));
        }
    }
    loadBufferFromUrl(pBufferId, pUrl) {
        if (isNullOrEmpty(pUrl))
            return (Promise.resolve());
        const tPromise = GltfMeshImport.LoadArrayBuffer(pUrl).then((pBuffer) => {
            this.onBufferLoaded(pBufferId, pBuffer);
        });
        return (tPromise);
    }
    importMesh(pGltfMesh) {
        if (pGltfMesh == null)
            return (null);
        const tVertexData = new VertexData();
        // Positions, normals and UVs
        for (let i = 0; i < pGltfMesh.primitives.length; i++) {
            const tCurrentVertexData = i === 0 ? tVertexData : new VertexData();
            const tPrimitive = pGltfMesh.primitives[i];
            if (tPrimitive.mode != null && tPrimitive.mode !== 4) {
                continue;
            }
            const tAttributes = tPrimitive.attributes;
            let tAccessor = null;
            let tBuffer = null;
            // Set positions, normal and uvs
            for (const tSemantic in tAttributes) {
                // Link accessor and buffer view
                tAccessor = this.mGltf.accessors[tAttributes[tSemantic]];
                tBuffer = this.getBufferFromAccessor(tAccessor);
                if (tSemantic === "NORMAL") {
                    tCurrentVertexData.normals = this.bufferToFloat32Array(tBuffer, this.mConvertToLeftHanded);
                }
                else if (tSemantic === "POSITION") {
                    tCurrentVertexData.positions = this.bufferToFloat32Array(tBuffer, this.mConvertToLeftHanded);
                }
                else if (tSemantic.indexOf("TEXCOORD_") !== -1) {
                    const tChannel = Number(tSemantic.split("_")[1]);
                    const tUvKind = VertexBuffer.UVKind + (tChannel === 0 ? "" : (tChannel + 1));
                    tCurrentVertexData.set(this.bufferToFloat32Array(tBuffer, false), tUvKind);
                }
            }
            // Indices
            if (tPrimitive.indices != null) {
                tAccessor = this.mGltf.accessors[tPrimitive.indices];
                tBuffer = this.getBufferFromAccessor(tAccessor);
                if (tBuffer instanceof Uint16Array || tBuffer instanceof Uint32Array)
                    tCurrentVertexData.indices = tBuffer;
                else
                    tCurrentVertexData.indices = new Uint32Array(tBuffer);
            }
            else {
                //generate indices
                if (tCurrentVertexData.positions != null) {
                    const tLength = tCurrentVertexData.positions.length / 3;
                    tCurrentVertexData.indices = new Uint32Array(tLength);
                    for (let i = 0; i < tLength; ++i)
                        tCurrentVertexData.indices[i] = i;
                }
            }
            // merge primitives
            if (i > 0) {
                tVertexData.merge(tCurrentVertexData);
            }
        }
        //finally create mesh and geometry
        const tMesh = new Mesh("", this.mScene);
        new Geometry(Geometry.RandomId(), this.mScene, tVertexData, false, tMesh);
        return (tMesh);
    }
    bufferToFloat32Array(pBuffer, pNegateZ) {
        const tArray = (pBuffer instanceof Float32Array) ? pBuffer :
            new Float32Array(pBuffer);
        if (pNegateZ) { //right to left handed
            for (let j = 0; j < tArray.length; j += 3) {
                //negate z
                tArray[j + 2] = -tArray[j + 2];
            }
        }
        return (tArray);
    }
    getByteStrideFromType(pAccessor) {
        if (pAccessor.byteStride != null)
            return (pAccessor.byteStride);
        switch (pAccessor.type) {
            case "VEC2":
                return (2);
            case "VEC3":
                return (3);
            default:
                return (1);
        }
    }
    getBufferFromAccessor(pAccessor) {
        if (this.mArrayBuffers == null)
            throw new Error("undefined mArrayBuffers");
        if (pAccessor.bufferView == null)
            throw new Error("undefined bufferView");
        const tBufferView = this.mGltf.bufferViews[pAccessor.bufferView];
        const tArrayBuffer = this.mArrayBuffers[tBufferView.buffer];
        const tAccessorOffset = pAccessor.byteOffset != null ? pAccessor.byteOffset : 0;
        const tBufferViewOffset = tBufferView.byteOffset != null ? tBufferView.byteOffset : 0;
        const tByteOffset = this.mBinaryGltfOffset + tAccessorOffset + tBufferViewOffset;
        const tCount = pAccessor.count * this.getByteStrideFromType(pAccessor);
        switch (pAccessor.componentType) {
            case GltfTypes.ComponentType.BYTE:
                return (new Int8Array(tArrayBuffer, tByteOffset, tCount));
            case GltfTypes.ComponentType.UNSIGNED_BYTE:
                return (new Uint8Array(tArrayBuffer, tByteOffset, tCount));
            case GltfTypes.ComponentType.SHORT:
                return (new Int16Array(tArrayBuffer, tByteOffset, tCount));
            case GltfTypes.ComponentType.UNSIGNED_SHORT:
                return (new Uint16Array(tArrayBuffer, tByteOffset, tCount));
            case GltfTypes.ComponentType.UNSIGNED_INT:
                return (new Uint32Array(tArrayBuffer, tByteOffset, tCount));
            default:
                return (new Float32Array(tArrayBuffer, tByteOffset, tCount));
        }
    }
}
