import { Color3 } from "@babylonjs/core/Maths/math.color";
import type { PropertyProvider } from "./PropertyProvider";
import { PropertyChangedResult } from "./PropertyProvider";
/**
 * Key of the default property class which will be used for properties without a defined property class.
 */
export declare const DEFAULT_PROPERTYCLASS_KEY = "WCF_PROPCLASS_DEFAULT";
export type PropertyValueType = string | boolean | number | null;
export declare const PropertyType: {
    readonly String: 0;
    readonly Bool: 1;
    readonly Number: 2;
    readonly Length: 3;
    readonly Color3: 4;
    readonly Angle: 5;
};
export type PropertyType = typeof PropertyType[keyof typeof PropertyType];
export interface PropertyValueSurcharge {
    value: number;
    currency: string;
}
export declare class PropertyValue {
    value: PropertyValueType;
    text: string | null;
    smallIcon: string | null;
    largeIcon: string | null;
    image: string | null;
    /**
     * Defines if this value is selectable by the user or not.
     * This is only applicable for choice values not for the current value of a property.
     */
    selectable: boolean;
    /** optional surcharge for this property value */
    surcharge?: PropertyValueSurcharge | null;
    constructor(pValue?: PropertyValueType, pText?: string | null);
    clone(): PropertyValue;
}
export declare class PropertyClass {
    key: string;
    name: string | null;
    constructor(pKey?: string | null, pName?: string | null);
    /**
     * Returns the class name if it is not empty, otherwise the key is returned.
     */
    getName(): string;
}
export declare abstract class Property {
    get provider(): PropertyProvider;
    key: string;
    name: string | null;
    class: string | null;
    editable: boolean;
    visible: boolean;
    undoable: boolean;
    choiceList: boolean;
    allowCustomChoice: boolean;
    userData: unknown;
    abstract readonly type: PropertyType;
    constructor(pProvider: PropertyProvider, pKey: string, pName?: string | null);
    /**
     * Callback which returns the property value.
     */
    getValue: () => PropertyValue | null;
    /**
     * Callback which sets the property value.
     */
    setValue: (pValue: PropertyValueType) => Promise<PropertyChangedResult>;
    /**
     * Callback which returns the available value choices.
     */
    getChoices: () => Promise<Array<PropertyValue> | null>;
    /**
     * Returns the property name if it is not empty, otherwise the key is returned.
     */
    getName(): string;
    /**
     * Returns pOtherValue as a string. If pOther = undefined, the current value of this
     * property will be returned as a string.
     */
    getValueAsString(pOther?: PropertyValue): string | null;
    isValueEqual(pOther: PropertyValue | null): boolean;
    abstract clone(): Property;
    protected mProvider: PropertyProvider;
    protected copyFrom(pOther: Property): void;
}
/**
 * StringProperty (PropertyValueType = string)
 */
export declare class StringProperty extends Property {
    maxLength: number | null;
    multiLine: boolean;
    readonly type: PropertyType;
    constructor(pProvider: PropertyProvider, pKey: string, pName: string | null);
    clone(): Property;
    static ConvertValueToString(pValue: PropertyValueType): string | null;
    protected copyFrom(pOther: Property): void;
}
/**
 * BoolProperty (PropertyValueType = boolean)
 */
export declare class BoolProperty extends Property {
    readonly type: PropertyType;
    clone(): Property;
    static ConvertValueToBool(pValue: PropertyValueType): boolean | null;
}
/**
 * NumberProperty (PropertyValueType = number)
 */
export declare class NumberProperty extends Property {
    minValue: number | null;
    maxValue: number | null;
    numDigits: number | null;
    numDecDigits: number;
    readonly type: PropertyType;
    constructor(pProvider: PropertyProvider, pKey: string, pName: string | null);
    isValueEqual(pOther: PropertyValue): boolean;
    clone(): Property;
    static ConvertValueToNumber(pValue: PropertyValueType): number | null;
    protected copyFrom(pOther: Property): void;
}
/**
 * LengthProperty (PropertyValueType = number)
 * The value represents a length in meter.
 */
export declare class LengthProperty extends NumberProperty {
    readonly type: PropertyType;
    constructor(pProvider: PropertyProvider, pKey: string, pName: string | null);
    clone(): Property;
}
/**
 * AngleProperty (PropertyValueType = number)
 * The value represents an angle in degree.
 */
export declare class AngleProperty extends NumberProperty {
    readonly type: PropertyType;
    constructor(pProvider: PropertyProvider, pKey: string, pName: string | null);
    clone(): Property;
}
/**
 * Color3Property (PropertyValueType = string)
 * The color is represented as an RGB hex string, e.g. #FF0000
 */
export declare class Color3Property extends Property {
    readonly type: PropertyType;
    clone(): Property;
    /**
     * Converts pValue into a hex color string (if it is not already a string).
     */
    static ConvertValueToString(pValue: PropertyValueType): string | null;
    /**
     * Converts pValue into a Color3
     */
    static ConvertValueToColor3(pValue: PropertyValueType): Color3 | null;
}
