import { Camera as BJSCamera } from "@babylonjs/core/Cameras/camera";
import { TargetCamera } from "@babylonjs/core/Cameras/targetCamera";
import { Engine } from "@babylonjs/core/Engines/engine";
import { Effect } from "@babylonjs/core/Materials/effect";
import { Material } from "@babylonjs/core/Materials/material";
import { ShaderMaterial } from "@babylonjs/core/Materials/shaderMaterial";
import { PBRMaterial } from "@babylonjs/core/Materials/PBR/pbrMaterial";
import { RenderTargetTexture } from "@babylonjs/core/Materials/Textures/renderTargetTexture";
import { Texture as BJSTexture } from "@babylonjs/core/Materials/Textures/texture";
import { Color3, Color4 } from "@babylonjs/core/Maths/math.color";
import { Vector3 } from "@babylonjs/core/Maths/math.vector";
import { CreatePlane } from "@babylonjs/core/Meshes/Builders/planeBuilder";
import { InstancedMesh } from "@babylonjs/core/Meshes/instancedMesh";
import { LinesMesh } from "@babylonjs/core/Meshes/linesMesh";
import { Mesh } from "@babylonjs/core/Meshes/mesh";
import { PostProcess } from "@babylonjs/core/PostProcesses/postProcess";
import { eq, PI_DIV2 } from "../../utils/math";
import { BoundingBox } from "../mdl/BoundingBox";
import { removeFromScene, updateFrozenWorldMatrix } from "../utils/MeshUtils";
import { CameraControl, CameraModeMask } from "../view/CameraControl";
import { ShadowGenerator } from "./ShadowGenerator";
export class ShadowPlane {
    get planeYOffset() {
        return (this.mPlane.position.y - this.mPosY);
    }
    set planeYOffset(pOffset) {
        this.mPlane.position.y = this.mPosY + pOffset;
        updateFrozenWorldMatrix(this.mPlane);
    }
    get planeYPosition() {
        return (this.mPosY);
    }
    get transparency() {
        return (this.mTransparency);
    }
    get enabled() {
        return (this.mEnabled);
    }
    constructor(pApp, pResolution = 1024) {
        this.onModelModified = () => {
            this.mShadowsNeedUpdate = true;
        };
        this.onWebGLContextRestored = () => {
            this.mShadowsNeedUpdate = true;
        };
        this.onViewerBeforeRender = () => {
            this.updateShadows();
        };
        this.mApp = pApp;
        this.mScene = pApp.scene;
        this.mShadowsNeedUpdate = false;
        this.mEnabled = true;
        this.mShadowUpdateTimeout = null;
        const tCamViewingRange = 0.25;
        this.mMinPlaneSize = (pResolution / 4096) * 10; //(512 = 1.25, 1024 = 2.5, 2048 = 5.0, 4096 = 10.0)
        this.mPlaneSize = this.mMinPlaneSize;
        this.mPosY = 0;
        this.mTransparency = 0.7;
        this.createCamera(tCamViewingRange);
        this.createRenderTarget(pResolution);
        this.createPlane();
        //register events
        this.mApp.model.eventModified.addListener(this.onModelModified);
        this.mApp.viewer.eventBeforeRender.addListener(this.onViewerBeforeRender);
        this.mApp.engine.onContextRestoredObservable.add(this.onWebGLContextRestored);
    }
    dispose() {
        this.mPlane.dispose();
        this.mPlaneMaterial.dispose();
        this.mCamera.dispose();
        this.mRenderTarget.dispose();
        this.mDepthMaterial.dispose();
        this.mDummyMesh.dispose();
        this.mApp.model.eventModified.removeListener(this.onModelModified);
        this.mApp.viewer.eventBeforeRender.removeListener(this.onViewerBeforeRender);
        this.mShadowsNeedUpdate = false;
    }
    setEnabled(pEnabled) {
        if (this.mEnabled === pEnabled)
            return;
        this.mPlane.isVisible = pEnabled;
        this.mApp.viewer.requestRenderFrame();
    }
    setPlaneYPosition(pPos) {
        if (this.mPosY === pPos)
            return;
        const tOffset = this.planeYOffset;
        this.mPosY = pPos;
        this.mCamera.position.y = this.mPosY + ShadowPlane.CAMERA_Y_OFFSET;
        this.mPlane.position.y = this.mPosY + tOffset;
        updateFrozenWorldMatrix(this.mPlane);
        this.mShadowsNeedUpdate = true;
        this.mApp.viewer.requestRenderFrame();
    }
    setTransparency(pTransparency) {
        if (this.mTransparency === pTransparency)
            return;
        this.mTransparency = pTransparency;
        this.mShadowsNeedUpdate = true;
        this.mApp.viewer.requestRenderFrame();
    }
    notifyModelInteraction() {
        if (!this.mEnabled)
            return;
        this.mPlane.isVisible = false;
        this.clearUpdateTimeout();
        this.mShadowUpdateTimeout = window.setTimeout(() => {
            this.mShadowUpdateTimeout = null;
            this.mShadowsNeedUpdate = true;
            this.mApp.viewer.requestRenderFrame();
        }, 250);
    }
    prepareRenderFrame(pMultiSamplingStep, pMaxSteps) {
    }
    finalizeRenderFrame(pMultiSamplingStep, pMaxSteps) {
    }
    clearUpdateTimeout() {
        if (this.mShadowUpdateTimeout != null) {
            clearTimeout(this.mShadowUpdateTimeout);
            this.mShadowUpdateTimeout = null;
        }
    }
    addMeshToRenderList(pNode, pRenderList) {
        if (pNode == null || pNode instanceof LinesMesh)
            return;
        if (pNode instanceof Mesh && pNode.geometry != null || pNode instanceof InstancedMesh) {
            if (ShadowGenerator.isMeshShadowDisabled(pNode))
                return;
            const tBInfo = pNode.getBoundingInfo();
            //ignore objects which are positioned at y = this.mPosY and have zero height (eg. planes)
            if (tBInfo != null && (!eq(tBInfo.boundingBox.minimumWorld.y, this.mPosY) ||
                !eq(tBInfo.boundingBox.maximumWorld.y, this.mPosY))) {
                pRenderList.push(pNode);
            }
        }
    }
    updateShadows() {
        if (!this.mEnabled || !this.mShadowsNeedUpdate)
            return;
        this.clearUpdateTimeout();
        this.mShadowsNeedUpdate = false;
        this.mPlane.isVisible = true; //maybe was disabled in notifyModelInteraction()
        //update renderlist
        const tRenderList = new Array();
        tRenderList.push(this.mDummyMesh);
        for (const tElement of this.mApp.model.elements) {
            if (tElement.node != null) {
                this.addMeshToRenderList(tElement.node, tRenderList);
                const tChilds = tElement.node.getChildTransformNodes(false);
                for (const tMesh of tChilds)
                    this.addMeshToRenderList(tMesh, tRenderList);
            }
        }
        this.mRenderTarget.renderList = tRenderList;
        this.mRenderTarget.setMaterialForRendering(tRenderList, this.mDepthMaterial);
        this.updatePlaneSize();
        //render shadow texture
        this.mScene.incrementRenderId();
        this.mScene.resetCachedMaterial();
        this.mScene.updateTransformMatrix(true);
        this.mRenderTarget.render(true);
    }
    createCamera(pViewingRange) {
        this.mCamera = new TargetCamera("shadowPlaneCamera", new Vector3(0, this.mPosY + ShadowPlane.CAMERA_Y_OFFSET, 0), this.mScene);
        this.mCamera.rotation.x = -PI_DIV2;
        this.mCamera.upVector = new Vector3(0, 0, -1);
        this.mCamera.mode = BJSCamera.ORTHOGRAPHIC_CAMERA;
        this.mCamera.minZ = 0;
        this.mCamera.maxZ = pViewingRange;
    }
    createRenderTarget(pTextureSize) {
        const tTexelSize = 1.0 / pTextureSize;
        this.mRenderTarget = new RenderTargetTexture("shadowPlane", pTextureSize, this.mScene, true, true, Engine.TEXTURETYPE_UNSIGNED_INT, false, BJSTexture.TRILINEAR_SAMPLINGMODE, false, false);
        this.mRenderTarget.clearColor = new Color4(1, 1, 1, 1);
        this.mRenderTarget.anisotropicFilteringLevel = this.mApp.engine.getCaps().maxAnisotropy;
        this.mRenderTarget.hasAlpha = true;
        this.mRenderTarget.noPrePassRenderer = true;
        this.mRenderTarget.activeCamera = this.mCamera;
        let tPostProcess = new PostProcess("shadowPlaneDof", "wcf.core.ShadowPlaneDof", ["dofAmount", "transparency"], [], { width: pTextureSize, height: pTextureSize }, this.mCamera);
        tPostProcess.onApply = (pEffect) => {
            //reduce dof amount with increasing plane size
            const tRatio = this.mPlaneSize / pTextureSize;
            const tAmount = 12.0 * (ShadowPlane.REFERENCE_RATIO / tRatio);
            pEffect.setFloat("dofAmount", tAmount * tTexelSize);
            pEffect.setFloat("transparency", this.mTransparency);
        };
        tPostProcess = new PostProcess("shadowPlaneBlur", "wcf.core.ShadowPlaneBlur", ["blurAmount"], [], { width: pTextureSize, height: pTextureSize }, this.mCamera);
        tPostProcess.onApply = (pEffect) => {
            pEffect.setFloat("blurAmount", 0.75 * tTexelSize);
        };
        //dummy mesh is used to ensure that the render target will update
        //if a render target has an empty renderList it will not render
        this.mDummyMesh = new Mesh("shadowPlaneDummy", this.mScene);
        this.mDummyMesh.freezeWorldMatrix();
        this.mDummyMesh.isPickable = false;
        this.mDummyMesh.isVisible = false;
        this.mDummyMesh.setEnabled(false);
        removeFromScene(this.mDummyMesh, this.mScene);
        //create depth material
        this.mDepthMaterial = new ShaderMaterial("shadowPlaneDepth", this.mScene, "wcf.core.ShadowPlaneDepth", {
            attributes: ["position", "world0", "world1", "world2", "world3"],
            uniforms: ["viewProjection", "world"]
        });
        this.mDepthMaterial.backFaceCulling = false;
        this.mDepthMaterial.freeze(); //do not freeze the material if instances are used, to allow switching between instances and non-instances
    }
    createPlane() {
        this.mPlaneMaterial = new PBRMaterial("shadowPlane", this.mScene);
        this.mPlaneMaterial.albedoColor = Color3.White();
        this.mPlaneMaterial.sideOrientation = Material.ClockWiseSideOrientation;
        this.mPlaneMaterial.albedoTexture = this.mRenderTarget;
        this.mPlaneMaterial.useAlphaFromAlbedoTexture = true;
        this.mPlaneMaterial.disableLighting = true;
        this.mPlaneMaterial.unlit = true;
        this.mPlaneMaterial.zOffset = -0.01;
        this.mPlaneMaterial.freeze();
        this.mPlane = CreatePlane("shadowPlane", { width: 1.0, height: 1.0 }, this.mScene);
        this.mPlane.position.y = this.mPosY + 0.0005; //add additional 0.5mm offset to prevent z-fighting
        this.mPlane.rotation.x = -PI_DIV2;
        this.mPlane.isPickable = false;
        this.mPlane.material = this.mPlaneMaterial;
        this.mPlane.freezeWorldMatrix();
        //only visiblie in ortho and perspective views
        CameraControl.SetMeshCameraModeMask(this.mPlane, CameraModeMask.Perspective | CameraModeMask.Orthographic);
    }
    updatePlaneSize() {
        const tTotalBounds = new BoundingBox();
        for (const tMesh of this.mRenderTarget.renderList) {
            if (tMesh === this.mDummyMesh)
                continue;
            const tBInfo = tMesh.getBoundingInfo();
            if (tBInfo != null) {
                tTotalBounds.expandByPointInPlace(tBInfo.boundingBox.minimumWorld);
                tTotalBounds.expandByPointInPlace(tBInfo.boundingBox.maximumWorld);
            }
        }
        if (tTotalBounds.isValid()) {
            const tCenter = tTotalBounds.getCenter();
            this.mPlaneSize = Math.max(tTotalBounds.max.x - tTotalBounds.min.x, tTotalBounds.max.z - tTotalBounds.min.z);
            if (this.mPlaneSize < this.mMinPlaneSize)
                this.mPlaneSize = this.mMinPlaneSize;
            this.mPlaneSize += 0.1; //add offset for blurring
            this.mPlane.scaling.x = this.mPlaneSize;
            this.mPlane.scaling.y = this.mPlaneSize;
            this.mPlane.position.x = tCenter.x;
            this.mPlane.position.z = tCenter.z;
            const tOrtho = this.mPlaneSize * 0.5;
            this.mCamera.orthoTop = tOrtho;
            this.mCamera.orthoBottom = -tOrtho;
            this.mCamera.orthoLeft = -tOrtho;
            this.mCamera.orthoRight = tOrtho;
            this.mCamera.position.x = tCenter.x;
            this.mCamera.position.z = tCenter.z;
            updateFrozenWorldMatrix(this.mPlane);
        }
    }
}
//// private ///////////////////////////////////////////////////////////////////////////////
ShadowPlane.REFERENCE_RATIO = 10.0 / 1024.0;
ShadowPlane.CAMERA_Y_OFFSET = -0.0005;
//register shaders
Effect.ShadersStore["wcf.core.ShadowPlaneDepthVertexShader"] = `
precision highp float;

#include<instancesDeclaration>
attribute vec3 position;
uniform mat4 viewProjection;

void main(void)
{
    #include<instancesVertex>
    gl_Position = viewProjection * finalWorld * vec4(position, 1.0);
}`;
Effect.ShadersStore["wcf.core.ShadowPlaneDepthFragmentShader"] = `
precision highp float;

void main(void)
{
    float depth = gl_FragCoord.z;
    gl_FragColor = vec4(depth, depth, depth, 1.0);
}`;
Effect.ShadersStore["wcf.core.ShadowPlaneBlurFragmentShader"] = `
precision highp float;

varying vec2 vUV;
uniform sampler2D textureSampler;
uniform float blurAmount;

void main(void)
{
    float alpha = 0.0;
    alpha += texture2D(textureSampler, vUV + vec2(-1.0, -1.0) * blurAmount).a;
    alpha += texture2D(textureSampler, vUV + vec2( 0.0, -1.0) * blurAmount).a;
    alpha += texture2D(textureSampler, vUV + vec2( 1.0, -1.0) * blurAmount).a;

    alpha += texture2D(textureSampler, vUV + vec2(-1.0,  0.0) * blurAmount).a;
    alpha += texture2D(textureSampler, vUV + vec2( 0.0,  0.0) * blurAmount).a;
    alpha += texture2D(textureSampler, vUV + vec2( 1.0,  0.0) * blurAmount).a;

    alpha += texture2D(textureSampler, vUV + vec2(-1.0,  1.0) * blurAmount).a;
    alpha += texture2D(textureSampler, vUV + vec2( 0.0,  1.0) * blurAmount).a;
    alpha += texture2D(textureSampler, vUV + vec2( 1.0,  1.0) * blurAmount).a;

    alpha = alpha / 9.0;
    gl_FragColor = vec4(0.0, 0.0, 0.0, alpha);
}`;
//this shader is based on https://github.com/mrdoob/three.js/blob/dev/examples/js/shaders/BokehShader.js
Effect.ShadersStore["wcf.core.ShadowPlaneDofFragmentShader"] = `
precision highp float;

varying vec2 vUV;
uniform sampler2D textureSampler;
uniform float dofAmount;
uniform float transparency;

void main(void)
{
    float depth = texture2D( textureSampler, vUV ).x;
    float dofblur = clamp( depth * dofAmount, 0.0, 1.0 );

    float dofblur9 = dofblur * 0.7; //0.9
    float dofblur7 = dofblur * 0.5; //0.7
    float dofblur4 = dofblur * 0.25; //0.4

    float col = 0.0;

    col += depth;
    col += texture2D( textureSampler, vUV + vec2(  0.0,   0.4  ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.15,  0.37 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.29,  0.29 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.37,  0.15 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.40,  0.0  ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.37, -0.15 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.29, -0.29 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.15, -0.37 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.0,  -0.4  ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.15,  0.37 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.29,  0.29 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.37,  0.15 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.4,   0.0  ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.37, -0.15 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.29, -0.29 ) * dofblur ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.15, -0.37 ) * dofblur ).x;

    col += texture2D( textureSampler, vUV + vec2(  0.15,  0.37 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.37,  0.15 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.37, -0.15 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.15, -0.37 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.15,  0.37 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.37,  0.15 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.37, -0.15 ) * dofblur9 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.15, -0.37 ) * dofblur9 ).x;

    col += texture2D( textureSampler, vUV + vec2(  0.29,  0.29 ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.40,  0.0  ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.29, -0.29 ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.0,  -0.4  ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.29,  0.29 ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.4,   0.0  ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.29, -0.29 ) * dofblur7 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.0,   0.4  ) * dofblur7 ).x;

    col += texture2D( textureSampler, vUV + vec2(  0.29,  0.29 ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.4,   0.0  ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.29, -0.29 ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.0,  -0.4  ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.29,  0.29 ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.4,   0.0  ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2( -0.29, -0.29 ) * dofblur4 ).x;
    col += texture2D( textureSampler, vUV + vec2(  0.0,   0.4  ) * dofblur4 ).x;

    col = col / 41.0;

    float tra = col + (1.0 - col) * transparency;
    gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0 - tra);
}`;
