import { Color3 } from "@babylonjs/core/Maths/math.color";
import { Vector2, Vector3 } from "@babylonjs/core/Maths/math.vector";
import { VertexBuffer } from "@babylonjs/core/Buffers/buffer";
import { InstancedMesh } from "@babylonjs/core/Meshes/instancedMesh";
import { InstancedLinesMesh, LinesMesh } from "@babylonjs/core/Meshes/linesMesh";
import { Mesh } from "@babylonjs/core/Meshes/mesh";
import { isPointInFrustum, isTriangleInFrustum, Rectangle, segmentRectangleIntersecting, triangleRectangleIntersecting } from "../utils/MathUtils";
import { CoordinatesType, View } from "../view/View";
import { AbstractTool } from "./AbstractTool";
import { InteractionMode } from "./Tool";
export class RectangleSelect extends AbstractTool {
    constructor(pApp) {
        super(pApp);
        this.multiSelectMode = false;
        this.mode = "Complete";
        this.mPartialMode = false;
        this.mRectDiv = null;
        this.mView = this.mApp.viewer.view;
    }
    start(pParameter) {
        return (super.start(pParameter));
    }
    stop() {
        if (this.mRectDiv != null) {
            this.mView.overlayDiv.removeChild(this.mRectDiv);
            this.mRectDiv = null;
        }
        super.stop();
    }
    handlePointerDown(pEvent, pState) {
        if (!pState.isPrimaryPointer(pEvent))
            return;
        this.startSelection();
    }
    handlePointerMove(pEvent, pState) {
        if (!pState.isPrimaryPointer(pEvent) || this.mRectDiv == null)
            return;
        const tPointer = pState.pointers.get(pEvent.pointerId);
        if (tPointer == null)
            return;
        const tP1 = tPointer.initialPosition;
        const tP2 = tPointer.position;
        const tX = Math.min(tP1.x, tP2.x);
        const tY = Math.min(tP1.y, tP2.y);
        const tWidth = Math.abs(tP1.x - tP2.x);
        const tHeight = Math.abs(tP1.y - tP2.y);
        const tPos = new Vector2(tX, tY);
        this.mView.convertCoordsToRef(tPos, CoordinatesType.Client, CoordinatesType.Viewer, tPos);
        this.mRectDiv.style.left = tPos.x.toString() + "px";
        this.mRectDiv.style.top = tPos.y.toString() + "px";
        this.mRectDiv.style.width = tWidth.toString() + "px";
        this.mRectDiv.style.height = tHeight.toString() + "px";
        if (this.mode === "Both") {
            this.setPartialMode(tP2.x < tP1.x, false);
        }
    }
    handlePointerUp(pEvent, pState) {
        if (!pState.isPrimaryPointer(pEvent))
            return;
        const tPointer = pState.pointers.get(pEvent.pointerId);
        if (tPointer != null && this.mRectDiv != null) {
            const tCenter = Vector2.Center(tPointer.initialPosition, tPointer.position);
            const tSizeX = Math.abs(tPointer.initialPosition.x - tPointer.position.x);
            const tSizeY = Math.abs(tPointer.initialPosition.y - tPointer.position.y);
            if (tSizeX > 1 && tSizeY > 1) {
                const tFrustum = this.mView.createPickingFrustum(tCenter, tSizeX, tSizeY, CoordinatesType.Client);
                const tRect = new Rectangle(tPointer.initialPosition, tPointer.position);
                this.select(tFrustum, tRect, this.multiSelectMode || pEvent.ctrlKey);
            }
        }
        this.stop();
    }
    //// internal //////////////////////////////////////////////////////////////////////////////
    /** @internal */
    startSelection() {
        this.mRectDiv = document.createElement("div");
        this.mRectDiv.style.position = "absolute";
        this.mRectDiv.style.boxSizing = "border-box";
        this.mView.overlayDiv.appendChild(this.mRectDiv);
        this.setPartialMode(this.mode === "Partial", true);
        this.setInteractionMode(InteractionMode.Primary);
    }
    setPartialMode(pMode, pForce) {
        if (pMode === this.mPartialMode && !pForce)
            return;
        this.mPartialMode = pMode;
        this.applyRectDivColor(this.mPartialMode);
    }
    applyRectDivColor(pPartialMode) {
        if (this.mRectDiv != null) {
            const tAlpha = pPartialMode ? 0.2 : 0.3;
            const tBorderMode = pPartialMode ? "dashed" : "solid";
            const tCol = Color3.FromHexString(this.mApp.theme.mainColorBright).scale(256);
            const tBGCol = `rgba(${Math.round(tCol.r)}, ${Math.round(tCol.g)}, ${Math.round(tCol.b)}, ${tAlpha})`;
            this.mRectDiv.style.backgroundColor = tBGCol;
            this.mRectDiv.style.border = "1px " + tBorderMode + " " + this.mApp.theme.mainColorBright;
        }
    }
    select(pFrustum, pRect, pMultiSelect) {
        const tNewSelection = new Array();
        for (const tElement of this.mApp.model.elements) {
            if (this.maySelectElement(pFrustum, pRect, tElement)) {
                tNewSelection.push(tElement);
            }
            else {
                if (tElement.isSelected && pMultiSelect)
                    tNewSelection.push(tElement);
            }
        }
        this.mApp.model.setSelection(tNewSelection);
    }
    maySelectElement(pFrustum, pRect, pElement) {
        if (pFrustum == null || !pElement.canSelect)
            return (false);
        //check bounds first
        if (!pElement.boundingBoxVisible.isInFrustum(pFrustum, pElement.transform))
            return (false);
        //check all meshes
        const tNodes = pElement.getNodes(true);
        if (tNodes == null || tNodes.length === 0)
            return (false);
        if (this.mPartialMode) {
            for (const tNode of tNodes) {
                if (this.isNodePartiallyInFrustum(pFrustum, pRect, tNode)) {
                    return (true);
                }
            }
            return (false);
        }
        else {
            let tIgnoredNodes = 0;
            for (const tNode of tNodes) {
                const tResult = this.isNodeCompletelyInFrustum(pFrustum, tNode);
                if (tResult === 0) //found a node which is not in frustum -> abort
                    return (false);
                if (tResult < 0)
                    tIgnoredNodes++;
            }
            return (tIgnoredNodes < tNodes.length);
        }
    }
    isNodeCompletelyInFrustum(pFrustum, pNode) {
        if (pNode == null || !pNode.isEnabled())
            return (-1); //ignored
        const tVertex = Vector3.Zero();
        if ((pNode instanceof Mesh && pNode.geometry != null) || pNode instanceof InstancedMesh) {
            if (!pNode.isVisible || !pNode.isPickable)
                return (-1); //ignored
            //check bounds first
            if (pNode.isInFrustum(pFrustum)) {
                //check all vertices
                const tPositions = pNode.getVerticesData(VertexBuffer.PositionKind);
                if (tPositions == null)
                    return (-1); //ignored
                //get local frustum
                const tLocalFrustum = View._GetLocalFrustum(pFrustum, pNode);
                //check if all vertices are inside the frustum
                for (let i = 0; i < tPositions.length; i += 3) {
                    tVertex.set(tPositions[i], tPositions[i + 1], tPositions[i + 2]);
                    if (!isPointInFrustum(tVertex, tLocalFrustum)) {
                        return (0); //not in frustum
                    }
                }
            }
            else {
                return (0); //not in frustum
            }
            return (1); //in frustum
        }
        return (-1); //ignored
    }
    isNodePartiallyInFrustum(pFrustum, pRect, pNode) {
        if (pNode == null || !pNode.isEnabled())
            return (false);
        if ((pNode instanceof Mesh && pNode.geometry != null) || pNode instanceof InstancedMesh) {
            if (!pNode.isVisible || !pNode.isPickable)
                return (false);
            //check bounds first
            if (pNode.isInFrustum(pFrustum)) {
                //get vertex positions
                if (!pNode._generatePointsArray())
                    return (false);
                const tPositions = pNode._positions;
                if (tPositions == null)
                    return (false);
                //get indices
                const tIndices = pNode.getIndices();
                if (tIndices == null)
                    return (false);
                //get local frustum
                const tLocalFrustum = View._GetLocalFrustum(pFrustum, pNode);
                if (pNode instanceof LinesMesh || pNode instanceof InstancedLinesMesh) {
                    //check if a line segment is partially in the frustum
                    for (let i = 0; i < tIndices.length; i += 2) {
                        const tP1 = tPositions[tIndices[i]];
                        const tP2 = tPositions[tIndices[i + 1]];
                        //check if segment intersects the selection rectangle
                        if (this.segmentRectIntersects(pRect, pNode, tP1, tP2))
                            return (true);
                    }
                }
                else {
                    //check if a triangle is partially in the frustum
                    for (let i = 0; i < tIndices.length; i += 3) {
                        const tP1 = tPositions[tIndices[i]];
                        const tP2 = tPositions[tIndices[i + 1]];
                        const tP3 = tPositions[tIndices[i + 2]];
                        //check if triangle is in frustum (fast check only)
                        if (!isTriangleInFrustum(tP1, tP2, tP3, tLocalFrustum))
                            continue;
                        //check if triangle intersects the selection rectangle
                        if (this.triangleRectIntersects(pRect, pNode, tP1, tP2, tP3))
                            return (true);
                    }
                }
            }
            else
                return (false);
        }
        return (false);
    }
    triangleRectIntersects(pRect, pNode, pP1, pP2, pP3) {
        //project triangle to client coordinates
        const tMW = pNode.getWorldMatrix();
        const tP1 = this.mView.projectPoint2(pP1, CoordinatesType.Client, tMW, true);
        const tP2 = this.mView.projectPoint2(pP2, CoordinatesType.Client, tMW, true);
        const tP3 = this.mView.projectPoint2(pP3, CoordinatesType.Client, tMW, true);
        if (tP1 == null || tP2 == null || tP3 == null)
            return (false);
        return (triangleRectangleIntersecting(tP1, tP2, tP3, pRect));
    }
    segmentRectIntersects(pRect, pNode, pP1, pP2) {
        //project triangle to client coordinates
        const tMW = pNode.getWorldMatrix();
        const tP1 = this.mView.projectPoint2(pP1, CoordinatesType.Client, tMW, true);
        const tP2 = this.mView.projectPoint2(pP2, CoordinatesType.Client, tMW, true);
        if (tP1 == null || tP2 == null)
            return (false);
        return (segmentRectangleIntersecting(tP1, tP2, pRect));
    }
}
