import { Color4 } from "@babylonjs/core/Maths/math.color";
import { Matrix, Vector2, Vector3 } from "@babylonjs/core/Maths/math.vector";
import { ListenableEvent } from "../../utils";
import { clone } from "../../utils/array";
import { EPSILON, eq } from "../../utils/math";
import { BoundingBox } from "../mdl/BoundingBox";
import { SceneElement } from "../mdl/SceneElement";
import { removeFromScene, setLocalMatrix } from "../utils/MeshUtils";
import { ThickLinesMesh } from "../utils/ThickLinesMesh";
import { Camera } from "../view/Camera";
import { CameraMode } from "../view/CameraControl";
import { PointerInfo } from "../view/InputState";
import { CoordinatesType, ViewLayerMaskFlags } from "../view/View";
import { Viewer } from "../view/Viewer";
import { AbstractTool } from "./AbstractTool";
import { RectangleSelect } from "./RectangleSelect";
import { InteractionMode } from "./Tool";
export class SelectionInfo {
    constructor() {
        this.bounds = new BoundingBox();
        this.worldMatrix = Matrix.Identity();
        this.canRotate = true;
        this.canTranslate = true;
        this.canDelete = true;
        this.update();
    }
    update() {
        // prevent numerical issues with flat objects
        const tSize = this.bounds.getSize();
        if (eq(tSize.x, 0)) {
            this.bounds.min.x -= EPSILON;
            this.bounds.max.x += EPSILON;
        }
        if (eq(tSize.y, 0)) {
            this.bounds.min.y -= EPSILON;
            this.bounds.max.y += EPSILON;
        }
        if (eq(tSize.z, 0)) {
            this.bounds.min.z -= EPSILON;
            this.bounds.max.z += EPSILON;
        }
        this.worldBounds = this.bounds.transform(this.worldMatrix);
        this.worldPoints = this.bounds.getPoints(this.worldMatrix);
    }
}
export class SelectionDefault extends AbstractTool {
    get multiSelectMode() {
        return (this.mMultiSelectMode);
    }
    set multiSelectMode(pEnabled) {
        this.setMultiSelectMode(pEnabled);
    }
    get showMainElementSelection() {
        return (this.mShowMainElementSelection);
    }
    set showMainElementSelection(pShow) {
        if (this.mShowMainElementSelection !== pShow) {
            this.mShowMainElementSelection = pShow;
            this.updateSelectionIfRunning();
        }
    }
    get showSubElementSelection() {
        return (this.mShowSubElementSelection);
    }
    set showSubElementSelection(pShow) {
        if (this.mShowSubElementSelection !== pShow) {
            this.mShowSubElementSelection = pShow;
            this.updateSelectionIfRunning();
        }
    }
    get showElementInteractors() {
        return (this.mShowElementInteractors);
    }
    set showElementInteractors(pShow) {
        if (this.mShowElementInteractors !== pShow) {
            this.mShowElementInteractors = pShow;
            this.updateSelectionIfRunning();
        }
    }
    constructor(pApp) {
        super(pApp);
        this.eventRectangleSelectFinished = new ListenableEvent(this);
        /**
         * Will be triggered (after the double click delay) if the user clicks/tabs
         * on the background and there is currently no selection.
         */
        this.eventBackgroundClick = new ListenableEvent(this);
        /** Will be triggered immediately if the user clicks/tabs on the background and there is currently no selection. */
        this.eventBackgroundClickImmediate = new ListenableEvent(this);
        /** Will be triggered if the user clicks/tabs on a SceneElement. */
        this.eventElementClick = new ListenableEvent(this);
        /** Will be triggered if the state of the multi selection mode has changed. */
        this.eventMultiSelectModeChanged = new ListenableEvent(this);
        this.onSelectionChanged = (pParameter) => {
            this.mSelectionChanged = true;
            //stop multi selection mode if selection is empty
            if (this.mMultiSelectMode && this.multiSelectAutoStartEnabled && this.mApp.model.selection.length === 0) {
                this.setMultiSelectMode(false);
            }
            this.mApp.viewer.requestRenderFrame();
        };
        this.onCameraChanged = (pParameter) => {
            this.mCameraChanged = true;
            this.mApp.viewer.requestRenderFrame();
        };
        this.onCameraModeChanged = (pParameter) => {
            this.mCameraModeChanged = true;
            this.mApp.viewer.requestRenderFrame();
        };
        this.onBeforeRender = (pParameter) => {
            let tUpdateInteractors = false;
            if (this.mSelectionChanged) {
                tUpdateInteractors = true;
                this.mSelectionChanged = false;
                this.updateSelection();
            }
            if (this.mCameraModeChanged) {
                tUpdateInteractors = true;
                this.mCameraModeChanged = false;
            }
            if (tUpdateInteractors || this.mCameraChanged) {
                this.updateInteractors(!tUpdateInteractors);
                this.mCameraChanged = false;
            }
        };
        this.onBeforeRenderCamera = (pCamera) => {
            if (pCamera instanceof Camera) {
                this.updateSelectionBoxSides(pCamera.view);
            }
        };
        this.onTriggerBackgroundClick = () => {
            this.eventBackgroundClick.trigger();
            this.mBackgroundClickTimer = null;
        };
        this.onHandleLongTouch = () => {
            this.mLongTouchTimer = null;
            //multi selection
            if (this.mPickedElement != null) {
                if (this.handleSelection(this.mPickedElement, true)) {
                    this.mPickedElement = null;
                    this.mHandleSelection = false;
                    this.mHandleClick = false;
                }
            }
        };
        this.multiSelectAutoStartEnabled = false;
        this.multiSelectDeselectAllEnabled = true;
        this.subElementSelectionEnabled = true;
        this.deselectionEnabled = true;
        this.leftMouseButtonRectangleSelectEnabled = false;
        this.rectangleSelectMode = "Complete";
        this.mSelectionBoxSides = [];
        this.mSelectionBoxSideGeom = [];
        this.mPickedElement = null;
        this.mSelection = [];
        this.mInteractorElements = [];
        this.mView = this.mApp.viewer.view;
        this.mPickedPos = Vector3.Zero();
        this.mSelectionInfo = new SelectionInfo();
        this.mRectSelectTool = null;
        this.mSelectionChanged = false;
        this.mCameraChanged = false;
        this.mCameraModeChanged = false;
        this.mHandleSelection = false;
        this.mHandleClick = false;
        this.mLongTouchTimer = null;
        this.mMultiSelectMode = false;
        this.mShowMainElementSelection = true;
        this.mShowSubElementSelection = true;
        this.mShowElementInteractors = true;
        this.mBackgroundClickTimer = null;
        this.prepareSelectionBoxSides();
    }
    getInteractionMode() {
        if (this.mRectSelectTool != null)
            return (this.mRectSelectTool.getInteractionMode());
        else
            return (super.getInteractionMode());
    }
    start(pParameter) {
        this.mApp.model.eventSelectionChanged.addListener(this.onSelectionChanged);
        this.mApp.viewer.eventBeforeRender.addListener(this.onBeforeRender);
        this.mApp.model.scene.onBeforeCameraRenderObservable.add(this.onBeforeRenderCamera);
        this.mView.cameraControl.eventCameraChanged.addListener(this.onCameraChanged);
        this.mView.cameraControl.eventCameraModeChanged.addListener(this.onCameraModeChanged);
        this.mPickedElement = null;
        this.mSelectionChanged = true;
        this.mCameraChanged = false;
        this.mCameraModeChanged = false;
        this.mHandleSelection = false;
        this.mHandleClick = false;
        this.mBackgroundClickTimer = null;
        this.mLongTouchTimer = null;
        this.mApp.viewer.requestRenderFrame();
        //stop multi selection mode if selection is empty
        if (this.mMultiSelectMode && this.multiSelectAutoStartEnabled && this.mApp.model.selection.length === 0) {
            this.setMultiSelectMode(false);
        }
        return (super.start(pParameter));
    }
    stop() {
        //rectangle select
        if (this.mRectSelectTool != null)
            this.mRectSelectTool.stop();
        this.mApp.model.eventSelectionChanged.removeListener(this.onSelectionChanged);
        this.mApp.viewer.eventBeforeRender.removeListener(this.onBeforeRender);
        this.mApp.model.scene.onBeforeCameraRenderObservable.removeCallback(this.onBeforeRenderCamera);
        this.mView.cameraControl.eventCameraChanged.removeListener(this.onCameraChanged);
        this.mView.cameraControl.eventCameraModeChanged.removeListener(this.onCameraModeChanged);
        this.clearSelection();
        this.resetBackgroundClickTimer();
        this.mApp.viewer.requestRenderFrame();
        super.stop();
    }
    handlePointerDown(pEvent, pState) {
        //rectangle select
        if (this.mRectSelectTool != null) {
            this.mRectSelectTool.handlePointerDown(pEvent, pState);
            return;
        }
        this.mHandleClick = true;
        this.mHandleSelection = false;
        const tIsPrimaryPointer = pState.isPrimaryPointer(pEvent);
        const tPointer = pState.pointers.get(pEvent.pointerId);
        if (tPointer == null)
            return;
        if (tIsPrimaryPointer) {
            //pick interactors
            if (this.pickInteractors(tPointer))
                return;
            //start long touch
            this.resetLongTouchTimer();
            this.mLongTouchTimer = window.setTimeout(this.onHandleLongTouch, SelectionDefault.LONG_TOUCH_TIMEOUT);
        }
        //pick model
        if (this.mPickedElement == null) {
            this.mPickedElement = this.getPickedElement(tPointer.position, !tPointer.isMousePointer, this.mPickedPos);
        }
        //disable camera control and enable selection handling
        if (this.mPickedElement != null || this.mSelection.length > 0) {
            this.setInteractionMode(InteractionMode.Primary);
            this.mHandleSelection = true;
        }
    }
    handlePointerMove(pEvent, pState) {
        //rectangle select
        if (this.mRectSelectTool != null) {
            this.mRectSelectTool.handlePointerMove(pEvent, pState);
            return;
        }
        if (pState.isMouseHoverPointer(pEvent)) {
            this.updateInteractorHover(new Vector2(pEvent.clientX, pEvent.clientY));
            return;
        }
        //pointer moved
        const tPointer = pState.getPointer(pEvent.pointerId);
        if (this.pointerWasMoved(tPointer)) {
            //stop long touch
            this.resetLongTouchTimer();
            //start rectangle selection with left mouse button
            if (this.leftMouseButtonRectangleSelectEnabled &&
                (this.mPickedElement == null || !this.mPickedElement.getTopLevelElement().isSelected) &&
                pState.isPrimaryPointer(pEvent) && PointerInfo.IsMousePointer(pEvent)) {
                this.startPointerDownRectangleSelect();
                return;
            }
            //reenable camera control and disable click handling
            if (this.mHandleClick) {
                this.setInteractionMode(InteractionMode.None);
                this.mHandleClick = false;
            }
            this.startPointerMoveTool(pEvent, pState);
        }
    }
    handlePointerUp(pEvent, pState) {
        //rectangle select
        if (this.mRectSelectTool != null) {
            this.mRectSelectTool.handlePointerUp(pEvent, pState);
            return;
        }
        const tPickedElement = this.mPickedElement;
        this.resetPickingStates();
        const tPointer = pState.pointers.get(pEvent.pointerId);
        if (tPointer == null || !pState.isPrimaryPointer(pEvent))
            return;
        //pointer was not moved and no long touch ==> handle click/tap
        if (!this.mHandleClick)
            return;
        let tBackgroundClick = true;
        //selection handling
        tBackgroundClick = !this.handleSelection(tPickedElement, pEvent.ctrlKey);
        //trigger background click event
        if (tBackgroundClick) {
            this.eventBackgroundClickImmediate.trigger();
            //trigger delayed background click event (to ignore double clicks)
            if (this.mBackgroundClickTimer == null) {
                this.mBackgroundClickTimer = window.setTimeout(this.onTriggerBackgroundClick, Viewer.MAX_DOUBLE_CLICK_DELAY * 1.2); //360ms
            }
            else {
                this.resetBackgroundClickTimer();
            }
        }
    }
    startRectangleSelect() {
        this.mRectSelectTool = new RectangleSelect(this.mApp);
        this.mRectSelectTool.multiSelectMode = this.mMultiSelectMode;
        this.mRectSelectTool.mode = this.rectangleSelectMode;
        this.mRectSelectTool.eventStopped.addListener(() => {
            this.mRectSelectTool = null;
            this.eventRectangleSelectFinished.trigger();
        });
    }
    updateInteractors(pOnlyCameraChanged) { }
    updateInteractorHover(pPos) { }
    pickInteractors(pPointer) { return (false); }
    startPointerMoveTool(pEvent, pState) { }
    getDirectionToCamera(pPos, pView) {
        let tCameraVec;
        if (pView.cameraControl.cameraMode === CameraMode.Perspective) {
            tCameraVec = pView.cameraControl.position.subtract(pPos);
            tCameraVec.normalize();
        }
        else
            tCameraVec = pView.cameraControl.direction.negate();
        return (tCameraVec);
    }
    resetBackgroundClickTimer() {
        if (this.mBackgroundClickTimer != null) {
            clearTimeout(this.mBackgroundClickTimer);
            this.mBackgroundClickTimer = null;
        }
    }
    resetLongTouchTimer() {
        if (this.mLongTouchTimer != null) {
            clearTimeout(this.mLongTouchTimer);
            this.mLongTouchTimer = null;
        }
    }
    pointerWasMoved(pPointer) {
        return (pPointer != null &&
            (Math.abs(pPointer.initialPosition.x - pPointer.position.x) > Viewer.POINTER_MOVE_TOLERANCE ||
                Math.abs(pPointer.initialPosition.y - pPointer.position.y) > Viewer.POINTER_MOVE_TOLERANCE));
    }
    setMultiSelectMode(pEnabled) {
        if (this.mMultiSelectMode !== pEnabled) {
            this.mMultiSelectMode = pEnabled;
            if (this.mRectSelectTool != null)
                this.mRectSelectTool.multiSelectMode = this.mMultiSelectMode;
            if (!this.mMultiSelectMode && this.mApp.model.selection.length > 1)
                this.mApp.model.setSelection(null);
            this.eventMultiSelectModeChanged.trigger();
        }
    }
    handleSelection(pPickedElement, pMultiSelect) {
        if (!this.mHandleSelection)
            return (false);
        let tHandled = false;
        this.mApp.model.beginSelectionChange();
        if (pPickedElement != null) {
            //trigger click event
            this.eventElementClick.trigger(pPickedElement);
            const tTopLevelElement = pPickedElement.getTopLevelElement();
            const tMainElement = pPickedElement.getMainElement();
            let tSubElement = pPickedElement.getSelectableElement();
            if (!this.subElementSelectionEnabled || tMainElement !== tTopLevelElement || tMainElement === tSubElement)
                tSubElement = null;
            if (pMultiSelect || this.mMultiSelectMode) {
                //multi selection
                //start multiSelection mode (if not already started)
                let tMultiSelectStarted = false;
                if (this.multiSelectAutoStartEnabled && !this.mMultiSelectMode) {
                    this.setMultiSelectMode(true);
                    tMultiSelectStarted = true;
                }
                if (tTopLevelElement.isSelected) {
                    //remove from selection
                    if (this.deselectionEnabled || this.mApp.model.selection.length > 1) {
                        if (!tMultiSelectStarted) {
                            this.mApp.model.removeFromSelection(tTopLevelElement);
                        }
                        tHandled = true;
                    }
                }
                else {
                    //add to selection
                    this.mApp.model.addToSelection(tTopLevelElement);
                    this.mApp.model.setSubElementSelection(tMainElement, tSubElement);
                    tHandled = true;
                }
            }
            else {
                //single selection
                if (!tTopLevelElement.isSelected || this.mApp.model.selection.length > 1)
                    this.mApp.model.setSelection([tTopLevelElement]);
                this.mApp.model.setSubElementSelection(tMainElement, tSubElement);
                tHandled = true;
            }
        }
        else {
            //deselect
            if (!pMultiSelect && (!this.mMultiSelectMode || this.multiSelectDeselectAllEnabled)) {
                if (this.deselectionEnabled) {
                    this.mApp.model.setSelection(null);
                    tHandled = true;
                }
                else {
                    if (this.mApp.model.hasSubElementSelection()) {
                        this.mApp.model.clearSubElementSelection();
                        tHandled = true;
                    }
                }
            }
        }
        this.mApp.model.endSelectionChange();
        return (tHandled);
    }
    getPickedElement(pPos, pUseFrustum, rPickPos) {
        let tPickInfo = null;
        if (pUseFrustum)
            tPickInfo = this.mView.pickSceneWithFrustum(pPos, SelectionDefault.PICK_FRUSTUM_SIZE, SelectionDefault.PICK_FRUSTUM_SIZE, CoordinatesType.Client);
        else
            tPickInfo = this.mView.pickScene(pPos, CoordinatesType.Client);
        if (tPickInfo != null && tPickInfo.hit && tPickInfo.pickedMesh != null) {
            if (rPickPos != null && tPickInfo.pickedPoint != null)
                rPickPos.copyFrom(tPickInfo.pickedPoint);
            return (SceneElement.FromNode(tPickInfo.pickedMesh, true));
        }
        return (null);
    }
    clearSelection() {
        for (const tBox of this.mSelectionBoxSides) {
            tBox.dispose();
        }
        this.mSelectionBoxSides = [];
        for (const tElement of this.mSelection) {
            //unregister events
            tElement.eventBoundingBoxChanged.removeListener(this.onSelectionChanged);
            tElement.eventTransformationChanged.removeListener(this.onSelectionChanged);
        }
        this.mSelection = [];
        //hide interactors
        for (const tElement of this.mInteractorElements) {
            tElement.showInteractors(false);
        }
        this.mInteractorElements = [];
    }
    updateSelectionIfRunning() {
        if (this.isRunning()) {
            this.updateSelection();
            this.mApp.viewer.requestRenderFrame();
        }
    }
    updateSelection() {
        this.clearSelection();
        this.mSelection = clone(this.mApp.model.selection);
        this.mSelectionInfo.bounds.reset();
        Matrix.IdentityToRef(this.mSelectionInfo.worldMatrix);
        this.mSelectionInfo.canDelete = true;
        this.mSelectionInfo.canRotate = true;
        this.mSelectionInfo.canTranslate = true;
        for (const tElement of this.mSelection) {
            //register events
            tElement.eventBoundingBoxChanged.addListener(this.onSelectionChanged);
            tElement.eventTransformationChanged.addListener(this.onSelectionChanged);
            const tBBox = tElement.getSelectionHighlightBoundingBox().clone();
            const tTransform = tElement.worldTransform;
            if (this.mShowMainElementSelection && tElement.isSelectionHighlightEnabled()) {
                this.createSelectionBox(tBBox, tTransform, false);
            }
            if (this.mSelection.length === 1) {
                this.mSelectionInfo.bounds = tBBox;
                this.mSelectionInfo.worldMatrix = tTransform.clone();
            }
            else {
                tBBox.transformInPlace(tTransform);
                this.mSelectionInfo.bounds.expandInPlace(tBBox);
            }
            this.mSelectionInfo.canDelete = this.mSelectionInfo.canDelete && tElement.canDelete();
            this.mSelectionInfo.canRotate = this.mSelectionInfo.canRotate && tElement.canRotate();
            this.mSelectionInfo.canTranslate = this.mSelectionInfo.canTranslate && tElement.canTranslate();
            //sub selection
            const tSubElement = this.mApp.model.getSubElementSelection(tElement);
            if (tSubElement != null) {
                if (this.mShowSubElementSelection && tSubElement.isSelectionHighlightEnabled()) {
                    this.createSelectionBox(tSubElement.boundingBoxVisible, tSubElement.worldTransform, true);
                }
                //show interactors
                if (this.mShowElementInteractors) {
                    tSubElement.showInteractors(true);
                    this.mInteractorElements.push(tSubElement);
                }
            }
            else {
                //show interactors
                if (this.mShowElementInteractors) {
                    tElement.showInteractors(true);
                    this.mInteractorElements.push(tElement);
                }
            }
        }
        this.mSelectionInfo.update();
    }
    prepareSelectionBoxSides() {
        const tPoints = new BoundingBox(new Vector3(-0.5, -0.5, -0.5), new Vector3(0.5, 0.5, 0.5)).getPoints();
        this.mSelectionBoxSideGeom = [];
        this.prepareSelectionBoxSide(tPoints, BoundingBox.SIDE_MIN_X);
        this.prepareSelectionBoxSide(tPoints, BoundingBox.SIDE_MAX_X);
        this.prepareSelectionBoxSide(tPoints, BoundingBox.SIDE_MIN_Y);
        this.prepareSelectionBoxSide(tPoints, BoundingBox.SIDE_MAX_Y);
        this.prepareSelectionBoxSide(tPoints, BoundingBox.SIDE_MIN_Z);
        this.prepareSelectionBoxSide(tPoints, BoundingBox.SIDE_MAX_Z);
    }
    prepareSelectionBoxSide(pBBPoints, pIndices) {
        const tLines = new Array();
        tLines.push(pBBPoints[pIndices[0]]);
        tLines.push(pBBPoints[pIndices[1]]);
        tLines.push(pBBPoints[pIndices[1]]);
        tLines.push(pBBPoints[pIndices[2]]);
        tLines.push(pBBPoints[pIndices[2]]);
        tLines.push(pBBPoints[pIndices[3]]);
        tLines.push(pBBPoints[pIndices[3]]);
        tLines.push(pBBPoints[pIndices[0]]);
        const tV1 = pBBPoints[pIndices[1]].subtract(pBBPoints[pIndices[0]]);
        const tV2 = pBBPoints[pIndices[2]].subtract(pBBPoints[pIndices[0]]);
        const tNormal = Vector3.Cross(tV1, tV2).normalize();
        const tCenter = Vector3.Center(pBBPoints[pIndices[0]], pBBPoints[pIndices[2]]);
        const tSide = ThickLinesMesh.CreateThickLines("selectionBoxSideGeom", this.mApp.model.scene, tLines);
        tSide.isPickable = false;
        tSide.thickness = 2.0;
        tSide.renderingGroupId = 1;
        tSide.layerMask = ViewLayerMaskFlags.All & ~ViewLayerMaskFlags.Screenshot;
        tSide.boxSide_normal = tNormal;
        tSide.boxSide_center = tCenter;
        removeFromScene(tSide, this.mApp.model.scene);
        this.mSelectionBoxSideGeom.push(tSide);
    }
    createSelectionBox(pBoundingBox, pTransform, pSubSelection) {
        if (pBoundingBox == null || pTransform == null)
            return;
        const tCenter = pBoundingBox.getCenter();
        const tColor = pSubSelection ? Color4.FromHexString("#868686FF") :
            Color4.FromHexString(this.mApp.theme.mainColorBright + "FF");
        //build pivot matrix
        const tMPivot = Matrix.Scaling((pBoundingBox.max.x - pBoundingBox.min.x) + SelectionDefault.SELECTION_BOX_OFFSET, (pBoundingBox.max.y - pBoundingBox.min.y) + SelectionDefault.SELECTION_BOX_OFFSET, (pBoundingBox.max.z - pBoundingBox.min.z) + SelectionDefault.SELECTION_BOX_OFFSET);
        tMPivot.multiplyToRef(Matrix.Translation(tCenter.x, tCenter.y, tCenter.z), tMPivot);
        const tTransform = tMPivot.multiply(pTransform);
        for (const tSideGeom of this.mSelectionBoxSideGeom) {
            const tSide = tSideGeom.clone("selectionBoxSide");
            tSide.color = tColor;
            setLocalMatrix(tSide, tTransform);
            tSide.boxSide_normal = Vector3.TransformNormal(tSideGeom.boxSide_normal, tTransform);
            tSide.boxSide_center = Vector3.TransformCoordinates(tSideGeom.boxSide_center, tTransform);
            this.mSelectionBoxSides.push(tSide);
        }
    }
    updateSelectionBoxSides(pView) {
        for (const tSide of this.mSelectionBoxSides) {
            const tNormal = tSide.boxSide_normal;
            const tCenter = tSide.boxSide_center;
            //get vector pointing from interactor to camera
            const tCameraVec = this.getDirectionToCamera(tCenter, pView);
            const tDot = Vector3.Dot(tNormal, tCameraVec);
            if (tDot < 0.0) {
                tSide.isVisible = false;
            }
            else {
                tSide.isVisible = true;
            }
        }
    }
    startPointerDownRectangleSelect() {
        this.resetPickingStates();
        this.startRectangleSelect();
        if (this.mRectSelectTool != null) {
            this.mRectSelectTool.startSelection();
        }
    }
    resetPickingStates() {
        this.mPickedElement = null;
        this.resetLongTouchTimer();
        this.setInteractionMode(InteractionMode.None);
    }
}
//// private ///////////////////////////////////////////////////////////////////////////////
SelectionDefault.PICK_FRUSTUM_SIZE = 30;
SelectionDefault.SELECTION_BOX_OFFSET = 0.025;
SelectionDefault.LONG_TOUCH_TIMEOUT = 750;
