import { Vector2 } from "@babylonjs/core/Maths/math.vector";
import { ERR_INVALID_STATE, newError } from "../../utils/error";
import { BoundingBox } from "../mdl/BoundingBox";
import { CoordinatesType } from "../view/View";
import { SelectionDefault } from "./SelectionDefault";
/**
 * The SelectionMenu can be used to provide a custom menu for the selected SceneElements.
 * A show and hide callback has to be provided to implement the custom menu.
 * The SelectionMenu only works in combination with a default tool derived from the SelectionDefault tool.
 */
export class SelectionMenu {
    get active() {
        return (this.mActive);
    }
    set active(pValue) {
        this.setActive(pValue);
    }
    /**
     * Creates and registers a new SelectionMenu.
     * @param pApp Reference to the core application.
     * @param pShowCallback Called if the menu should be shown. A 2d bounding rect in viewer coordinates is given for the current selection.
     * @param pHideCallback Called if the menu should be hidden.
     */
    constructor(pApp, pShowCallback, pHideCallback) {
        this.onElementClicked = (pElement) => {
            if (pElement == null)
                return;
            const tTopLevel = pElement.getTopLevelElement();
            if (tTopLevel.isSelected) {
                this.setActive(!this.active);
            }
        };
        this.onToolStarted = () => {
            if (this.mApp.tools.activeTool === this.mApp.tools.defaultTool)
                this.show();
            else {
                if (this.deactivateOnToolStart)
                    this.setActive(false);
                else
                    this.hide();
            }
        };
        this.onInteractorTriggered = () => {
            this.setActive(false);
        };
        this.onSelectionChanged = () => {
            this.mActive = true; //force to be active
            this.show(); //update position / hide if selection is empty
        };
        this.onCameraChanged = () => {
            this.hide();
            this.resetCameraTimeout();
            this.mCameraTimeout = window.setTimeout(() => this.show(), 200);
        };
        this.mApp = pApp;
        this.mView = this.mApp.viewer.view;
        this.mVisible = false;
        this.mActive = true;
        this.deactivateOnToolStart = true;
        this.mShowCallback = pShowCallback;
        this.mHideCallback = pHideCallback;
        this.mApp.model.eventSelectionChanged.addListener(this.onSelectionChanged);
        this.mApp.tools.eventToolStarted.addListener(this.onToolStarted);
        this.mApp.tools.eventInteractorTriggered.addListener(this.onInteractorTriggered);
        this.mView.cameraControl.eventCameraChanged.addListener(this.onCameraChanged);
        if (this.mApp.tools.defaultTool == null)
            throw newError(ERR_INVALID_STATE, "no default tool defined");
        if (this.mApp.tools.defaultTool instanceof SelectionDefault) {
            this.mApp.tools.defaultTool.eventElementClick.addListener(this.onElementClicked);
        }
        else
            throw newError(ERR_INVALID_STATE, "default tool not derived from SelectionDefault");
        this.show();
    }
    /** Disposes this menu. */
    dispose() {
        this.hide();
        this.mApp.model.eventSelectionChanged.removeListener(this.onSelectionChanged);
        this.mApp.tools.eventToolStarted.removeListener(this.onToolStarted);
        this.mApp.tools.eventInteractorTriggered.removeListener(this.onInteractorTriggered);
        if (this.mApp.tools.defaultTool instanceof SelectionDefault) {
            this.mApp.tools.defaultTool.eventElementClick.removeListener(this.onElementClicked);
        }
        this.mView.cameraControl.eventCameraChanged.removeListener(this.onCameraChanged);
    }
    setActive(pActive) {
        if (pActive === this.mActive)
            return;
        if (this.mActive) {
            this.mActive = false;
            this.hide();
        }
        else {
            this.mActive = true;
            this.show();
        }
    }
    hide() {
        if (this.mVisible) {
            this.mVisible = false;
            this.mHideCallback();
        }
    }
    resetCameraTimeout() {
        if (this.mCameraTimeout != null) {
            clearTimeout(this.mCameraTimeout);
            this.mCameraTimeout = undefined;
        }
    }
    show() {
        this.resetCameraTimeout();
        if (!this.mActive)
            return;
        if (!this.mApp.tools.isDefaultToolRunning() || this.mApp.model.selection.length === 0) {
            this.hide();
            return;
        }
        let tBoundsTransform;
        let tBounds;
        if (this.mApp.model.selection.length > 1) {
            tBounds = new BoundingBox();
            for (const tElement of this.mApp.model.selection) {
                let tElementBounds = tElement.getSelectionHighlightBoundingBox();
                tElementBounds = tElementBounds.transform(tElement.worldTransform);
                tBounds.expandInPlace(tElementBounds);
            }
            tBoundsTransform = null;
        }
        else {
            tBounds = this.mApp.model.selection[0].getSelectionHighlightBoundingBox();
            tBoundsTransform = this.mApp.model.selection[0].worldTransform;
        }
        if (!tBounds.isValid()) {
            this.hide();
            return;
        }
        const tPoints3d = tBounds.getPoints();
        const tBBox = new BoundingBox();
        for (const tPoint3d of tPoints3d) {
            const tPoint = this.mView.projectPoint3(tPoint3d, CoordinatesType.Viewer, tBoundsTransform, false);
            if (tPoint != null) {
                tBBox.expandByPointInPlace(tPoint);
            }
        }
        if (tBBox.isValid()) {
            this.mVisible = true;
            this.mShowCallback(new Vector2(tBBox.min.x, tBBox.min.y), new Vector2(tBBox.max.x, tBBox.max.y), this.mView);
        }
        else {
            this.hide();
        }
    }
}
