import type { ListenableEvent } from "../../utils";
import type { InputState } from "../view/InputState";
/**
 * The tool interaction mode defines which camera controls are available while the tool is running.
 **/
export declare const InteractionMode: {
    /**
     * The tool is performing no interactions.
     * Camera controls is available while tool is running.
     **/
    readonly None: 0;
    /**
     * The tool is performing primary interactions (left mouse button, one finger touch).
     * Only secondary camera controls are possible.
     **/
    readonly Primary: 1;
    /**
     * The tool is performing primary and secondary interactions.
     * Camera controls are completely disabled.
     **/
    readonly All: 2;
};
export type InteractionMode = typeof InteractionMode[keyof typeof InteractionMode];
export interface Tool {
    readonly eventStarted: ListenableEvent<void, Tool>;
    readonly eventStopped: ListenableEvent<void, Tool>;
    isAvailable(): boolean;
    /**
     * Should return true if the tool is running.
     * False should be returned after a call to stop() or if the
     * tool is finished with its operations.
     */
    isRunning(): boolean;
    /**
     * Should return the current interaction mode.
     * The interaction mode will be checked continuously while the tool is running.
     */
    getInteractionMode(): InteractionMode;
    /**
     * Starts the tool. Returns false if the start was canceled/aborted.
     */
    start(pParameter: unknown): boolean;
    stop(): void;
    handlePointerDown(pEvent: PointerEvent, pState: InputState): void;
    handlePointerUp(pEvent: PointerEvent, pState: InputState): void;
    handlePointerDblClick(pEvent: PointerEvent, pState: InputState): void;
    handlePointerMove(pEvent: PointerEvent, pState: InputState): void;
    handlePointerCancel(pEvent: PointerEvent, pState: InputState): void;
    handleKeyUp(pEvent: KeyboardEvent, pState: InputState): void;
}
