import { ListenableEvent, Log } from "../../utils";
import { ERR_INVALID_ARGUMENT, ERR_NOT_AVAILABLE } from "../../utils/error";
export class ToolManager {
    get activeTool() {
        return (this.mActiveTool);
    }
    get defaultTool() {
        return (this.mDefaultTool);
    }
    set defaultTool(pTool) {
        this.mDefaultTool = pTool;
    }
    constructor() {
        this.eventToolStarted = new ListenableEvent(this);
        this.eventToolStopped = new ListenableEvent(this);
        this.eventInteractorTriggered = new ListenableEvent(this);
        this.mDefaultTool = null;
        this.mActiveTool = null;
        this.mActiveDraggerInteractor = null;
        this.mActiveDraggerPointerId = null;
    }
    /**
     * Starts pTool and stops the current active tool.
     */
    startTool(pTool, pToolParameter = null) {
        if (pTool == null)
            throw (new Error(ERR_INVALID_ARGUMENT));
        if (!pTool.isAvailable())
            throw (new Error(ERR_NOT_AVAILABLE));
        const tStarted = pTool.start(pToolParameter);
        if (tStarted) {
            this.stopTool(this.mActiveTool);
            this.mActiveTool = pTool;
            this.eventToolStarted.trigger(pTool);
        }
        return (tStarted);
    }
    stopActiveTool() {
        this.stopTool(this.mActiveTool);
        this.startDefaultTool();
    }
    startDefaultTool() {
        if (this.mDefaultTool == null)
            return (false);
        if (this.isDefaultToolRunning())
            return (true); //already running
        return (this.startTool(this.mDefaultTool, null));
    }
    isDefaultToolRunning() {
        return (this.mActiveTool != null && this.mActiveTool === this.mDefaultTool);
    }
    //// internal //////////////////////////////////////////////////////////////////////////////
    /** @internal */
    get _activeDraggerInteractor() {
        return (this.mActiveDraggerInteractor);
    }
    /** @internal */
    _updateActiveToolState() {
        if (this.mActiveTool != null && !this.mActiveTool.isRunning()) {
            this.mActiveTool = null;
            this.startDefaultTool();
        }
    }
    /** @internal */
    _handlePointerDown(pEvent, pState) {
        const tTool = this.getActiveRunningTool();
        if (tTool != null)
            tTool.handlePointerDown(pEvent, pState);
    }
    /** @internal */
    _handlePointerUp(pEvent, pState) {
        this.finishInteractorDragging(pEvent);
        const tTool = this.getActiveRunningTool();
        if (tTool != null)
            tTool.handlePointerUp(pEvent, pState);
    }
    /** @internal */
    _handlePointerDblClick(pEvent, pState) {
        const tTool = this.getActiveRunningTool();
        if (tTool != null)
            tTool.handlePointerDblClick(pEvent, pState);
    }
    /** @internal */
    _handlePointerMove(pEvent, pState) {
        const tTool = this.getActiveRunningTool();
        if (tTool != null)
            tTool.handlePointerMove(pEvent, pState);
    }
    /** @internal */
    _handlePointerCancel(pEvent, pState) {
        this.finishInteractorDragging(pEvent);
        const tTool = this.getActiveRunningTool();
        if (tTool != null)
            tTool.handlePointerCancel(pEvent, pState);
    }
    /** @internal */
    _handleKeyUp(pEvent, pState) {
        const tTool = this.getActiveRunningTool();
        if (tTool != null)
            tTool.handleKeyUp(pEvent, pState);
        //esc -> abort active tool
        if (pEvent.which === 27)
            this.startDefaultTool();
    }
    /** @internal */
    _setActiveDraggerInteractor(pDragger, pPointerId) {
        this.mActiveDraggerInteractor = pDragger;
        this.mActiveDraggerPointerId = pPointerId;
    }
    stopTool(pTool) {
        if (pTool == null)
            return;
        try {
            if (pTool.isRunning()) {
                pTool.stop();
                this.eventToolStopped.trigger(pTool);
            }
        }
        catch (e) {
            Log.error("Failed to stop tool.", e);
        }
    }
    getActiveRunningTool() {
        if (this.mActiveTool != null && this.mActiveTool.isRunning())
            return (this.mActiveTool);
        return (null);
    }
    finishInteractorDragging(pEvent) {
        if (this.mActiveDraggerPointerId === pEvent.pointerId && this.mActiveDraggerInteractor != null) {
            const tDragger = this.mActiveDraggerInteractor;
            this.mActiveDraggerInteractor = null;
            this.mActiveDraggerPointerId = null;
            tDragger._finishDragging();
        }
    }
}
