import { Plane } from "@babylonjs/core/Maths/math.plane";
import type { Matrix } from "@babylonjs/core/Maths/math.vector";
import { Vector3 } from "@babylonjs/core/Maths/math.vector";
import { AbstractMesh } from "@babylonjs/core/Meshes/abstractMesh";
import type { TransformNode } from "@babylonjs/core/Meshes/transformNode";
import { ListenableEvent } from "../../utils";
import type { Serializable } from "../../utils/object";
import type { SceneElement } from "../mdl";
import type { CameraSetting } from "./io/Types";
import type { View } from "./View";
import type { Viewer } from "./Viewer";
export declare const CameraMode: {
    readonly Top: 0;
    readonly Bottom: 1;
    readonly Left: 2;
    readonly Right: 3;
    readonly Front: 4;
    readonly Back: 5;
    readonly Orthographic: 6;
    readonly Perspective: 7;
};
export type CameraMode = typeof CameraMode[keyof typeof CameraMode];
export declare const CameraModeMask: {
    readonly None: 0;
    readonly Top: 1;
    readonly Bottom: 2;
    readonly Left: 4;
    readonly Right: 8;
    readonly Front: 16;
    readonly Back: 32;
    readonly Orthographic: 64;
    readonly Perspective: 128;
    readonly All: 4294967295;
};
export type CameraModeMask = number;
export interface CameraZoomToFitOptions {
    /**
     * margin for the zoom (range: [0, 1], default: 0.05)
     * A value of 0.1 means a 5% border on all sides.
     **/
    margin?: number;
    /**
     * minimum margin in pixel for the zoom (default: 0)
     * A value of 10 means a 5 pixel border on all sides.
     **/
    minMarginPx?: number;
    /**
     * horizontal margin for the zoom (range: [0, 1], default: 0.05)
     * A value of 0.1 means a 5% border on all sides.
     * If undefined `margin` will be used.
     **/
    marginH?: number;
    /**
     * minimum horizontal margin in pixel for the zoom (default: 0)
     * A value of 10 means a 5 pixel border on all sides.
     * If undefined `minMarginPx` will be used.
     **/
    minMarginPxH?: number;
    /** if true the bounding spheres instead of the bounding boxes will be used for the calculation (default: false) */
    useBoundingSphere?: boolean;
    /** if true and the camera has a fixed target, the targed will be adjusted (default: true) */
    adjustFixedTarget?: boolean;
}
export declare class CameraControl implements Serializable<Array<CameraSetting>> {
    /** Default camera ortho/zoom. Used for all camera modes except Perspective. */
    static DEFAULT_CAMERA_ORTHO: number;
    /** Default camera position. Used for camera modes Orthographic and Perspective. */
    static DEFAULT_CAMERA_POSITION: Vector3;
    orbitSpeed: number;
    wheelZoomSpeed: number;
    touchZoomSpeed: number;
    controlsEnabled: boolean;
    orbitEnabled: boolean;
    panningEnabled: boolean;
    zoomEnabled: boolean;
    /**
     * Enables snapping for the orthographic camera mode during orbit. (Default: false)
     */
    orthoOrbitSnappingEnabled: boolean;
    dblClickZoomToFitEnabled: boolean;
    dblClickZoomToFitOptions: CameraZoomToFitOptions;
    /**
     * Can be used to disable camera navigation with left mouse button.
     * This way the left mouse button is free to use for other interactions.
     * Note: One finger touch navigation will still be possible. (Default: true)
     */
    leftMouseButtonEnabled: boolean;
    pitchMin: number;
    pitchMax: number;
    yawMin: number;
    yawMax: number;
    get view(): View;
    get cameraMode(): CameraMode;
    set cameraMode(pMode: CameraMode);
    get position(): Vector3;
    get direction(): Vector3;
    get right(): Vector3;
    get viewPlane(): Plane;
    get fixedTarget(): Vector3 | null;
    get transformMatrix(): Matrix;
    get orbitInertia(): number;
    set orbitInertia(pValue: number);
    get navigationArea(): {
        radius: number;
        center: Vector3;
    };
    readonly eventCameraModeChanged: ListenableEvent<void, CameraControl>;
    readonly eventCameraChanged: ListenableEvent<void, CameraControl>;
    constructor(pViewer: Viewer, pView: View);
    /**
     * Copies the camera settings (position, direction, target, mode ...) from the given CameraControl to this.
     */
    copyCameraSettingsFrom(pOther: CameraControl): void;
    setPosition(pPosition: Vector3, pKeepFixedTarget?: boolean): void;
    setDirection(pDirection: Vector3, pUpVector?: Vector3): boolean;
    setTarget(pTarget: Vector3, pUpVector?: Vector3): boolean;
    setFixedTarget(pFixedTarget: Vector3 | null): boolean;
    setCameraOrtho(pOrthoTop: number): void;
    setFov(pFov: number): void;
    /**
     * The navigation area is used to limit certain camera interactions.
     * E.g. the maximum camera distance (zoom) depends on it.
     * The area will not be updated automatically, it should be defined by the host application.
     */
    setNavigationArea(pRadius: number, pCenter?: Vector3): void;
    getFrustum(): Array<Plane>;
    notifyChange(pTriggerChangeEvent?: boolean): void;
    beginChange(): void;
    endChange(): void;
    /**
     * Performs zoom to fit on SceneElement basis.
     * If pElements is null all elements in the scene will be used.
     */
    zoomToFitElements(pElements: Array<SceneElement> | null, pOptions?: CameraZoomToFitOptions): void;
    zoomToFitMeshes(pMeshes: Array<AbstractMesh>, pOptions?: CameraZoomToFitOptions): void;
    zoomToFitNodes(pNodes: Array<TransformNode>, pOptions?: CameraZoomToFitOptions): void;
    /**
     * Performs a zoom step targeting the view center with the given amount.
     * Use a positive amount to zoom out and a negative amount to zoom in.
     * The amount typically should be in range [-1.0, 1.0]. 0.15 can be used as a reference value.
     */
    zoomBy(pAmount: number): void;
    serialize(): Array<CameraSetting>;
    deserialize(pIOCameraSettings: Array<CameraSetting>): boolean;
    resetCamera(pResetOrthoZoom: boolean, pResetCameraModeSettings: boolean): void;
    stopOrbitInertia(): void;
    static SetMeshCameraModeMask(pMesh: AbstractMesh, pMask: CameraModeMask | null): void;
    static GetMeshCameraModeMask(pMesh: AbstractMesh): CameraModeMask | null;
    private static readonly NODEDATA_CAMERAMODEMASK;
    private static readonly ZOOM_NO_HIT_MIN_TARGETDISTANCE;
    private static readonly OTHO_SNAP_TOLERANCE;
    private static readonly MIN_ORTHO;
    private mViewer;
    private mView;
    private mCamera;
    private mModel;
    private mToolManager;
    private mViewPlane;
    private mTargetPlane;
    private mGroundPlane;
    private mTargetDistance;
    private mDirection;
    private mRight;
    private mFixedTarget;
    private mTransformMatrix;
    private mNavAreaCenter;
    private mNavAreaRadius;
    private mMaxZoomOrtho;
    private mMaxZoomDistance;
    private mCameraMode;
    private mCameraModeMask;
    private mIsDirectionLocked;
    private mWasCameraChanged;
    private mChanged;
    private mChangeDepth;
    private mPanningPickPos;
    private mZoomPickPos;
    private mZoomPickDistance;
    private mToolInteractionMode;
    private mOrbitPickPos;
    private mOrbitRawRotation;
    private mOrbitInertia;
    private mOrbitDelta;
    private mOrbitStartTime;
    private mOrbitStarted;
    private mOrbitStoppedTimer;
    private mApplyOrbitInertia;
    private mCameraModeSettings;
    private cameraModeToString;
    private cameraModeFromString;
    private setCameraMode;
    private updateDirection;
    private updateViewPlane;
    private updateTargetPlane;
    private getNearestElementCenter;
    private updateTargetDistance;
    private updateToolInteractionMode;
    private getPlanePickPos;
    private resetPickState;
    private onOrbitStopped;
    private orbit;
    private updateOrbit;
    private panning;
    private wheelZoom;
    private touchZoomPan;
    private updatePerspectiveZoom;
    private calculateZoomToFitBounds;
    private zoomToFitOrtho;
    private zoomToFitPerspective;
    private getZoomToFitMarginV;
    private getZoomToFitMarginH;
}
