import { PickingInfo } from "@babylonjs/core/Collisions/pickingInfo";
import { Ray } from "@babylonjs/core/Culling/ray";
import { Matrix, Vector2, Vector3 } from "@babylonjs/core/Maths/math.vector";
import type { AbstractMesh } from "@babylonjs/core/Meshes/abstractMesh";
import "@babylonjs/core/Rendering/geometryBufferRendererSceneComponent";
import "@babylonjs/core/Rendering/prePassRendererSceneComponent";
import type { Serializable } from "../../utils/object";
import type { Frustum } from "../utils/MathUtils";
import { Camera } from "./Camera";
import { CameraControl } from "./CameraControl";
import type { View as IOView } from "./io/Types";
import type { Viewer } from "./Viewer";
export declare const ViewLayerMaskFlags: {
    readonly None: 0;
    readonly Screenshot: 1;
    readonly All: 4294967295;
};
export declare const CoordinatesType: {
    readonly Client: 1;
    readonly Viewer: 2;
};
export type CoordinatesType = typeof CoordinatesType[keyof typeof CoordinatesType];
export declare class View implements Serializable<IOView> {
    get viewer(): Viewer;
    get camera(): Camera;
    get cameraControl(): CameraControl;
    get overlayDiv(): HTMLDivElement;
    get width(): number;
    get height(): number;
    get renderWidth(): number;
    get renderHeight(): number;
    /** pixel scale factor is used to scale certain elements like 2D text */
    pixelScaleFactor: number;
    /**
     * Should be called after Viewer.removeView() was called and the view won't be used again.
     */
    dispose(): void;
    resize(): void;
    pickScene(pPos: Vector2, pCoordType: CoordinatesType, pPredicate?: (pMesh: AbstractMesh) => boolean): PickingInfo | null;
    pickSceneWithFrustum(pPos: Vector2, pWidth: number, pHeight: number, pCoordType: CoordinatesType, pPredicate?: (pMesh: AbstractMesh) => boolean): PickingInfo | null;
    createPickingRay(pPos: Vector2, pCoordType: CoordinatesType, pWorldMatrix?: Matrix | null): Ray;
    createPickingRayToRef(pPos: Vector2, pCoordType: CoordinatesType, pWorldMatrix: Matrix | null, pResult: Ray): void;
    createPickingFrustum(pPos: Vector2, pWidth: number, pHeight: number, pCoordType: CoordinatesType): Frustum;
    /**
     * Projects a point from 3d world space to 2d view/viewer space and returns it as Vector2.
     */
    projectPoint2(pPos: Vector3, pTargetCoordsType: CoordinatesType, pWorldMatrix?: Matrix | null, pCheckIfBehindCamera?: boolean): Vector2 | null;
    /**
     * Projects a point from 3d world space to 2d view/viewer space and returns it as Vector3.
     */
    projectPoint3(pPos: Vector3, pTargetCoordsType: CoordinatesType, pWorldMatrix?: Matrix | null, pCheckIfBehindCamera?: boolean): Vector3 | null;
    /**
     * Unprojects a point from 2d view/viewer space to 3d world space.
     */
    unprojectPoint(pPos: Vector3, pSourceCoordsType: CoordinatesType, pWorldMatrix?: Matrix | null): Vector3;
    /**
     * Unprojects a point from 2d view/viewer space to 3d world space.
     */
    unprojectPointToRef(pPos: Vector3, pSourceCoordsType: CoordinatesType, pWorldMatrix: Matrix | null, pResult: Vector3): void;
    /**
     * Projects the points p1 and p2 from world coordinates to viewer coordinates and returns its distance (pixel distance).
     */
    projectedDistance(pP1: Vector3, pP2: Vector3, pWorldMatrix?: Matrix | null): number | null;
    convertCoords(pCoords: Vector2 | Vector3, pFrom: CoordinatesType, pTo: CoordinatesType): Vector2;
    convertCoordsToRef(pCoords: Vector2 | Vector3, pFrom: CoordinatesType, pTo: CoordinatesType, pTarget: Vector2 | Vector3): void;
    serialize(): IOView;
    deserialize(pIOView: IOView): boolean;
    private static readonly sProjectViewport;
    private static readonly sIdentityMatrix;
    private static readonly sSamplingOffsets;
    private mViewer;
    private mCamera;
    private mCameraControl;
    private mOverlayDiv;
    private mSizeOverride;
    private mMSAASamplingPass;
    private mMSAAResultPass1;
    private mMSAAResultPass2;
    private mMSAASamplingPassAttached;
    private mMSAAResultPass1Attached;
    private mMSAAResultPass2Attached;
    private mMultiSamplingStep;
    private mFXAA;
    private mFXAAAttached;
    private mSSAO;
    private mSSAOMinSamples;
    private mSSAOMaxSamples;
    private mSSAOAttached;
    private constructor();
    private onBeforeCameraRender;
    private onAfterCameraRender;
    private setupMeshVisibility;
    private restoreMeshVisibility;
    private updateOrthoAspectRatio;
    private internalProjectPoint;
    private setupPostProcessing;
    private attachPostProcess;
    private detachPostProcess;
    private pickMeshWithFrustum;
}
