import { Color4 } from "@babylonjs/core/Maths/math.color";
import type { Vector3 } from "@babylonjs/core/Maths/math.vector";
import { Vector2 } from "@babylonjs/core/Maths/math.vector";
import type { AbstractMesh } from "@babylonjs/core/Meshes/abstractMesh";
import type { Scene } from "@babylonjs/core/scene";
import type { Serializable } from "../../utils/object";
import { ListenableEvent } from "../../utils";
import type { Application } from "../Application";
import { InputState } from "./InputState";
import type { Viewer as IOViewer } from "./io/Types";
import { View } from "./View";
/**
 * Base screenshot generation options.
 */
export interface BaseScreenshotOptions {
    /** width of the image in pixel */
    width: number;
    /** height of the image in pixel */
    height: number;
    /** pixel scale factor is used to scale certain elements like 2D text (default: 1.0) */
    pixelScaleFactor?: number;
    /** MIME-Type of the data uri (default: "image/png") */
    mimeType?: string;
    /** callback which will be called before the rendering */
    onBeforeRender?: () => void;
    /** callback which will be called after the rendering */
    onAfterRender?: () => void;
    /** Allows do draw a background for the generated screenshot. Only called if the current background color is transparent. */
    onDrawBackground?: (pContext: CanvasRenderingContext2D) => Promise<void>;
    /** allows do draw an overlay on the generated screenshot */
    onDrawOverlay?: (pContext: CanvasRenderingContext2D) => Promise<void>;
    /** if true the rendering will be performed asynchronously (default: false) */
    renderAsync?: boolean;
    /** defines the antialiasing quality, higher is better. (range: [0, 32], default: 32) */
    antiAliasingQuality?: number;
    /** can be used to override the viewer ssaoQuality */
    ssaoQuality?: SSAOQuality;
    /** can be used to override the viewer backgroundColor, for transparent backgrounds a premultiplied alpha color has to be used e.g. (0, 0, 0, 0) */
    backgroundColor?: Color4;
    /**
     * Optional target canvas. If defined the screenshot will be rendered into this canvas.
     * `Viewer.createScreenshot()` will return an empty string in this case.
     */
    targetCanvas?: HTMLCanvasElement;
}
/**
 * Screenshot generation options.
 */
export interface ScreenshotOptions extends BaseScreenshotOptions {
    /** list of meshes which should be rendered, if null all meshes in the scene will be renderd (default: null) */
    meshesToRender?: Array<AbstractMesh>;
}
export type SSAOQuality = "Off" | "Low" | "Medium" | "High";
export declare class Viewer implements Serializable<IOViewer> {
    static readonly MAX_DOUBLE_CLICK_DELAY = 300;
    static readonly POINTER_MOVE_TOLERANCE: number;
    get app(): Application;
    get container(): HTMLDivElement;
    get canvas(): HTMLCanvasElement;
    get scene(): Scene;
    get width(): number;
    get height(): number;
    get view(): View;
    get inputState(): InputState;
    get backgroundColor(): Color4;
    set backgroundColor(pColor: Color4);
    get renderingEnabled(): boolean;
    /**
     * defines the antialiasing quality, higher is better.
     * @valueRange [0, 32]
     * @defaultValue 32
     */
    get antiAliasingQuality(): number;
    set antiAliasingQuality(pQuality: number);
    /**
     * defines the SSAO quality or turns it off
     * @defaultValue "Medium"
     */
    get ssaoQuality(): SSAOQuality;
    set ssaoQuality(pValue: SSAOQuality);
    /**
     * Enables/disables FXAA. FXAA is used during user interaction to reduce aliasing.
     * Note: It is not recommended to use FXAA in combination with a transparent background because this leads to artifacts in some cases.
     * @defaultValue false
     */
    get fxaaEnabled(): boolean;
    set fxaaEnabled(pValue: boolean);
    showPointerFeedback: boolean;
    /**
     * If false the `wheel` event will not be handled. Camera zoom in/out will be disabled in this
     * case and default page scrolling will be performed instead. (default: true)
     */
    handleWheelEvent: boolean;
    /**
     * Can be used to enable permanent rendering (mainly for debugging purpose). Disabling the
     * on-demand rendering will also disable MSAA.
     */
    onDemandRenderingEnabled: boolean;
    /**
     * true if the viewer is currently creating a screenshot
     */
    get isCreatingScreenshot(): boolean;
    readonly eventResize: ListenableEvent<void, Viewer>;
    readonly eventBeforeRender: ListenableEvent<void, Viewer>;
    readonly eventAfterRender: ListenableEvent<void, Viewer>;
    constructor(pApp: Application, pContainer: HTMLDivElement);
    initialize(pAutoResize: boolean): boolean;
    dispose(): void;
    /**
     * Resets the camera settings of the view.
     */
    resetView(): void;
    /**
     * Resizes the gl context to fit the current container/canvas size.
     * @param pImmediateResize If true the resize will be performed immediately otherwise it
     * will be delayed in order to merge consecutive resize() calls. (default: false)
     */
    resize(pImmediateResize?: boolean): void;
    /**
     * Has to be called if the viewer should render the scene. (e.g. after the camera has changed)
     */
    requestRenderFrame(): void;
    /**
     * Forces immediate/synchronous rendering of the scene. This function should be used with care.
     * Under normal circumstances requestRenderFrame() should be used instead.
     */
    forceRenderFrame(): void;
    /**
     * Can be used to disable the rendering. Nested calls are supported.
     */
    beginDisableRendering(): void;
    /**
     * Can be used to re-enable the rendering. Nested calls are supported.
     */
    endDisableRendering(): void;
    clientToViewerCoords(pCoords: Vector2 | Vector3): Vector2;
    clientToViewerCoordsToRef(pCoords: Vector2 | Vector3, pTarget: Vector2 | Vector3): void;
    viewerToClientCoords(pCoords: Vector2 | Vector3): Vector2;
    viewerToClientCoordsToRef(pCoords: Vector2 | Vector3, pTarget: Vector2 | Vector3): void;
    /**
     * Creates a screenshot and returns it as data uri or blob.
     */
    createScreenshot(pOptions: ScreenshotOptions): Promise<string>;
    createScreenshot(pOptions: ScreenshotOptions, pAsBlob: false): Promise<string>;
    createScreenshot(pOptions: ScreenshotOptions, pAsBlob: true): Promise<Blob>;
    createScreenshot(pOptions: ScreenshotOptions, pAsBlob: boolean | undefined): Promise<string | Blob>;
    serialize(): IOViewer;
    deserialize(pIOViewer?: IOViewer | null): boolean;
    private static readonly MSAA_DELAY_FRAMES;
    private static readonly MSAA_MAX_QUALITY;
    private static readonly MIN_CANVAS_SIZE;
    private mApp;
    private mContainer;
    private mCanvas;
    private mScene;
    private mTools;
    private mWidth;
    private mHeight;
    private mResizeTimeOut;
    private mRenderFrameRequested;
    private mDisableRenderingDepth;
    private mView;
    private mInputState;
    private mDoubleClickCounter;
    private mDoubleClickTimer;
    private mClientOffset;
    private mNextMSAAStep;
    private mMSAAQuality;
    private mMSAADelayFrame;
    private mSSAOQuality;
    private mFXAAEnabled;
    private mScreenshotLock;
    private mPerformResizeAfterScreenshot;
    private mDumpFrameBufferCanvas;
    private resetDoubleClickState;
    private getClientOffset;
    private getContainerWidth;
    private getContainerHeight;
    private onResize;
    private onWebGLContextRestored;
    private onPointerDown;
    private onPointerUp;
    private onPointerMove;
    private onPointerCancel;
    private onKeyUp;
    private onMouseWheel;
    private onContextMenu;
    private onScroll;
    private performResize;
    private createPointerFeedback;
    private internalCreateScreenshot;
    private markScreenshotMeshes;
    private createScreenshotOverlayCanvas;
    private reomveScreenshotOverlayCanvas;
    private dumpFramebuffer;
}
