import { Color4 } from "@babylonjs/core/Maths/math.color";
import { Vector2 } from "@babylonjs/core/Maths/math.vector";
import { ListenableEvent } from "../../utils";
import { Lock } from "../../utils/async";
import { ERR_INVALID_ARGUMENT, newError } from "../../utils/error";
import { canvasToBlob } from "../../utils/image";
import { clamp } from "../../utils/math";
import { InputState, PointerInfo } from "./InputState";
import { View, ViewLayerMaskFlags } from "./View";
export class Viewer {
    get app() {
        return (this.mApp);
    }
    get container() {
        return (this.mContainer);
    }
    get canvas() {
        return (this.mCanvas);
    }
    get scene() {
        return (this.mScene);
    }
    get width() {
        return (this.mWidth);
    }
    get height() {
        return (this.mHeight);
    }
    get view() {
        return (this.mView);
    }
    get inputState() {
        return (this.mInputState);
    }
    get backgroundColor() {
        return (this.mScene.clearColor);
    }
    set backgroundColor(pColor) {
        this.mScene.clearColor = pColor;
    }
    get renderingEnabled() {
        return (this.mDisableRenderingDepth === 0);
    }
    /**
     * defines the antialiasing quality, higher is better.
     * @valueRange [0, 32]
     * @defaultValue 32
     */
    get antiAliasingQuality() {
        return (this.mMSAAQuality);
    }
    set antiAliasingQuality(pQuality) {
        this.mMSAAQuality = clamp(pQuality, 0, Viewer.MSAA_MAX_QUALITY);
    }
    /**
     * defines the SSAO quality or turns it off
     * @defaultValue "Medium"
     */
    get ssaoQuality() {
        return (this.mSSAOQuality);
    }
    set ssaoQuality(pValue) {
        if (this.mSSAOQuality === pValue)
            return;
        this.mSSAOQuality = pValue;
        if (this.mView != null) {
            this.mView._setSSAOQuality(this.mSSAOQuality);
        }
    }
    /**
     * Enables/disables FXAA. FXAA is used during user interaction to reduce aliasing.
     * Note: It is not recommended to use FXAA in combination with a transparent background because this leads to artifacts in some cases.
     * @defaultValue false
     */
    get fxaaEnabled() {
        return (this.mFXAAEnabled);
    }
    set fxaaEnabled(pValue) {
        if (this.mFXAAEnabled === pValue)
            return;
        this.mFXAAEnabled = pValue;
        if (this.mView != null) {
            this.mView._updateFXAA();
        }
    }
    /**
     * true if the viewer is currently creating a screenshot
     */
    get isCreatingScreenshot() {
        return (this.mScreenshotLock.isLocked);
    }
    constructor(pApp, pContainer) {
        this.eventResize = new ListenableEvent(this);
        this.eventBeforeRender = new ListenableEvent(this);
        this.eventAfterRender = new ListenableEvent(this);
        //// internal //////////////////////////////////////////////////////////////////////////////
        /** @internal */
        this._renderOnlyScreenshotMeshes = false;
        this.resetDoubleClickState = () => {
            this.mDoubleClickCounter = 0;
            if (this.mDoubleClickTimer != null) {
                clearTimeout(this.mDoubleClickTimer);
                this.mDoubleClickTimer = null;
            }
        };
        this.onResize = () => {
            this.resize();
        };
        this.onWebGLContextRestored = () => {
            this.requestRenderFrame();
        };
        this.onPointerDown = (pEvent) => {
            if (this.mInputState.registerPointer(pEvent)) {
                this.mTools._handlePointerDown(pEvent, this.mInputState);
                this.mView.cameraControl._handlePointerDown(pEvent, this.mInputState);
                this.mCanvas.setPointerCapture(pEvent.pointerId);
                if (document.activeElement != this.mCanvas)
                    this.mCanvas.focus({ preventScroll: true }); //set focus on canvas
                //double click handling
                if (this.mDoubleClickCounter === 0) {
                    this.mDoubleClickTimer = window.setTimeout(this.resetDoubleClickState, Viewer.MAX_DOUBLE_CLICK_DELAY);
                }
                this.mDoubleClickCounter++;
                this.createPointerFeedback(pEvent);
                pEvent.preventDefault();
            }
        };
        this.onPointerUp = (pEvent) => {
            if (this.mInputState.updatePointer(pEvent)) {
                this.mTools._handlePointerUp(pEvent, this.mInputState);
                this.mView.cameraControl._handlePointerUp(pEvent, this.mInputState);
                //double click event
                if (this.mDoubleClickCounter >= 2) {
                    this.resetDoubleClickState();
                    this.mTools._handlePointerDblClick(pEvent, this.mInputState);
                    this.mView.cameraControl._handlePointerDblClick(pEvent, this.mInputState);
                }
                //deregister pointer
                this.mInputState.deregisterPointer(pEvent.pointerId);
                pEvent.preventDefault();
            }
        };
        this.onPointerMove = (pEvent) => {
            if (this.mInputState.updatePointer(pEvent)) {
                //reset double click state if pointer was moved
                const tPointer = this.mInputState.pointers.get(pEvent.pointerId);
                if (tPointer != null &&
                    (Math.abs(tPointer.initialPosition.x - tPointer.position.x) > Viewer.POINTER_MOVE_TOLERANCE ||
                        Math.abs(tPointer.initialPosition.y - tPointer.position.y) > Viewer.POINTER_MOVE_TOLERANCE)) {
                    this.resetDoubleClickState();
                }
            }
            this.mTools._handlePointerMove(pEvent, this.mInputState);
            this.mView.cameraControl._handlePointerMove(pEvent, this.mInputState);
            pEvent.preventDefault();
        };
        this.onPointerCancel = (pEvent) => {
            if (this.mInputState.pointers.has(pEvent.pointerId)) {
                this.mView.cameraControl._handlePointerCancel(pEvent, this.mInputState);
                this.mTools._handlePointerCancel(pEvent, this.mInputState);
                this.mInputState.deregisterPointer(pEvent.pointerId);
                this.mCanvas.releasePointerCapture(pEvent.pointerId);
                this.resetDoubleClickState();
            }
        };
        this.onKeyUp = (pEvent) => {
            this.mTools._handleKeyUp(pEvent, this.mInputState);
        };
        this.onMouseWheel = (pEvent) => {
            if (!this.handleWheelEvent) {
                return;
            }
            this.mInputState._updatePointerPosition(pEvent);
            this.mView.cameraControl._handleMouseWheel(pEvent, this.mInputState);
            pEvent.preventDefault();
        };
        this.onContextMenu = (pEvent) => {
            //prevent browser context menu
            pEvent.preventDefault();
        };
        this.onScroll = (pEvent) => {
            this.mClientOffset = null; //reset cached client offset
        };
        this.mApp = pApp;
        this.mContainer = pContainer;
        this.mRenderFrameRequested = true;
        this.mDisableRenderingDepth = 0;
        this.showPointerFeedback = true;
        this.handleWheelEvent = true;
        this.onDemandRenderingEnabled = true;
        this.mInputState = new InputState();
        this.mDoubleClickCounter = 0;
        this.mDoubleClickTimer = null;
        this.mWidth = -1;
        this.mHeight = -1;
        this.mClientOffset = null;
        this.mResizeTimeOut = null;
        this.mNextMSAAStep = -1;
        this.mMSAAQuality = Viewer.MSAA_MAX_QUALITY;
        this.mMSAADelayFrame = 0;
        this.mSSAOQuality = "Medium";
        this.mFXAAEnabled = false;
        this.mScreenshotLock = new Lock();
        this.mPerformResizeAfterScreenshot = false;
        //container must be positioned relative
        this.mContainer.style.position = "relative";
        //create canvas
        this.mCanvas = document.createElement("canvas");
        this.mCanvas.style.width = this.getContainerWidth().toString() + "px";
        this.mCanvas.style.height = this.getContainerHeight().toString() + "px";
        this.mCanvas.style.display = "block";
        this.mCanvas.style.outline = "none";
        this.mCanvas.style.msUserSelect = "none";
        this.mCanvas.style.webkitUserSelect = "none";
        this.mCanvas.style.userSelect = "none";
        this.mCanvas.tabIndex = 0; //make canvas focusable
        this.mContainer.appendChild(this.mCanvas);
    }
    initialize(pAutoResize) {
        if (this.mApp.model == null || this.mApp.tools == null)
            return (false);
        this.mApp.engine.canvasTabIndex = this.mCanvas.tabIndex;
        this.mScene = this.mApp.scene;
        this.mScene.clearColor = new Color4(1, 1, 1, 1);
        this.mTools = this.mApp.tools;
        this.mWidth = this.mCanvas.clientWidth;
        this.mHeight = this.mCanvas.clientHeight;
        this.mView = View._CreateView(this);
        this.mView.resize();
        this.mView._setSSAOQuality(this.mSSAOQuality);
        this.mScene.activeCamera = this.mView.camera;
        //register global resize event
        if (pAutoResize)
            window.addEventListener("resize", this.onResize);
        //register canvas events
        this.mCanvas.addEventListener("pointerdown", this.onPointerDown, false);
        this.mCanvas.addEventListener("pointerup", this.onPointerUp, false);
        this.mCanvas.addEventListener("pointermove", this.onPointerMove, false);
        //pointer cancel gets triggered only in very rare cases from IOS (e.g if the iPad is rotated while a pointer is down)
        this.mCanvas.addEventListener("pointercancel", this.onPointerCancel, false);
        this.mCanvas.addEventListener("keyup", this.onKeyUp, false);
        this.mCanvas.addEventListener("wheel", this.onMouseWheel, { capture: false, passive: false });
        this.mCanvas.addEventListener("contextmenu", this.onContextMenu, false);
        //Workaround:
        //IOS doesn't trigger the up event on the canvas in some cases.
        //e.g. if two or more fingers are moved outside of the screen. (IOS bug?)
        //up and cancel are registered also on the window, to make sure that the pointer gets correctly deregistered
        window.addEventListener("pointerup", this.onPointerUp, false);
        window.addEventListener("pointercancel", this.onPointerCancel, false);
        //capture all scroll events to reset cached client offset
        window.addEventListener("scroll", this.onScroll, true);
        this.mApp.engine.onContextRestoredObservable.add(this.onWebGLContextRestored);
        /**
         * BabylonJs calls focus() on the canvas without `preventScroll: true` when preventDefault is used.
         * This should not have side effects as preventDefault() is also called our event handlers of the Viewer.
         */
        this.mScene.preventDefaultOnPointerDown = false;
        this.mScene.preventDefaultOnPointerUp = false;
        return (true);
    }
    dispose() {
        //not initialized?
        if (this.mScene == null)
            return;
        if (this.mView != null) {
            this.mView.dispose();
        }
        //remove events
        window.removeEventListener("resize", this.onResize);
        this.mCanvas.removeEventListener("pointerdown", this.onPointerDown, false);
        this.mCanvas.removeEventListener("pointerup", this.onPointerUp, false);
        this.mCanvas.removeEventListener("pointermove", this.onPointerMove, false);
        this.mCanvas.removeEventListener("pointercancel", this.onPointerCancel, false);
        this.mCanvas.removeEventListener("keyup", this.onKeyUp, false);
        this.mCanvas.removeEventListener("wheel", this.onMouseWheel, false);
        this.mCanvas.removeEventListener("contextmenu", this.onContextMenu, false);
        window.removeEventListener("pointerup", this.onPointerUp, false);
        window.removeEventListener("pointercancel", this.onPointerCancel, false);
        window.removeEventListener("scroll", this.onScroll, true);
    }
    /**
     * Resets the camera settings of the view.
     */
    resetView() {
        this.mView.cameraControl.resetCamera(true, true);
    }
    /**
     * Resizes the gl context to fit the current container/canvas size.
     * @param pImmediateResize If true the resize will be performed immediately otherwise it
     * will be delayed in order to merge consecutive resize() calls. (default: false)
     */
    resize(pImmediateResize = false) {
        if (pImmediateResize) {
            this.performResize();
        }
        else {
            if (this.mResizeTimeOut == null) {
                this.mResizeTimeOut = window.setTimeout(() => {
                    this.performResize();
                }, 100);
            }
        }
    }
    /**
     * Has to be called if the viewer should render the scene. (e.g. after the camera has changed)
     */
    requestRenderFrame() {
        this.mMSAADelayFrame = 0;
        this.mRenderFrameRequested = true;
    }
    /**
     * Forces immediate/synchronous rendering of the scene. This function should be used with care.
     * Under normal circumstances requestRenderFrame() should be used instead.
     */
    forceRenderFrame() {
        this.mMSAADelayFrame = 0;
        this._renderFrame(true);
    }
    /**
     * Can be used to disable the rendering. Nested calls are supported.
     */
    beginDisableRendering() {
        this.mDisableRenderingDepth++;
    }
    /**
     * Can be used to re-enable the rendering. Nested calls are supported.
     */
    endDisableRendering() {
        if (this.mDisableRenderingDepth > 0)
            this.mDisableRenderingDepth--;
    }
    clientToViewerCoords(pCoords) {
        if (pCoords == null)
            throw new Error(ERR_INVALID_ARGUMENT);
        const tCoords = Vector2.Zero();
        this.clientToViewerCoordsToRef(pCoords, tCoords);
        return (tCoords);
    }
    clientToViewerCoordsToRef(pCoords, pTarget) {
        if (pCoords == null || pTarget == null)
            return;
        const tOffset = this.getClientOffset();
        pTarget.x = pCoords.x - tOffset.x;
        pTarget.y = pCoords.y - tOffset.y;
    }
    viewerToClientCoords(pCoords) {
        if (pCoords == null)
            throw new Error(ERR_INVALID_ARGUMENT);
        const tCoords = Vector2.Zero();
        this.viewerToClientCoordsToRef(pCoords, tCoords);
        return (tCoords);
    }
    viewerToClientCoordsToRef(pCoords, pTarget) {
        if (pCoords == null || pTarget == null)
            return;
        const tOffset = this.getClientOffset();
        pTarget.x = pCoords.x + tOffset.x;
        pTarget.y = pCoords.y + tOffset.y;
    }
    async createScreenshot(pOptions, pAsBlob) {
        if (pOptions == null)
            throw newError(ERR_INVALID_ARGUMENT);
        const tLockContext = await this.mScreenshotLock.acquireLock();
        try {
            const tImage = await this.internalCreateScreenshot(pOptions, pAsBlob);
            return (tImage);
        }
        finally {
            tLockContext.unlock();
            if (this.mPerformResizeAfterScreenshot) {
                this.performResize();
            }
        }
    }
    serialize() {
        const tIOViewer = {};
        tIOViewer.views = [];
        const tIOView = this.mView.serialize();
        if (tIOView != null)
            tIOViewer.views.push(tIOView);
        return (tIOViewer);
    }
    deserialize(pIOViewer) {
        if (pIOViewer != null && pIOViewer.views.length > 0) {
            this.mView.deserialize(pIOViewer.views[0]);
        }
        return (true);
    }
    /** @internal */
    _isScreenshotMesh(pMesh) {
        return (pMesh._wcf_core_isScreenshotMesh === true);
    }
    /** @internal */
    _renderFrame(pForce) {
        //disable rendering while creating screenshots
        if ((!this.renderingEnabled && !pForce) || this.mScreenshotLock.isLocked)
            return;
        //get camera changed flag
        const tCameraChanged = this.mView.cameraControl._beforeRender();
        const tNeedUpdateFrame = !this.onDemandRenderingEnabled || this.mRenderFrameRequested || tCameraChanged || pForce;
        if (tNeedUpdateFrame) {
            this.mNextMSAAStep = -1; //reset antialiasing
            this.eventBeforeRender.trigger();
        }
        this.mMSAADelayFrame++;
        const tNeedUpdateMSAA = this.mMSAADelayFrame > Viewer.MSAA_DELAY_FRAMES &&
            this.mNextMSAAStep <= this.mMSAAQuality && this.mMSAAQuality > 0;
        if (tNeedUpdateFrame || tNeedUpdateMSAA) {
            this.mView._prepareRenderFrame(this.mNextMSAAStep, this.mMSAAQuality);
            this.mScene.render();
            this.mView._finalizeRenderFrame(this.mNextMSAAStep, this.mMSAAQuality);
            this.mNextMSAAStep++;
            this.mRenderFrameRequested = false;
            if (tNeedUpdateFrame)
                this.eventAfterRender.trigger();
        }
    }
    /** @internal */
    _forwardEvent(pEvent, pOnlyUpdatePointerPosition) {
        if (pEvent == null)
            return;
        let tUpdatePointerPosition = false;
        switch (pEvent.type) {
            case "pointerdown":
                if (pOnlyUpdatePointerPosition)
                    tUpdatePointerPosition = true;
                else
                    this.onPointerDown(pEvent);
                break;
            case "wheel":
                if (pOnlyUpdatePointerPosition)
                    tUpdatePointerPosition = true;
                else
                    this.onMouseWheel(pEvent);
                break;
            case "click": //forwarded from button interactors
                tUpdatePointerPosition = true;
                break;
        }
        if (tUpdatePointerPosition) {
            this.mInputState._updatePointerPosition(pEvent);
        }
    }
    getClientOffset() {
        if (this.mClientOffset == null) {
            const tRect = this.mCanvas.getBoundingClientRect();
            this.mClientOffset = new Vector2(tRect.left, tRect.top);
        }
        return (this.mClientOffset);
    }
    getContainerWidth() {
        return (this.mContainer.clientWidth >= Viewer.MIN_CANVAS_SIZE ? this.mContainer.clientWidth : Viewer.MIN_CANVAS_SIZE);
    }
    getContainerHeight() {
        return (this.mContainer.clientHeight >= Viewer.MIN_CANVAS_SIZE ? this.mContainer.clientHeight : Viewer.MIN_CANVAS_SIZE);
    }
    performResize() {
        //do not resize while creating a screenshot
        if (this.mScreenshotLock.isLocked) {
            this.mPerformResizeAfterScreenshot = true;
            return;
        }
        else {
            this.mPerformResizeAfterScreenshot = false;
        }
        //Workaround:
        //If canvas deviceRatio is != 1 the canvas is rendered to big
        //after minimizing/maximzing the browser on IOS.
        //Using fixed pixel dimensions and delaying the resizing seems (see resize()) to prevent this issue.
        this.mResizeTimeOut = null;
        this.mClientOffset = null; //reset cached client offset
        const tNewWidth = this.getContainerWidth();
        const tNewHeight = this.getContainerHeight();
        this.mCanvas.style.width = tNewWidth.toString() + "px";
        this.mCanvas.style.height = tNewHeight.toString() + "px";
        if (this.mWidth != tNewWidth || this.mHeight != tNewHeight) {
            this.mWidth = tNewWidth;
            this.mHeight = tNewHeight;
            this.mApp.engine.resize();
            this.mView.resize();
            this.eventResize.trigger();
            this.requestRenderFrame();
        }
    }
    createPointerFeedback(pEvent) {
        if (!this.showPointerFeedback)
            return;
        if (pEvent == null || PointerInfo.IsMousePointer(pEvent))
            return;
        const tPointer = this.mInputState.pointers.get(pEvent.pointerId);
        if (tPointer == null)
            return;
        const tViewerPos = this.clientToViewerCoords(tPointer.position);
        const tPointerFb = document.createElement("div");
        tPointerFb.className = "wcf-core pointer-feedback";
        tPointerFb.style.left = tViewerPos.x.toString() + "px";
        tPointerFb.style.top = tViewerPos.y.toString() + "px";
        this.mContainer.appendChild(tPointerFb);
        setTimeout(() => {
            this.mContainer.removeChild(tPointerFb);
        }, 1000);
    }
    async internalCreateScreenshot(pOptions, pAsBlob) {
        //setup camera and post processing
        const tOldActiveCamera = this.mScene.activeCamera;
        const tOldLayerMask = this.mView.camera.layerMask;
        const tOldClearColor = this.mScene.clearColor;
        const tOldSizeOverride = this.mView._getSizeOverride();
        const tOldPixelScaleFactor = this.mView.pixelScaleFactor;
        const tOldCanvasWidth = this.mCanvas.width;
        const tOldCanvasHeight = this.mCanvas.height;
        const tOldSSAOQuality = this.mSSAOQuality;
        let tImageData = null;
        const tOverlayCanvas = this.createScreenshotOverlayCanvas();
        try {
            let tWidth = Math.round(pOptions.width);
            let tHeight = Math.round(pOptions.height);
            //adjust view size for correct rendering and projection calculations
            this.mApp.engine.setSize(tWidth, tHeight);
            this.mView._setSizeOverride(new Vector2(tWidth, tHeight), false);
            //trigger before render
            if (pOptions.onBeforeRender != null) {
                pOptions.onBeforeRender();
                //check if onBeforeRender changed the rendersize
                const tNewWidth = Math.round(pOptions.width);
                const tNewHeight = Math.round(pOptions.height);
                if (tNewWidth != tWidth || tNewHeight != tHeight) {
                    tWidth = tNewWidth;
                    tHeight = tNewHeight;
                    this.mApp.engine.setSize(tWidth, tHeight);
                    this.mView._setSizeOverride(new Vector2(tWidth, tHeight), false);
                }
            }
            this.mScene.activeCamera = this.mView.camera;
            this.mView.pixelScaleFactor = pOptions.pixelScaleFactor != null ? pOptions.pixelScaleFactor : 1.0;
            this.mView.camera.layerMask = ViewLayerMaskFlags.Screenshot;
            this.mView.camera.notifyProjectionChanged();
            if (pOptions.backgroundColor != null)
                this.mScene.clearColor = pOptions.backgroundColor;
            if (pOptions.ssaoQuality != null) {
                this.mView._setSSAOQuality(pOptions.ssaoQuality);
            }
            let tCurrentMSAAStep = 0;
            const tRenderAsync = pOptions.renderAsync != null ? pOptions.renderAsync : false;
            let tMaxAASteps = pOptions.antiAliasingQuality != null ? pOptions.antiAliasingQuality : Viewer.MSAA_MAX_QUALITY;
            tMaxAASteps = clamp(tMaxAASteps, 0, Viewer.MSAA_MAX_QUALITY);
            //prepare meshes to render
            this.markScreenshotMeshes(pOptions.meshesToRender, true);
            this._renderOnlyScreenshotMeshes = pOptions.meshesToRender != null;
            //setup render promise
            const tRenderPromise = new Promise((resolve, reject) => {
                const renderFrame = () => {
                    this.mView._prepareRenderFrame(tCurrentMSAAStep, tMaxAASteps);
                    this.mScene.render();
                    this.mView._finalizeRenderFrame(tCurrentMSAAStep, tMaxAASteps);
                    if (tCurrentMSAAStep === tMaxAASteps) {
                        this.dumpFramebuffer(tWidth, tHeight, pOptions, pAsBlob)
                            .then((pResult) => resolve(pResult))
                            .catch((pError) => reject(pError));
                    }
                    tCurrentMSAAStep++;
                    if (tCurrentMSAAStep <= tMaxAASteps) {
                        //render next
                        if (tRenderAsync)
                            setTimeout(renderFrame);
                        else
                            renderFrame();
                    }
                };
                //render frames
                renderFrame();
            });
            tImageData = await tRenderPromise;
            return (tImageData);
        }
        finally {
            //clean up
            this.mScene.clearColor = tOldClearColor;
            this.mScene.activeCamera = tOldActiveCamera;
            this.mView.pixelScaleFactor = tOldPixelScaleFactor;
            this.mView._setSSAOQuality(tOldSSAOQuality);
            this.mView.camera.layerMask = tOldLayerMask;
            this.mView.camera.notifyProjectionChanged();
            this.markScreenshotMeshes(pOptions.meshesToRender, false);
            this._renderOnlyScreenshotMeshes = false;
            //reset render size
            //has to be done before onAfterRender to prevent wrong calculations
            this.mApp.engine.setSize(tOldCanvasWidth, tOldCanvasHeight);
            this.mView._setSizeOverride(tOldSizeOverride, false);
            if (pOptions.onAfterRender != null) {
                pOptions.onAfterRender();
            }
            //restore original rendering
            this.requestRenderFrame();
            this.reomveScreenshotOverlayCanvas(tOverlayCanvas);
        }
    }
    markScreenshotMeshes(pMeshes, pValue) {
        if (pMeshes != null) {
            for (const tMesh of pMeshes) {
                tMesh._wcf_core_isScreenshotMesh = pValue;
            }
        }
    }
    createScreenshotOverlayCanvas() {
        const tCanvas = document.createElement("canvas");
        tCanvas.width = this.mCanvas.width;
        tCanvas.height = this.mCanvas.height;
        tCanvas.style.position = "absolute";
        tCanvas.style.top = "0px";
        tCanvas.style.left = "0px";
        tCanvas.style.width = "100%";
        tCanvas.style.height = "100%";
        const tContext = tCanvas.getContext("2d");
        if (tContext == null)
            throw new Error("Failed to get canvas 2d context");
        tContext.drawImage(this.mCanvas, 0, 0);
        this.mView.overlayDiv.insertBefore(tCanvas, this.mView.overlayDiv.firstChild);
        this.mCanvas.style.visibility = "hidden";
        return (tCanvas);
    }
    reomveScreenshotOverlayCanvas(pCanvas) {
        this.mCanvas.style.visibility = "visible";
        this.mView.overlayDiv.removeChild(pCanvas);
    }
    async dumpFramebuffer(pWidth, pHeight, pOptions, pAsBlob = false) {
        const tMimeType = pOptions.mimeType != null ? pOptions.mimeType : "image/png";
        //Create a 2D canvas to store the result
        let tCanvas = pOptions.targetCanvas;
        if (tCanvas == null) {
            if (this.mDumpFrameBufferCanvas == null)
                this.mDumpFrameBufferCanvas = document.createElement('canvas');
            tCanvas = this.mDumpFrameBufferCanvas;
        }
        tCanvas.width = pWidth;
        tCanvas.height = pHeight;
        const tContext = tCanvas.getContext('2d');
        if (tContext == null)
            throw new Error("Failed to get canvas 2d context");
        //draw background
        if (pOptions.onDrawBackground != null && this.mScene.clearColor.a < 0.99) {
            await pOptions.onDrawBackground(tContext);
        }
        //draw the screenshot on the 2D canvas
        tContext.drawImage(this.mCanvas, 0, 0);
        //draw overlay
        if (pOptions.onDrawOverlay != null)
            await pOptions.onDrawOverlay(tContext);
        //resolve with an empty string if a target canvas was defined
        if (pOptions.targetCanvas != null) {
            return ("");
        }
        if (pAsBlob) {
            return (canvasToBlob(tCanvas, tMimeType));
        }
        else {
            return (tCanvas.toDataURL(tMimeType));
        }
    }
}
Viewer.MAX_DOUBLE_CLICK_DELAY = 300;
Viewer.POINTER_MOVE_TOLERANCE = (window.devicePixelRatio != null && window.devicePixelRatio > 1.0) ?
    window.devicePixelRatio * 3 : 3;
//// private ///////////////////////////////////////////////////////////////////////////////
Viewer.MSAA_DELAY_FRAMES = 20;
Viewer.MSAA_MAX_QUALITY = 32;
Viewer.MIN_CANVAS_SIZE = 4;
