import { ListenableEvent } from "../utils/ListenableEvent";
import { BasketService } from "./basket";
import { CatalogService } from "./catalog";
import { ProjectService } from "./project";
import type { SessionCacheFileType } from "./session";
import { SessionService } from "./session";
/**
 * Options for `EaiwsSession.open()`
 */
export interface OpenSessionOptions {
    /** name of the startup file for the session */
    startup?: string;
    /** list of package groups for the session */
    packageGroups?: string;
    /** user token which will be used to authorize the session */
    userToken?: string;
    /** optional locale which can be used to override the locale defined in the startup file */
    locale?: string;
    /** data domain for this session, used from project store */
    dataDomain?: string;
    /** Interval in milliseconds which is used to call `keepAlive()` at the session. A value of 0 or null will disabled the autmoatic keep alive handling. (default: 60000) */
    keepAliveInterval?: number | null;
}
export declare class EaiwsSession {
    get baseUrl(): string;
    get sessionId(): string;
    get session(): SessionService;
    get catalog(): CatalogService;
    get basket(): BasketService;
    get project(): ProjectService;
    get isValid(): boolean;
    /**
     * Will be triggered if this instance was connected to an EAIWS session.
     */
    readonly eventConnected: ListenableEvent<void, EaiwsSession>;
    /**
     * Will be triggered if this instance was disconnected from an EAIWS session.
     */
    readonly eventDisconnected: ListenableEvent<void, EaiwsSession>;
    /**
     * Will be triggered from `session.setLocale()`.
     */
    readonly eventSessionLocaleChanged: ListenableEvent<void, EaiwsSession>;
    /**
     * Will be triggered from `catalog.setLanguages()`.
     */
    readonly eventCatalogLanguagesChanged: ListenableEvent<void, EaiwsSession>;
    /**
     * Will be triggered from `basket.setLanguages()`.
     */
    readonly eventBasketLanguagesChanged: ListenableEvent<void, EaiwsSession>;
    constructor();
    /**
     * Opens and connects a new EAIWS session.
     */
    open(pBaseUrl: string, pOptions: OpenSessionOptions): Promise<boolean>;
    /**
     * Closes the currently connected EAIWS session and disconnects.
     */
    close(): Promise<boolean>;
    /**
     * Connects to an already opened EAIWS session. If pKeepAliveInterval is null or 0 the keep
     * alive handling will be disabled (default: 60000).
     */
    connect(pBaseUrl: string, pSessionId: string, pKeepAliveInterval?: number | null): boolean;
    /**
     * Disconnects from the currently connected EAIWS session. The session will not be closed.
     */
    disconnect(): boolean;
    uploadFileToUrl(pTargetUrl: string, pFile: Blob, pOnProgress?: (pEvent: ProgressEvent, pXhr: XMLHttpRequest) => any): Promise<string>;
    uploadFile(pType: SessionCacheFileType, pFile: Blob, pOnProgress?: (pEvent: ProgressEvent, pXhr: XMLHttpRequest) => any, pUrlSuffix?: string): Promise<string | null>;
    deleteFile(pUrl: string): Promise<void>;
    private mBaseUrl;
    private mSessionId;
    private mKeepAliveInterval;
    private mSession;
    private mCatalog;
    private mBasket;
    private mProject;
    private keepAlive;
}
