import { ajax, AjaxAbortedError } from "../utils/async";
import { ERR_INVALID_STATE, rejectError } from "../utils/error";
import { ListenableEvent } from "../utils/ListenableEvent";
import { isNotNullOrEmpty, isNullOrEmpty } from "../utils/string";
import { BasketService } from "./basket";
import { CatalogService } from "./catalog";
import { ProjectService } from "./project";
import { SessionService, StringPair } from "./session";
export class EaiwsSession {
    get baseUrl() {
        return (this.mBaseUrl);
    }
    get sessionId() {
        return (this.mSessionId);
    }
    get session() {
        return (this.mSession);
    }
    get catalog() {
        return (this.mCatalog);
    }
    get basket() {
        return (this.mBasket);
    }
    get project() {
        return (this.mProject);
    }
    get isValid() {
        return (this.mSession != null);
    }
    constructor() {
        /**
         * Will be triggered if this instance was connected to an EAIWS session.
         */
        this.eventConnected = new ListenableEvent(this);
        /**
         * Will be triggered if this instance was disconnected from an EAIWS session.
         */
        this.eventDisconnected = new ListenableEvent(this);
        /**
         * Will be triggered from `session.setLocale()`.
         */
        this.eventSessionLocaleChanged = new ListenableEvent(this);
        /**
         * Will be triggered from `catalog.setLanguages()`.
         */
        this.eventCatalogLanguagesChanged = new ListenableEvent(this);
        /**
         * Will be triggered from `basket.setLanguages()`.
         */
        this.eventBasketLanguagesChanged = new ListenableEvent(this);
        this.mBaseUrl = undefined;
        this.mSessionId = undefined;
        this.mSession = undefined;
        this.mCatalog = undefined;
        this.mBasket = undefined;
        this.mProject = undefined;
        this.mKeepAliveInterval = null;
    }
    /**
     * Opens and connects a new EAIWS session.
     */
    open(pBaseUrl, pOptions) {
        const tArgs = new Array();
        if (isNotNullOrEmpty(pOptions.startup)) {
            tArgs.push(new StringPair("-startup", pOptions.startup));
        }
        if (isNotNullOrEmpty(pOptions.packageGroups)) {
            tArgs.push(new StringPair("-package-groups", pOptions.packageGroups));
        }
        if (isNotNullOrEmpty(pOptions.userToken)) {
            tArgs.push(new StringPair("-user-token", pOptions.userToken));
        }
        if (isNotNullOrEmpty(pOptions.locale)) {
            tArgs.push(new StringPair("-locale", pOptions.locale));
        }
        if (isNotNullOrEmpty(pOptions.dataDomain)) {
            tArgs.push(new StringPair("-domain", pOptions.dataDomain));
        }
        const tPromise = SessionService.OpenSession(pBaseUrl, tArgs).then((pId) => {
            return (this.connect(pBaseUrl, pId, pOptions.keepAliveInterval));
        });
        return (tPromise);
    }
    /**
     * Closes the currently connected EAIWS session and disconnects.
     */
    close() {
        if (this.mSession == null)
            return (Promise.resolve(false));
        const tPromise = this.mSession.closeSession().then(() => {
            return (this.disconnect());
        });
        return (tPromise);
    }
    /**
     * Connects to an already opened EAIWS session. If pKeepAliveInterval is null or 0 the keep
     * alive handling will be disabled (default: 60000).
     */
    connect(pBaseUrl, pSessionId, pKeepAliveInterval = 60000) {
        if (isNullOrEmpty(pSessionId))
            return (false);
        this.disconnect();
        this.mBaseUrl = pBaseUrl;
        this.mSessionId = pSessionId;
        this.mSession = new SessionService(this, this.mBaseUrl, this.mSessionId);
        this.mCatalog = new CatalogService(this, this.mBaseUrl, this.mSessionId);
        this.mBasket = new BasketService(this, this.mBaseUrl, this.mSessionId);
        this.mProject = new ProjectService(this, this.mBaseUrl, this.mSessionId);
        if (pKeepAliveInterval != null && pKeepAliveInterval > 0) {
            this.mKeepAliveInterval = window.setInterval(() => {
                this.keepAlive();
            }, pKeepAliveInterval);
        }
        this.eventConnected.trigger();
        return (true);
    }
    /**
     * Disconnects from the currently connected EAIWS session. The session will not be closed.
     */
    disconnect() {
        if (this.mSession == null)
            return (false);
        this.mBaseUrl = undefined;
        this.mSessionId = undefined;
        this.mSession = undefined;
        this.mCatalog = undefined;
        this.mBasket = undefined;
        this.mProject = undefined;
        if (this.mKeepAliveInterval != null) {
            clearInterval(this.mKeepAliveInterval);
            this.mKeepAliveInterval = null;
        }
        this.eventDisconnected.trigger();
        return (true);
    }
    async uploadFileToUrl(pTargetUrl, pFile, pOnProgress) {
        try {
            const tBeforeAjax = (pXhr) => {
                if (pOnProgress != null && pXhr.upload != null) {
                    /* eslint-disable-next-line @typescript-eslint/no-unsafe-return */
                    pXhr.upload.onprogress = (pEvent) => pOnProgress(pEvent, pXhr);
                }
            };
            await ajax("PUT", pTargetUrl, pFile, {
                timeout: 0,
                responseType: "text",
                headers: {
                    "Content-Type": pFile.type
                }
            }, undefined, tBeforeAjax);
            return (pTargetUrl);
        }
        catch (pError) {
            if (pError instanceof AjaxAbortedError) {
                return ("");
            }
            else {
                throw pError;
            }
        }
    }
    uploadFile(pType, pFile, pOnProgress, pUrlSuffix) {
        if (!this.isValid)
            return (rejectError(ERR_INVALID_STATE, "session not valid"));
        const tPromise = this.mSession.getUploadURL(pType, pUrlSuffix).then((pUrl) => {
            if (pUrl != null && pUrl !== "") {
                return (this.uploadFileToUrl(pUrl, pFile, pOnProgress));
            }
            else {
                return (Promise.resolve(null));
            }
        });
        return (tPromise);
    }
    deleteFile(pUrl) {
        return (ajax("DELETE", pUrl, null, { responseType: "text" }));
    }
    //// internal //////////////////////////////////////////////////////////////////////////////
    /** @internal */
    _clearCache() {
        if (this.isValid) {
            this.mSession._clearCache();
            this.mCatalog._clearCache();
            this.mBasket._clearCache();
            this.mProject._clearCache();
        }
    }
    keepAlive() {
        if (this.isValid) {
            void this.mSession.keepAlive();
        }
    }
}
