import { CustomError } from "../error";
export type AjaxMethod = "GET" | "POST" | "PUT" | "DELETE" | "PATCH" | "HEAD";
export type AjaxDataType = "post" | "queryString" | "json" | "text" | "arraybuffer" | "blob" | "document" | "formdata" | "xml";
export interface AjaxOptions {
    /**
     * Defines the type of the request data.
     * Available types: `post`, `queryString`, `json`, `text`, `arraybuffer`, `blob`, `document`, `xml` or `formdata` (default: post)
     */
    dataType?: AjaxDataType;
    /**
     * The response type.
     * Available types: `auto`, `json`, `xml`, `text`, `arraybuffer`, `blob` or `document` (default: `auto`)
     * Note: use `xml` to parse the response always as xml using the DOMParser. If `document` is used the response will be parsed
     * as xml or as html by the browser depending on the content-type of the response.
     */
    responseType?: XMLHttpRequestResponseType | "xml" | "auto";
    /** browser caching (default: true) */
    cache?: boolean;
    /** the user to access to the URL, if needed */
    user?: string;
    /** the password to access to the URL, if needed */
    password?: string;
    /** object containing headers to be sent */
    headers?: {
        [key: string]: string;
    };
    /** sends credentials with the XHR request (default: false) */
    withCredentials?: boolean;
    /** the timeout for the request in ms (default: 30000) */
    timeout?: number;
    /** defines how many retry attempts will be performed in case of connection errors (default: 3) */
    retryAttempts?: number;
    /** if true the global `ajaxErrorHandler` will not be triggered in case of an error (default: false) */
    ignoreGlobalErrorHandler?: boolean;
}
/**
 * Will be thrown if the ajax request was aborted.
 */
export declare class AjaxAbortedError extends CustomError {
    constructor(pUrl: string);
}
/**
 * Will be thrown for ajax request timeouts.
 */
export declare class AjaxTimeoutError extends CustomError {
    constructor(pUrl: string);
}
/**
 * Will be thrown for ajax connection errors.
 */
export declare class AjaxConnectionError extends CustomError {
    constructor(pUrl: string);
}
export interface AjaxPromise<T> extends Promise<T> {
    /**
     * Aborts the currently running ajax request.
     * In this case the ajax promise will be rejected with `AjaxAbortedError`.
     */
    abort(): void;
}
export interface AjaxRequestInfo {
    url: string;
    method: AjaxMethod;
    data?: unknown;
    options?: AjaxOptions;
}
export interface AjaxErrorInfo {
    status: number;
    textStatus: string;
    errorThrown: Error | string;
    responseError: Error;
    requestInfo: AjaxRequestInfo;
    /** Shows how many retries were already performed. 0 for the first error. */
    retryCount: number;
    /** If called the ajax call will be executed again. */
    retry(): void;
    /** If retry was not called reject has to be called to reject the underlying promise. */
    reject(pCustomError?: Error): void;
}
export type AjaxCustomErrorCallback = (pError: Error | string, pXhr: XMLHttpRequest, pResponse: unknown) => string | Error | null;
export type AjaxBeforeCallback = (pXhr: XMLHttpRequest) => void;
/**
 * Wrapper function for XMLHttpRequests with optional custom error handling.
 * pErrorMessageCallback can be used to return a custom error message.
 */
export declare function ajax<T = unknown>(pMethod: AjaxMethod, pUrl: string, pData?: unknown, pOptions?: AjaxOptions, pCustomErrorCallback?: AjaxCustomErrorCallback, pBeforeCallback?: AjaxBeforeCallback): AjaxPromise<T>;
/**
 * This function performs an HEAD request to fetch the content-type of the given url.
 */
export declare function fetchContentType(pUrl: string): Promise<string | null>;
