import { wcfConfig } from "../WcfConfig";
import { CustomError } from "../error";
import { isNotNullOrEmpty } from "../string";
import { AjaxRequest } from "./internal/AjaxRequest";
/**
 * Will be thrown if the ajax request was aborted.
 */
export class AjaxAbortedError extends CustomError {
    constructor(pUrl) {
        super("request aborted: " + pUrl);
        Object.setPrototypeOf(this, AjaxAbortedError.prototype);
        this.name = "AjaxAbortedError";
    }
}
/**
 * Will be thrown for ajax request timeouts.
 */
export class AjaxTimeoutError extends CustomError {
    constructor(pUrl) {
        super("request timeout: " + pUrl);
        Object.setPrototypeOf(this, AjaxTimeoutError.prototype);
        this.name = "AjaxTimeoutError";
    }
}
/**
 * Will be thrown for ajax connection errors.
 */
export class AjaxConnectionError extends CustomError {
    constructor(pUrl) {
        super("request connection interrupted: " + pUrl);
        Object.setPrototypeOf(this, AjaxConnectionError.prototype);
        this.name = "AjaxConnectionError";
    }
}
/**
 * Wrapper function for XMLHttpRequests with optional custom error handling.
 * pErrorMessageCallback can be used to return a custom error message.
 */
export function ajax(pMethod, pUrl, pData, pOptions, pCustomErrorCallback, pBeforeCallback) {
    var _a, _b;
    const tOptions = pOptions == null ? {} : pOptions;
    const tMaxRetries = (_a = tOptions.retryAttempts) !== null && _a !== void 0 ? _a : wcfConfig.ajaxDefaultOptions.retryAttempts;
    const tIgnoreGlobalErrorHandler = (_b = tOptions.ignoreGlobalErrorHandler) !== null && _b !== void 0 ? _b : wcfConfig.ajaxDefaultOptions.ignoreGlobalErrorHandler;
    let tCurrentRequest = null;
    const getRequestInfo = () => {
        return ({
            url: pUrl,
            method: pMethod,
            data: pData,
            options: pOptions
        });
    };
    const rejectRequest = (pRejectData, pResponseError, pReject) => {
        if (wcfConfig.ajaxRejectedEvent != null) {
            wcfConfig.ajaxRejectedEvent(getRequestInfo(), pRejectData.xhr, pResponseError);
        }
        pReject(pResponseError);
    };
    const executeRequest = (pResolve, pReject, pTotalRetryCount, pRetryCount) => {
        tCurrentRequest = new AjaxRequest(pMethod, pUrl, pData, tOptions, pBeforeCallback);
        tCurrentRequest.execute().then((pResolveData) => {
            tCurrentRequest = null;
            if (wcfConfig.ajaxResolvedEvent != null) {
                wcfConfig.ajaxResolvedEvent(getRequestInfo(), pResolveData.xhr);
            }
            pResolve(pResolveData.response);
        }).catch((pRejectData) => {
            tCurrentRequest = null;
            //check if request was aborted, in this case directly foward the error
            //do not trigger ajaxRejectedEvent for aborted requests
            if (pRejectData.error instanceof AjaxAbortedError) {
                pReject(pRejectData.error);
                return;
            }
            //in case of network errors automatically retry the request
            if (pRejectData.error instanceof AjaxConnectionError && pRetryCount < tMaxRetries) {
                setTimeout(() => {
                    executeRequest(pResolve, pReject, pTotalRetryCount + 1, pRetryCount + 1);
                }, wcfConfig.ajaxGetRetryDelay(pRetryCount + 1));
                return;
            }
            const tResponseError = getResponseError(pRejectData, pCustomErrorCallback);
            //delegate error information to error handler
            if (wcfConfig.ajaxErrorHandler != null && !tIgnoreGlobalErrorHandler) {
                const tErrorInfo = {
                    status: pRejectData.xhr.status,
                    textStatus: pRejectData.xhr.statusText,
                    errorThrown: pRejectData.error,
                    responseError: tResponseError,
                    requestInfo: getRequestInfo(),
                    retryCount: pTotalRetryCount,
                    retry: () => executeRequest(pResolve, pReject, pTotalRetryCount + 1, 0),
                    reject: (pCustomError) => rejectRequest(pRejectData, pCustomError !== null && pCustomError !== void 0 ? pCustomError : tResponseError, pReject)
                };
                wcfConfig.ajaxErrorHandler(tErrorInfo);
            }
            else {
                rejectRequest(pRejectData, tResponseError, pReject);
            }
        });
    };
    const tPromise = new Promise((pResolve, pReject) => {
        executeRequest(pResolve, pReject, 0, 0);
    });
    tPromise.abort = () => {
        if (tCurrentRequest != null) {
            tCurrentRequest.abort();
        }
    };
    return (tPromise);
}
/**
 * This function performs an HEAD request to fetch the content-type of the given url.
 */
export async function fetchContentType(pUrl) {
    let tXhr;
    let tType = null;
    await ajax("HEAD", pUrl, undefined, undefined, undefined, (pXhr) => tXhr = pXhr);
    if (tXhr != null) {
        tType = tXhr.getResponseHeader("Content-Type");
    }
    return (tType);
}
function getResponseError(pRejectData, pCustomErrorCallback) {
    const tErrorMessage = pRejectData.error instanceof Error ? pRejectData.error.message : pRejectData.error;
    let tResponseError = null;
    if (pCustomErrorCallback != null)
        tResponseError = pCustomErrorCallback(pRejectData.error, pRejectData.xhr, pRejectData.response);
    if (tResponseError == null) {
        tResponseError = "[" + pRejectData.xhr.status + "] " + pRejectData.xhr.statusText;
        if (isNotNullOrEmpty(tErrorMessage))
            tResponseError += " - " + tErrorMessage;
    }
    if (tResponseError instanceof Error)
        return (tResponseError);
    else
        return Error(tResponseError);
}
